#include <iostream>
#include <pthread.h>
#include <errno.h>

void PrintError(char* msg, int status, char* fileName, int lineNumber) { 
	std::cout << msg << ' ' << fileName << ":" << lineNumber 
	<< " - " << strerror(status) << std::endl;
}
void* simpleThread (void* myID) { // myID is the fourth argument of pthread_create ()
	std::cout << "Thread " << (long) myID << " is running" << std::endl;
	return NULL;		// implicit call to pthread_exit(NULL);
}

int main() {
	pthread_t threadArray[2];	// array of thread IDs
	int status;	// error code
	pthread_attr_t threadAttribute;	// thread attribute
	status = pthread_attr_init(&threadAttribute);	// initialize attribute object
	if (status != 0) { PrintError("pthread_attr_init failed at", status, __FILE__, 
	__LINE__); exit(status);}			
	// set the scheduling scope attribute 
	status = pthread_attr_setscope(&threadAttribute,
	PTHREAD_SCOPE_SYSTEM);
	if (status != 0) { PrintError("pthread_attr_setscope failed at", status, __FILE__, 
	__LINE__); exit(status);}		
	// Create two threads and store their IDs in array threadArray
	status = pthread_create(&threadArray[0], &threadAttribute, simpleThread, 
	(void*) 1L);
	if (status != 0) { PrintError("pthread_create failed at", status, __FILE__, 
	__LINE__); exit(status);}		
	status = pthread_create(&threadArray[1], &threadAttribute, simpleThread, 
	(void*) 2L);
	if (status != 0) { PrintError("pthread_create failed at", status, __FILE__, 
	__LINE__); exit(status);}		
	status = pthread_attr_destroy(&threadAttribute); // destroy the attribute object
	if (status != 0) { PrintError("pthread_attr_destroy failed at", status, __FILE__, 
	__LINE__); exit(status);}		
	status = pthread_join(threadArray[0],NULL); // wait for threads to finish
	if (status != 0) { PrintError("pthread_join failed at", status, __FILE__,
	 __LINE__); exit(status);}		
	status = pthread_join(threadArray[1],NULL);
	if (status != 0) { PrintError("pthread_join failed at", status, __FILE__, 
	__LINE__); exit(status);}		
}

//Listing 1.4 A simple concurrent program using C++/Pthreads.

