#ifndef THREAD_H
#define THREAD_H

#include "runnable.h"
#include <pthread.h>
#include <memory>


class Thread {
public:
	Thread(std::auto_ptr<Runnable> runnable_, bool isDetached = false);
	Thread(bool isDetached = false);
	virtual ~Thread();
	void start();
	void* join();
private:
	pthread_t PthreadThreadID;	// thread ID
	bool detached;	// true if thread created in detached state; false otherwise
	pthread_attr_t threadAttribute;
	std::auto_ptr<Runnable> runnable;
	Thread(const Thread&);
	const Thread& operator= (const Thread&);
	void setCompleted();
	void* result;	// stores return value of run()
	virtual void* run() {}
	static void* startThreadRunnable(void* pVoid);
	static void* startThread(void* pVoid);
	void PrintError(char* msg, int status, char* fileName, int lineNumber);
};

#endif
