#pragma warning(disable:4786) 
#include "semaphore.h"
#include <queue>
#include "win32Semaphore.h"
#include "binarySemaphore.h"
#include "mutexLocker.h"
#include "SemaphoreIDGenerator.h"
#include "threadData.h"
#include "TDThread.h"
#include "deadlockWatchdog.h"
#include <assert.h>
#include <typeinfo>
#include <string.h>
#include <iostream>


binarySemaphore::binarySemaphore(int initialPermits, bool traceAndReplay)
 : semaphore(initialPermits) {
	if (traceAndReplay) {
	 char* env = getenv("MODE");
	 if (env!=NULL) {
		if (!(strcmp(env,"TRACE")))
			mode = TRACE;
		else if (!(strcmp(env,"REPLAY")))
			mode = REPLAY;
		else if (!(strcmp(env,"TEST")))
			mode = TEST;
		else if (!(strcmp(env,"SPECTEST")))
			mode = SPECTEST;
		else if (!(strcmp(env,"NONE")))
			mode = NONE;
		else mode = NONE;
	 }
	 else mode = NONE;
 	 if (mode!=TRACE && mode!=REPLAY  && mode != NONE) {
		std::cout << "Error: Only modes TRACE, REPLAY, and NONE are supported by semaphores." << std::endl;
		exit(1);
	 }
	 env = getenv("CONTROLLERS");
	 if (env!=NULL) {
		if (!(strcmp(env,"SINGLE")))
			numControllers = 1;
		else numControllers = 0;
	 }
	 else numControllers = 1;
	 env = getenv("RANDOMDELAY");
	 if (env!=NULL) {	
		 if	(!(strcmp(env,"ON")))
			 randomDelay = true;
		 else
			 randomDelay = false;
	 }
	 else randomDelay = false;

	 env = getenv("DEADLOCKDETECTION");
	 if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			detectDeadlock = true;
		else
			detectDeadlock = false;
	 }
	 else detectDeadlock = false;

	 if (mode) {
		semaphoreName = generateID();
		std::string traceFileName;
		if (numControllers == 0) 
			traceFileName = semaphoreName;
		else
			traceFileName = "semaphoreandlock";
		if (mode == TRACE || mode == REPLAY)
			if (numControllers == 1) 
				control = instance(mode,numControllers,traceFileName);
			else
				control = new semaphoreAndLockTracingAndReplay(mode,numControllers,
							traceFileName);
	 }
	}
	else {mode = NONE; randomDelay = false;}
}

binarySemaphore::~binarySemaphore() {
	if (mode == TRACE || mode == REPLAY)
		if (numControllers = 0) 
		  delete control;
	// else control is a singleton shared by many threads - need to ensure it is 
	// deleted only once if you decide to delete it (need to lock control)

}

std::string binarySemaphore::generateID() {
		const std::string thisClassName = "binarySemaphore";
		data d = threadData::instance().getThreadData(GetCurrentThreadId());	
		std::string semaphoreName;
		if (d.threadID != -1) 
		  semaphoreName = d.threadName;
		else
		  semaphoreName = "";
		int position = semaphoreName.find("_"); 
		if (position > -1 && (size_t) position < semaphoreName.length())
		  semaphoreName = semaphoreName + "_" + thisClassName;
		else
		  semaphoreName = "main_" + thisClassName;		
		std::string myID = SemaphoreIDGenerator::instance().getID(semaphoreName);
		return myID;
}


void binarySemaphore::P() {
   if (mode == REPLAY) {
     	control->requestPermit((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
   }
   else if (randomDelay && mode == TRACE) {
	   Sleep(rand()%delayAmount);	// default delayAmount is 750
   }
   lock();
   if (permits==1)  { // then no need to block thread
     if (waitingV.size() > 0) {  // signal thread blocked in V()
       win32Semaphore* oldestConditionVar = (waitingV.front()).s;
	   int oldestID = (waitingV.front()).ID;
       waitingV.pop();
	   if (mode == REPLAY)
		   assert(false);
	   if (mode == TRACE) {
		   //std::cout << "bSem tracing P" << std::endl;
    	   control->trace((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
   		   control->trace(oldestID); // ID of thread that did V
	   }
       oldestConditionVar->V();
	   if (detectDeadlock && (mode == RT || mode == TRACE)) {
		  TDThread::decNumBlockedThreads();
	   }
	 }
     else {
       permits=0;
	   if (mode == REPLAY)
	  	   control->releasePermit();
	   else if (mode == TRACE) {
		   //std::cout << "bSem tracing P" << std::endl;
    	   control->trace((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
	   }
	 }
	 unlock();
     return;
   }
   waitRecord rec;
   rec.s = pool.acquire();
   rec.ID = (threadData::instance().getThreadData(GetCurrentThreadId())).threadID;

   if (detectDeadlock && (mode == RT || mode == TRACE)) {
		TDThread::incNumBlockedThreads();
		deadlockWatchdog::changeStatus(GetCurrentThreadId(),"blocked on P operation of "+semaphoreName);
   }

   waitingP.push(rec); // otherwise append blocked thread
   unlock();
   rec.s->P();
   if (detectDeadlock && (mode == RT || mode == TRACE)) {
		deadlockWatchdog::changeStatus(GetCurrentThreadId(),"running");
   }
   if (mode == REPLAY)
       assert(false);
   pool.release(rec.s);
} // end P()

void binarySemaphore::V() {
   if (mode == REPLAY)
       control->requestPermit((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
   else if (randomDelay && mode == TRACE) {
	   Sleep(rand()%delayAmount);	// default delayAmount is 750
   }
   lock(); // lock semaphore
   if (permits==0) { // then no need to block thread
     if (waitingP.size() > 0) {  // signal thread blocked in P()
       win32Semaphore* oldestConditionVar = (waitingP.front()).s;
	   int oldestID = (waitingP.front()).ID;
       waitingP.pop();
	   if (mode == REPLAY)
		   assert(false);
	   if (mode == TRACE) { 
    	   control->trace((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
   		   control->trace(oldestID); // ID of thread that did P
	   }
       oldestConditionVar->V();
	   if (detectDeadlock && (mode == RT || mode == TRACE)) {
		  TDThread::decNumBlockedThreads();
	   }
     }
     else {
       permits=1;
	   if (mode == REPLAY)
	 	   control->releasePermit();
	   else if (mode == TRACE) 
    	   control->trace((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
	 }
	 unlock();
     return;
   }
   waitRecord rec;
   rec.s = pool.acquire();
   rec.ID = (threadData::instance().getThreadData(GetCurrentThreadId())).threadID;

   if (detectDeadlock && (mode == RT || mode == TRACE)) {
		TDThread::incNumBlockedThreads();
		deadlockWatchdog::changeStatus(GetCurrentThreadId(),"blocked on V operation of "+semaphoreName);
   }

   waitingV.push(rec); // otherwise append blocked thread
   unlock(); // end synchronized (this) to avoid conditionVar.wait() while holding lock on this
   rec.s->P();
   if (detectDeadlock && (mode == RT || mode == TRACE)) {
		deadlockWatchdog::changeStatus(GetCurrentThreadId(),"running");
   }
   if (mode == REPLAY)
       assert(false);
   pool.release(rec.s);
} // end V()


bool binarySemaphore::doP() {
// Called by VP() operation; checks permit and unblocks a thread waiting
// in V() if necessary. Returns true if P() should block; false otherwise.
   if (permits==1)  { // P() does not block
     if (waitingV.size() > 0) {  // signal thread blocked in V()
       win32Semaphore* oldestConditionVar = (waitingV.front()).s;
	   int oldestID = (waitingV.front()).ID;
       waitingV.pop();
       oldestConditionVar->V();
     }
     else
       permits=0;
     return false;
    } // P() does not block
    else
       return true;	// P() does block
}

void binarySemaphore::VP(semaphore* vSem) {
// execute {vSem.V(); this.P();} without any intervening 
// P() or V() operations on this or vSem.	

// lock semaphores in ascending order of IDs to prevent circular deadlock (i.e. T1 holds
// this's lock and waits for vSem's lock while T2 holds vSem's lock and waits for this's lock.)
   semaphore* first = this;
   semaphore* second = vSem;
   if (this->getSemaphoreID() > vSem->getSemaphoreID()) { 
     first = vSem;
     second = this;
   }
   first->lock(); 
   second->lock();
   //vSem.V() must not block
   //if (!strcmp(typeid(*vSem).name(),"binarySemaphore") && vSem->permits==1)
	if (vSem->permits==1 && vSem->isBinarySemaphore())
	   std::cout << "illegal VP operation" << std::endl;
   // perform vSem.V()
   vSem->V(); // it's okay that we already hold vSem's lock
   // perform this.P()
   bool blockingP = doP();
   if (!blockingP) {
     second->unlock();
     first->unlock();
     return;
   }
   waitRecord rec;
   rec.s = pool.acquire();
   rec.ID = (threadData::instance().getThreadData(GetCurrentThreadId())).threadID;
   waitingP.push(rec); // otherwise append blocked thread
   second->unlock();
   first->unlock();
   rec.s->P();
   pool.release(rec.s);
} // end VP()

