#include "lockData.h"
#include <stddef.h>
#include <windows.h>
#include "criticalSectionLock.h"
#include "mutexLocker.h"
#include <iostream>
#include <assert.h>
#include "threadData.h"

criticalSectionLock lockData::instanceLock;


lockData& lockData::instance(){
	mutexLocker<criticalSectionLock> mutex(instanceLock);  
	static lockData instance;
	return instance;
  }
	
std::multiset<int> lockData::getLocked(unsigned long currentThread) {
	mutexLocker<win32Critical_Section> mutex(dataLock);

	std::map<unsigned long, std::multiset<int>, ltms>::iterator pos;
	std::multiset<int> locksOfCurrentThread;

	pos = locked.find(currentThread);
	// get locks owned by currentThread
	if (pos != locked.end())
	 locksOfCurrentThread = pos->second;

	return locksOfCurrentThread;
}

void lockData::addLocked(int ID) {
// need to determine current thread  and add to its locked

  mutexLocker<win32Critical_Section> mutex(dataLock);
  unsigned long currentThread = GetCurrentThreadId();
  std::map<unsigned long, std::multiset<int>, ltms>::iterator pos;

  pos = locked.find(currentThread);
  if (pos != locked.end()) 
	  // currentThread already hold some locks, add lock ID
		// Note: this may be duplicate since a thread can lock 
		// a mutex more than once if the thread owns the mutex,
    pos->second.insert(ID);
  else {
		// first lock owned by currentThread so add entry for thread
		// which is pair (currentThread,multiset of lock IDs)
		std::multiset<int> m; 
		m.insert(ID);
		locked.insert(std::make_pair(currentThread,m));  
  }
  //std::cout << ID << " added to locked of " << 
	  //threadData::instance().getThreadData(currentThread).threadID << ", now : ";
  //pos = locked.find(currentThread);
  //for (std::multiset<int>::iterator it = pos->second.begin();
  //    it != pos->second.end(); it++)
  //		   std::cout << *it << " ";
  //	   std::cout << std::endl;
}

void lockData::removeLocked(int ID) {
// need to remove from current thread's locked
    bool found = false;
    mutexLocker<win32Critical_Section> mutex(dataLock);
	unsigned long currentThread = GetCurrentThreadId();

	std::map<unsigned long, std::multiset<int>, ltms>::iterator pos;

	pos = locked.find(currentThread);
	// Q:assuming that currentThread owns lockID sine alled from unlock()
	// remove lock ID from currentThreads ID set
	if (pos == locked.end())  {
		std::cout << "Warning: unlocked but no outstanding locked." << std::endl;
		// there was a user error: unlock without prior lock
		return;
	}

	pos->second.erase(ID);
	//std::cout << ID << " removed from locked of " << 
	//	threadData::instance().getThreadData(currentThread).threadID << ", now : ";
	//pos = locked.find(currentThread);
	//for (std::multiset<int>::iterator it = pos->second.begin();
    //   it != pos->second.end(); it++)
	//	   std::cout << *it << " ";
	//   std::cout << std::endl;
 
}
