#include "semaphoreAndLockTracingAndReplay.h"
#include "mutexLocker.h"
#include <iostream>	

criticalSectionLock semaphoreAndLockTracingAndReplay::instanceLock;

semaphoreAndLockTracingAndReplay::semaphoreAndLockTracingAndReplay(int mode_, int numControllers_, 
																   std::string traceFileName_) 
	: mode(mode_), numControllers(numControllers_), traceFileName(traceFileName_) {
		mutex = new win32Semaphore(1);

		if (mode == TRACE) { // trace
			index = 0;	
			outfile.open((traceFileName+"-replay.txt").c_str());
			if (outfile.fail()) {
				std::cout << "Error: Trace output file " 
						   << (traceFileName+"-replay.txt") << " not opened." << std::endl;
				exit(1);
			}
		}
	    else if (mode == REPLAY) { // replay
			//std::cout << "Reading trace file." << std::endl;
			index = 0;
			numberOfThreads = -1;
			infile.open((traceFileName+"-replay.txt").c_str());
			if (infile.fail()) {
				std::cout << "Warning: Trace input file " 
						  << (traceFileName+"-replay.txt") << " not opened." << std::endl;
				std::cout << "(Ignore this warning if there are no semaphores or locks in the program.)" << std::endl;
				//exit(1);
				return;
			}
			int ID;
			int max=0;
			while (infile >> ID) {
				SYNsequence.push_back(ID);
				if (ID>max) max = ID;
			}
			infile.close();
			//std::cout << "LockUnlock-sequence size is " << SYNsequence.size() << std::endl;
			numberOfThreads = max;
			hasRequested = new bool[numberOfThreads+1];
			threads = new win32Semaphore*[numberOfThreads+1];
			for (int i=0; i<(numberOfThreads+1);i++) {
				threads[i] = new win32Semaphore(0); 
				hasRequested[i] = false;
			}
		}
}

semaphoreAndLockTracingAndReplay::~semaphoreAndLockTracingAndReplay() {
	delete mutex;
	if (mode == REPLAY)
		for (int i=0; i<(numberOfThreads+1);i++)
			delete threads[i];
}

semaphoreAndLockTracingAndReplay* semaphoreAndLockTracingAndReplay::instance(int mode, 
												  int numControllers, std::string traceFileName) {
	mutexLocker<criticalSectionLock> mutex(instanceLock);  
	static semaphoreAndLockTracingAndReplay instance(mode,numControllers,traceFileName);
	return &instance;
}

void semaphoreAndLockTracingAndReplay::trace(int ID) {
	mutex->P();
	outfile << ID << std::endl;
    mutex->V();
}

void semaphoreAndLockTracingAndReplay::requestPermit(int ID) {
	mutex->P();
	if (ID != SYNsequence[index]) {
		hasRequested[ID] = true;
		mutex->V();
		threads[ID]->P();
		hasRequested[ID] = false;
	}
	else mutex->V();
}

void semaphoreAndLockTracingAndReplay::releasePermit() {
	mutex->P();
	++index;
	if (index < (int) SYNsequence.size()) {
		if (hasRequested[SYNsequence[index]]) {
			threads[SYNsequence[index]]->V();
		}
	}
	mutex->V();
}
