template<class T>
sharedVariable<T>::~sharedVariable() {if (mode&&tracingIsOn) delete SVcontrol;}

template<class T>
sharedVariable<T>::sharedVariable() : first(true), analysisIsOn(true), tracingIsOn(true),
dataLock(), lockSet(), value(0), writtenAtLeastOnce(false) { 
	 char* env = getenv("MODE");
	 if (env!=NULL) {
		if (!(strcmp(env,"TRACE")))
			mode = TRACE;
		else if (!(strcmp(env,"REPLAY")))
			mode = REPLAY;
		else if (!(strcmp(env,"TEST")))
			mode = TEST;
		else if (!(strcmp(env,"NONE")))
			mode = NONE;
		else mode = NONE;
	 }
	 else mode = NONE;
 	 if (mode!=NONE && mode!=TRACE && mode!=REPLAY) {
		std::cout << "Error: Only TRACE and REPLAY modes are supported by sharedVariables." << std::endl;
		exit(1);
	 }
	 env = getenv("DATARACEDETECTION");
	 if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			raceAnalysis = true;
		else raceAnalysis = false;
	 }
	 else raceAnalysis = false;
	if (raceAnalysis && mode != TRACE) {
		std::cout << "Error: DATARACEDETECTION must be performed in TRACE mode." << std::endl;
		exit(1);
	}
	 if (mode) {
		std::string traceFileName = generateID();
		//std::cout << "sv() traceFileName is " <<  traceFileName << std::endl;
	    SVcontrol = new sharedVariableController(traceFileName+"-replay.txt");
	 }
}

template<class T>
sharedVariable<T>::sharedVariable(analysisOnOff aOnOff, tracingOnOff tOnOff) : first(true),
dataLock(), lockSet(), analysisIsOn((aOnOff==analysisOn)?true:false), 
tracingIsOn((tOnOff==traceOn)?true:false), writtenAtLeastOnce(false) {
	 char* env = getenv("MODE");
	 if (env!=NULL) {
		if (!(strcmp(env,"TRACE")))
			mode = TRACE;
		else if (!(strcmp(env,"REPLAY")))
			mode = REPLAY;
		else if (!(strcmp(env,"TEST")))
			mode = TEST;
		else if (!(strcmp(env,"NONE")))
			mode = NONE;
		else mode = NONE;
	 }
	 else mode = NONE;
	 if (mode!=NONE && mode!=TRACE && mode!=REPLAY) {
		std::cout << "Error: Only TRACE and REPLAY modes are supported by sharedVariables." << std::endl;
		exit(1);
	 }
	 env = getenv("DATARACEDETECTION");
	 if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			raceAnalysis = true;
		else raceAnalysis = false;
	 }
	 else raceAnalysis = false;
	if (raceAnalysis && mode != TRACE) {
		std::cout << "Error: DATARACEDETECTION must be performed in TRACE mode." << std::endl;
		exit(1);
	}
	 if (mode&&tracingIsOn) {
		 std::string traceFileName = generateID();
	     SVcontrol = new sharedVariableController(traceFileName+"-replay.txt");
	 }
}


template<class T>
sharedVariable<T>::sharedVariable(const T& v) 
: value (v), first(true), analysisIsOn(true), tracingIsOn(true), dataLock(), 
  lockSet(), writtenAtLeastOnce(false) {
	 char* env = getenv("MODE");
	 if (env!=NULL) {
		if (!(strcmp(env,"TRACE")))
			mode = TRACE;
		else if (!(strcmp(env,"REPLAY")))
			mode = REPLAY;
		else if (!(strcmp(env,"TEST")))
			mode = TEST;
		else if (!(strcmp(env,"NONE")))
			mode = NONE;
		else mode = NONE;
	 }
	 else mode = NONE;
 	 if (mode!=NONE && mode!=TRACE && mode!=REPLAY) {
		std::cout << "Error: Only TRACE and REPLAY modes are supported by sharedVariables." << std::endl;
		exit(1);
	 }
	 env = getenv("DATARACEDETECTION");
	 if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			raceAnalysis = true;
		else raceAnalysis = false;
	 }
	 else raceAnalysis = false;
	 if (raceAnalysis && mode != TRACE) {
		std::cout << "Error: DATARACEDETECTION must be performed in TRACE mode." << std::endl;
		exit(1);
	 }
	 if (mode) {
		std::string traceFileName = generateID();
	    SVcontrol = new sharedVariableController(traceFileName+"-replay.txt");
	 }
}


template<class T>
sharedVariable<T>::sharedVariable(const T& v, analysisOnOff aOnOff, tracingOnOff tOnOff) 
: value (v), first(true), dataLock(), lockSet(), 
tracingIsOn((tOnOff==sharedVariable<T>::traceOn)?true:false), 
analysisIsOn((aOnOff==sharedVariable<T>::analysisOn)?true:false), writtenAtLeastOnce(false) {
	 char* env = getenv("MODE");
	 if (env!=NULL) {
		if (!(strcmp(env,"TRACE")))
			mode = TRACE;
		else if (!(strcmp(env,"REPLAY")))
			mode = REPLAY;
		else if (!(strcmp(env,"TEST")))
			mode = TEST;
		else if (!(strcmp(env,"NONE")))
			mode = NONE;
		else mode = NONE;
	 }
	 else mode = NONE;
 	 if (mode!=NONE && mode!=TRACE && mode!=REPLAY) {
		std::cout << "Error: Only TRACE and REPLAY modes are supported by sharedVariables." << std::endl;
		exit(1);
	 }
	 env = getenv("DATARACEDETECTION");
	 if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			raceAnalysis = true;
		else raceAnalysis = false;
	 }
	 else raceAnalysis = false;
 	 if (raceAnalysis && mode != TRACE) {
		std::cout << "Error: DATARACEDETECTION must be performed in TRACE mode." << std::endl;
		exit(1);
	 }
	 if (mode&&tracingIsOn) {
		std::string traceFileName = generateID();
	    SVcontrol = new sharedVariableController(traceFileName+"-replay.txt");
	 }
}

// copies have on(false) so that no race analysis occurs on temporaries.
template<class T>
sharedVariable<T>::sharedVariable(const sharedVariable<T>& sv) 
: first(true), analysisIsOn(false), tracingIsOn(false), value(sv.value), dataLock(), 
  lockSet(), writtenAtLeastOnce(false) { 
	char* env = getenv("MODE");
	if (env!=NULL) {
		if (!(strcmp(env,"TRACE")))
			mode = TRACE;
		else if (!(strcmp(env,"REPLAY")))
			mode = REPLAY;
		else if (!(strcmp(env,"TEST")))
			mode = TEST;
		else if (!(strcmp(env,"NONE")))
			mode = NONE;
		else mode = NONE;
	}
	else mode = NONE;
	if (mode!=NONE && mode!=TRACE && mode!=REPLAY) {
		std::cout << "Error: Only TRACE and REPLAY modes are supported by sharedVariables." << std::endl;
		exit(1);
	}
	env = getenv("DATARACEDETECTION");
	if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			raceAnalysis = true;
		else raceAnalysis = false;
	}
	else raceAnalysis = false;
	if (raceAnalysis && mode != TRACE) {
		std::cout << "Error: DATARACEDETECTION must be performed in TRACE mode." << std::endl;
		exit(1);
	}
	if (mode&&tracingIsOn) { // this condition will be false
		std::string traceFileName = generateID();
	    SVcontrol = new sharedVariableController(traceFileName+"-replay.txt");
	}

} 


template<class T>
sharedVariable<T>::operator T() const {
	T t;
	startRead(*this);
	t = value;
	endRead(*this);

	if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		if (race) {
			std::cout << "Warning: data race detected in casting operator." << std::endl;
		}
	}

	return t;
}


template<class T>
std::string sharedVariable<T>::generateID() {
		const std::string thisClassName = "sharedVariable";
		std::string threadName;
		data d = threadData::instance().getThreadData(GetCurrentThreadId());	
		if (d.threadID != -1) 
		  threadName = d.threadName;
		else
		  threadName = "";
		size_t position = threadName.find("_"); 
		if (position > -1 && position < threadName.length())
		  threadName = threadName + "_" + thisClassName;
		else
		  threadName = "main_" + thisClassName;
			
		std::string myID = sharedVariableIDGenerator::instance().getID(threadName);
		return myID;
	}


template<class T>
bool sharedVariable<T>::computeLockSet() const{


  unsigned long currentThread = GetCurrentThreadId();
	// getLocked is synchronized 
  std::multiset<int> currentThreadLocked = lockData::instance().getLocked(currentThread);

  //std::cout << "currentThreadLocked of thread " << threadData::instance().getThreadData(currentThread).threadID << " is: ";
  //for (std::multiset<int>::iterator it = currentThreadLocked.begin();
  //   it != currentThreadLocked.end(); it++)
  //std::cout << *it << " ";
  //std::cout << std::endl;

  { // open block	
  mutexLocker<win32Critical_Section> mutex(dataLock);

  if (first) {
	// lockSet is empty until the first time this is called. So if first
	// use the locked set of currentThread as lockSet

    first = false;
	std::copy(currentThreadLocked.begin(),currentThreadLocked.end(),std::back_inserter(lockSet));
	//std::cout << "lockSet is: ";
    //for (std::list<int>::iterator it = lockSet.begin();
      //it != lockSet.end(); it++)
    //std::cout << *it << " ";
    //std::cout << std::endl;
  }
  else {

    //std::cout << "lockSet is: ";
    //for (std::list<int>::iterator it = lockSet.begin();
      //it != lockSet.end(); it++)
    //std::cout << *it << " ";
    //std::cout << std::endl;

  	// compute intersection of lockSet and currentThreads locked
    	std::list<int>::iterator first = lockSet.begin();
  	std::list<int>::iterator last = lockSet.end();
  	while (first != last) {
  	  std::list<int>::iterator next = first;
  	  next++;
			//search for lock *first in currentThreadLocked 
	  bool found = false;
	  for (std::multiset<int>::iterator it = currentThreadLocked.begin();
      it != currentThreadLocked.end(); it++)
		if (*it == *first) {
		  found = true;
	  	  break;
		}
	  if (!found)
	  	 lockSet.erase(first);
   	  //list<int>::iterator pos = currentThreadLocked.find(*first);
	    //if (pos == currentThreadLocked.end())
	    //  lockSet.erase(first);
	  first = next;
	} //end while
  } // end else
  if (lockSet.size() == 0 && writtenAtLeastOnce) // warning only if this is not a read-only var
	return true;
  else
	return false;
  } // close block
  }

template<class T>
bool sharedVariable<T>::raceDetected() const {
	return computeLockSet();
}

template<class T>
void sharedVariable<T>::displayLockset(std::string msg) const {
	std::cout << msg;
	for (std::list<int>::iterator it = lockSet.begin();
		it != lockSet.end(); it++)
  	std::cout << *it << " ";
	std::cout << std::endl;
}

template<class T>
void sharedVariable<T>::startRead(const sharedVariable<T>& sv) const {
	if (sv.tracingIsOn)
		if (mode == TRACE)
			sv.SVcontrol->traceStartRead();
		else if (mode == REPLAY)
			sv.SVcontrol->requestReadPermit();
}

template<class T>
void sharedVariable<T>::endRead(const sharedVariable<T>& sv) const {
	if (sv.tracingIsOn)
		if (mode == TRACE)
			sv.SVcontrol->traceEndRead();
		else if (mode == REPLAY)
			sv.SVcontrol->releaseReadPermit();
}

template<class T>
void sharedVariable<T>::startWrite(const sharedVariable<T>& sv) const {
	if (sv.tracingIsOn)
		if (mode == TRACE)
			sv.SVcontrol->traceStartWrite();
		else if (mode == REPLAY)
			sv.SVcontrol->requestWritePermit();
}

template<class T>
void sharedVariable<T>::endWrite(const sharedVariable<T>& sv) const {
	if (sv.tracingIsOn)
		if (mode == TRACE)
			sv.SVcontrol->traceEndWrite();
		else if (mode == REPLAY)
			sv.SVcontrol->releaseWritePermit();
}

template<class T>
const sharedVariable<T>& sharedVariable<T>::operator+=(const sharedVariable<T>& sv){

 T svTemp;
 T thisTemp;

 { 
   startRead(*this);
   thisTemp = value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on +=." << std::endl;
		}
   }
 } // close block


 {
   startRead(sv);
   svTemp = sv.value;
   endRead(sv);

   if (raceAnalysis && sv.analysisIsOn) {
		bool race = sv.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			//std::cout << "no locks in lockSet of rhs of +" << std::endl;
			std::cout << "Warning: data race detected on +=." << std::endl;
		}
   }
 }

   startWrite(*this);
	value = thisTemp + svTemp;
	writtenAtLeastOnce = true;
	endWrite(*this);

	// No race detection here since we already checked lhs

	return *this;
}


template<class T>
const sharedVariable<T>& sharedVariable<T>::operator-=(const sharedVariable<T>& sv) {

 T svTemp;
 T thisTemp;
 {
   startRead(*this);
   thisTemp = value;
   endRead(*this);
   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on -=." << std::endl;
		}
   } 
 }
 
 {
   startRead(sv);
   svTemp = sv.value;
   endRead(sv);

   if (raceAnalysis && sv.analysisIsOn) {
		bool race = sv.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on -=." << std::endl;
		}
   }
 }

 startWrite(*this);
 value = thisTemp - svTemp;
 writtenAtLeastOnce = true;
 endWrite(*this);
 // No race detection here since we already checked lhs

 return *this;
}

template<class T>
const sharedVariable<T>& sharedVariable<T>::operator*=(const sharedVariable<T>& sv){

 T svTemp;
 T thisTemp;

 {
   startRead(*this);
   thisTemp = value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on *=." << std::endl;
		}
   }
 } // close block


 {
   startRead(sv);
   svTemp = sv.value;
   endRead(sv);

   if (raceAnalysis && sv.analysisIsOn) {
		bool race = sv.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on *=." << std::endl;
		}
   }
 }

   startWrite(*this);
	value = thisTemp * svTemp;
	writtenAtLeastOnce = true;
	endWrite(*this);

	// No race detection here since we already checked lhs

	return *this;
}

template<class T>
const sharedVariable<T>& sharedVariable<T>::operator/=(const sharedVariable<T>& sv){

 T svTemp;
 T thisTemp;

 {
   startRead(*this);
   thisTemp = value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on /=." << std::endl;
		}
   }
 } // close block


 {
   startRead(sv);
   svTemp = sv.value;
   endRead(sv);

   if (raceAnalysis && sv.analysisIsOn) {
		bool race = sv.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on /=." << std::endl;
		}
   }
 }

   startWrite(*this);
	value = thisTemp / svTemp;
	writtenAtLeastOnce = true;
	endWrite(*this);

	// No race detection here since we already checked lhs

	return *this;
}

template<class T>
const sharedVariable<T>& sharedVariable<T>::operator%=(const sharedVariable<T>& sv){

 T svTemp;
 T thisTemp;

 {
   startRead(*this);
   thisTemp = value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on %=." << std::endl;
		}
   }
 } // close block


 {
   startRead(sv);
   svTemp = sv.value;
   endRead(sv);

   if (raceAnalysis && sv.analysisIsOn) {
		bool race = sv.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on %=." << std::endl;
		}
   }
 }

   startWrite(*this);
	value = thisTemp % svTemp;
	writtenAtLeastOnce = true;
	endWrite(*this);

	// No race detection here since we already checked lhs

	return *this;
}

template<class T>
const sharedVariable<T>& sharedVariable<T>::operator&=(const sharedVariable<T>& sv){

 T svTemp;
 T thisTemp;

 {
   startRead(*this);
   thisTemp = value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on &=." << std::endl;
		}
   }
 } // close block


 {
   startRead(sv);
   svTemp = sv.value;
   endRead(sv);

   if (raceAnalysis && sv.analysisIsOn) {
		bool race = sv.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on &=." << std::endl;
		}
   }
 }

   startWrite(*this);
	value = thisTemp & svTemp;
	writtenAtLeastOnce = true;
	endWrite(*this);
    

	// No race detection here since we already checked lhs

	return *this;
}

template<class T>
const sharedVariable<T>& sharedVariable<T>::operator|=(const sharedVariable<T>& sv){
 //sharedVariable result;

 T svTemp;
 T thisTemp;

 {
   startRead(*this);
   thisTemp = value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on |=." << std::endl;
		}
   }
 } // close block


 {
   startRead(sv);
   svTemp = sv.value;
   endRead(sv);

   if (raceAnalysis && sv.analysisIsOn) {
		bool race = sv.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on |=." << std::endl;
		}
   }
 }

   startWrite(*this);
	value = thisTemp | svTemp;
	writtenAtLeastOnce = true;
	endWrite(*this);
    

	// No race detection here since we already checked lhs

	return *this;
}

template<class T>
const sharedVariable<T>& sharedVariable<T>::operator^=(const sharedVariable<T>& sv){

 T svTemp;
 T thisTemp;

 {
   startRead(*this);
   thisTemp = value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on ^=." << std::endl;
		}
   }
 } // close block


 {
   startRead(sv);
   svTemp = sv.value;
   endRead(sv);

   if (raceAnalysis && sv.analysisIsOn) {
		bool race = sv.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on ^=." << std::endl;
		}
   }
 }

   startWrite(*this);
	value = thisTemp ^ svTemp;
	writtenAtLeastOnce = true;
	endWrite(*this);
    

	// No race detection here since we already checked lhs

	return *this;
}

template<class T>
const sharedVariable<T>& sharedVariable<T>::operator>>=(const sharedVariable<T>& sv){

 T svTemp;
 T thisTemp;

 {
   startRead(*this);
   thisTemp = value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on >>=." << std::endl;
		}
   }
 } // close block


 {
   startRead(sv);
   svTemp = sv.value;
   endRead(sv);

   if (raceAnalysis && sv.analysisIsOn) {
		bool race = sv.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on >>=." << std::endl;
		}
   }
 }

   startWrite(*this);
	value = thisTemp >> svTemp;
	writtenAtLeastOnce = true;
	endWrite(*this);
    

	// No race detection here since we already checked lhs

	return *this;
}

template<class T>
const sharedVariable<T>& sharedVariable<T>::operator<<=(const sharedVariable<T>& sv){

 T svTemp;
 T thisTemp;

 {
   startRead(*this);
   thisTemp = value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on <<=." << std::endl;
		}
   }
 } // close block


 {
   startRead(sv);
   svTemp = sv.value;
   endRead(sv);

   if (raceAnalysis && sv.analysisIsOn) {
		bool race = sv.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on <<=." << std::endl;
		}
   }
 }

   startWrite(*this);
	value = thisTemp << svTemp;
	writtenAtLeastOnce = true;
	endWrite(*this);
    

	// No race detection here since we already checked lhs

	return *this;
}


template <class TT>
bool operator!=(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv1Temp;
 TT sv2Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on !=." << std::endl;
		}
   }
 } // close block


 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on !=." << std::endl;
		}
   }
 }

 result = (sv1Temp != sv2Temp);
 return result;
}

template <class TT>
bool operator!=(const sharedVariable<TT>& sv1, const TT& sv2) {
 bool result;
 TT sv1Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on !=." << std::endl;
		}
   }
 } // close block


 result = (sv1Temp != sv2);
 return result;
}

template <class TT>
bool operator!=(const TT& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv2Temp;

 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on !=." << std::endl;
		}
   }
 }

 result = (sv1 != sv2Temp);
 return result;
}

template <class TT>
bool operator<(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv1Temp;
 TT sv2Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on <." << std::endl;
		}
   }
 } // close block


 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on <." << std::endl;
		}
   }
 }

 result = (sv1Temp < sv2Temp);
 return result;
}

template <class TT>
bool operator<(const sharedVariable<TT>& sv1, const TT& sv2) {
 bool result;
 TT sv1Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on <." << std::endl;
		}
   }
 } // close block


 result = (sv1Temp < sv2);
 return result;
}

template <class TT>
bool operator<(const TT& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv2Temp;

 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on <." << std::endl;
		}
   }
 }

 result = (sv1 < sv2Temp);
 return result;
}

template <class TT>
bool operator>(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv1Temp;
 TT sv2Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on >." << std::endl;
		}
   }
 } // close block


 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on >." << std::endl;
		}
   }
 }

 result = (sv1Temp > sv2Temp);
 return result;
}

template <class TT>
bool operator>(const sharedVariable<TT>& sv1, const TT& sv2) {
 bool result;
 TT sv1Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on >." << std::endl;
		}
   }
 } // close block


 result = (sv1Temp > sv2);
 return result;
}

template <class TT>
bool operator>(const TT& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv2Temp;

 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on >." << std::endl;
		}
   }
 }

 result = (sv1 > sv2Temp);
 return result;
}

template <class TT>
bool operator<=(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv1Temp;
 TT sv2Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on <=." << std::endl;
		}
   }
 } // close block


 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on <=." << std::endl;
		}
   }
 }

 result = (sv1Temp <= sv2Temp);
 return result;
}

template <class TT>
bool operator<=(const sharedVariable<TT>& sv1, const TT& sv2) {
 bool result;
 TT sv1Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on <=." << std::endl;
		}
   }
 } // close block


 result = (sv1Temp <= sv2);
 return result;
}

template <class TT>
bool operator<=(const TT& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv2Temp;

 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on <=." << std::endl;
		}
   }
 }

 result = (sv1 <= sv2Temp);
 return result;
}

template <class TT>
bool operator>=(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv1Temp;
 TT sv2Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on >=." << std::endl;
		}
   }
 } // close block


 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on >=." << std::endl;
		}
   }
 }

 result = (sv1Temp >= sv2Temp);
 return result;
}

template <class TT>
bool operator>=(const sharedVariable<TT>& sv1, const TT& sv2) {
 bool result;
 TT sv1Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on >=." << std::endl;
		}
   }
 } // close block


 result = (sv1Temp >= sv2);
 return result;
}


template <class TT>
bool operator>=(const TT& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv2Temp;

 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {;
			std::cout << "Warning: data race detected on >=." << std::endl;
		}
   }
 }

 result = (sv1 >= sv2Temp);
 return result;
}


template <class TT>
bool operator&&(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv1Temp;
 TT sv2Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on &&." << std::endl;
		}
   }
 } // close block


 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on &&." << std::endl;
		}
   }
 }

 result = (sv1Temp && sv2Temp);
 return result;
}

template <class TT>
bool operator&&(const sharedVariable<TT>& sv1, const TT& sv2) {
 bool result;
 TT sv1Temp;
 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on &&." << std::endl;
		}
   }
 } // close block


 result = (sv1Temp && sv2);
 return result;
}

template <class TT>
bool operator&&(const TT& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv2Temp;

 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {;
			std::cout << "Warning: data race detected on &&." << std::endl;
		}
   }
 }

 result = (sv1 && sv2Temp);
 return result;
}

template <class TT>
bool operator&&(const sharedVariable<TT>& sv1, const bool& sv2) {
 bool result;
 TT sv1Temp;
 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on &&." << std::endl;
		}
   }
 } // close block


 result = (sv1Temp && sv2);
 return result;
}

template <class TT>
bool operator||(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv1Temp;
 TT sv2Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on ||." << std::endl;
		}
   }
 } // close block


 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on ||." << std::endl;
		}
   }
 }

 result = (sv1Temp || sv2Temp);
 return result;
}

template <class TT>
bool operator||(const sharedVariable<TT>& sv1, const TT& sv2) {
 bool result;
 TT sv1Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on ||." << std::endl;
		}
   }
 } // close block


 result = (sv1Temp || sv2);
 return result;
}

template <class TT>
bool operator||(const TT& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv2Temp;

 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on ||." << std::endl;
		}
   }
 }

 result = (sv1 || sv2Temp);
 return result;
}

template <class TT>
bool operator||(const sharedVariable<TT>& sv1, const bool& sv2) {
 bool result;
 TT sv1Temp;
 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on ||." << std::endl;
		}
   }
 } // close block


 result = (sv1Temp || sv2);
 return result;
}


template<class T>
const sharedVariable<T>& sharedVariable<T>::operator=(const sharedVariable<T>& sv) {
//void sharedVariable<T>::operator=(const sharedVariable<T>& sv) {
  T svTemp;
	{
	startRead(sv);
   svTemp = sv.value;
	endRead(sv);

	if (raceAnalysis && sv.analysisIsOn) {
		bool race = sv.raceDetected();
		//sv.displayLockset("lockSet of + rhs: ");
		if (race) {
			std::cout << "Warning: data race detected on operator =." << std::endl;
		}
	}
	}
	{
	startWrite(*this);
   value = svTemp;
	writtenAtLeastOnce = true;
	endWrite(*this);

	if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on operator =." << std::endl;
		}
	}
	return *this;
	}
}

template<class T>
const sharedVariable<T>& sharedVariable<T>::operator=(const T& v) {
//void sharedVariable<T>::operator=(const T& v) {
	startWrite(*this);
   value = v;
	writtenAtLeastOnce = true;
	endWrite(*this);

	if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on operator =." << std::endl;
		}
	}
	return *this;
}

template<class T>
const sharedVariable<T>& sharedVariable<T>::operator++() {
 T thisTemp;

 {

   startRead(*this);
   thisTemp = value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on ++." << std::endl;
		}
   }
 } // close block

   startWrite(*this);
	value = thisTemp + 1;
	writtenAtLeastOnce = true;
	endWrite(*this);
    

	// No race detection here since we already checked lhs

	return *this;
}

template<class T>
const sharedVariable<T> sharedVariable<T>::operator++(int) {
 T thisTemp;
 { 

   startRead(*this);
   thisTemp = value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on ++." << std::endl;
		}
   }
 } // close block

   startWrite(*this);
	value = thisTemp + 1;
	writtenAtLeastOnce = true;
	endWrite(*this);
    

	// No race detection here since we already checked lhs

	return sharedVariable<T>(thisTemp,analysisOff,traceOff);

}

template<class T>
const sharedVariable<T>& sharedVariable<T>::operator--() {
 T thisTemp;

 {

   startRead(*this);
   thisTemp = value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on --." << std::endl;
		}
   }
 } // close block

   startWrite(*this);
	value = thisTemp - 1;
	writtenAtLeastOnce = true;
	endWrite(*this);
    
	// No race detection here since we already checked lhs

	return *this;
}

template<class T>
const sharedVariable<T> sharedVariable<T>::operator--(int) {
 T thisTemp;
 {
   startRead(*this);
   thisTemp = value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on --." << std::endl;
		}
   }
 } // close block

   startWrite(*this);
	value = thisTemp - 1;
	writtenAtLeastOnce = true;
	endWrite(*this);
    

	// No race detection here since we already checked lhs

	return sharedVariable<T>(thisTemp,analysisOff,traceOff);
}


template<class T>
const sharedVariable<T> sharedVariable<T>::operator+() {
 T thisTemp;
 {
   startRead(*this);
   thisTemp = +value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on unary +." << std::endl;
		}
   }
 } // close block
    

	// No race detection here since we already checked lhs

	return sharedVariable<T>(thisTemp,analysisOff,traceOff);
}

template<class T>
const sharedVariable<T> sharedVariable<T>::operator-() {
 T thisTemp;
 {
   startRead(*this);
   thisTemp = -value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on unary -." << std::endl;
		}
   }
 } // close block
    

	// No race detection here since we already checked lhs

	return sharedVariable<T>(thisTemp,analysisOff,traceOff);
}

template<class T>
const sharedVariable<T> sharedVariable<T>::operator~() {
 T thisTemp;
 {
   startRead(*this);
   thisTemp = ~value;
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on ~." << std::endl;
		}
   }
 } // close block
    

	// No race detection here since we already checked lhs

	return sharedVariable<T>(thisTemp,analysisOff,traceOff);
}

template<class T>
bool sharedVariable<T>::operator!() {
 bool thisTemp;
 {
   startRead(*this);
   thisTemp = !(value);
   endRead(*this);

   if (raceAnalysis && analysisIsOn) {
		bool race = raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on !." << std::endl;
		}
   }
 } // close block
    

	// No race detection here since we already checked lhs

	return thisTemp;
}

/* access not allowed */
template<class T>
const sharedVariable<T>& sharedVariable<T>::operator&() {
	assert(false);
	return *this;
}

/* access not allowed */
template<class T>
const sharedVariable<T>& sharedVariable<T>::operator*() {
	assert(false);
	return *this;
}


template<class T>
void sharedVariable<T>::setAnalysis(analysisOnOff on_) {
	if (on_ == analysisOn) 
		analysisIsOn = true;
	else
		analysisIsOn = false;
}
template<class T>
bool sharedVariable<T>::getAnalysisOn() {return analysisIsOn;}

template<class T>
void sharedVariable<T>::setTrace(tracingOnOff on_) {
	if (on_ == traceOn) 
		tracingIsOn = true;
	else
		tracingIsOn = false;
}
template<class T>
bool sharedVariable<T>::getTraceOn() {return tracingIsOn;}


template<class TT>
std::ostream& operator<<(std::ostream& out, const sharedVariable<TT>& sv) {
		TT svTemp;
		{
		sv.startRead(sv);
		svTemp = sv.value;
		sv.endRead(sv);

		if (sv.raceAnalysis && sv.analysisIsOn) {
			bool race = sv.raceDetected();
			//sv.displayLockset("lockSet of + rhs: ");
			if (race) {
				std::cout << "Warning: data race detected on <<." << std::endl;
			}
		}

 
		}
		out << svTemp;
		return out;
}

template <class TT>
bool operator==(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv1Temp;
 TT sv2Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on ==." << std::endl;
		}
   }
 } // close block


 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on ==." << std::endl;
		}
   }
 }

 result = (sv1Temp == sv2Temp);
 return result;
}

template<class TT>
bool operator==(const sharedVariable<TT>& sv1, const TT& sv2) {
 bool result;
 TT sv1Temp;

 {
   sv1.startRead(sv1);
   sv1Temp = sv1.value;
   sv1.endRead(sv1);

   if (sv1.raceAnalysis && sv1.analysisIsOn) {
		bool race = sv1.raceDetected();
		//displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on ==." << std::endl;
		}
   }
 } // close block


 result = (sv1Temp == sv2);
 return result;
}

template <class TT>
bool operator==(const TT& sv1, const sharedVariable<TT>& sv2) {
 bool result;
 TT sv2Temp;

 {
   sv2.startRead(sv2);
   sv2Temp = sv2.value;
   sv2.endRead(sv2);

   if (sv2.raceAnalysis && sv2.analysisIsOn) {
		bool race = sv2.raceDetected();
		//sv.displayLockset("lockSet of + lhs: ");
		if (race) {
			std::cout << "Warning: data race detected on ==." << std::endl;
		}
   }
 }

 result = (sv1 == sv2Temp);
 return result;
}

template<class TT>
sharedVariable<TT> operator+(const sharedVariable<TT>& sv1, const TT& sv2) {
	TT temp = sv1;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp+sv2;
	return s;
}


template<class TT>
sharedVariable<TT> operator+(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
	TT temp1 = sv1;
   TT temp2 = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp1+temp2;
	return s;
}

template<class TT>
sharedVariable<TT> operator+(const TT& sv1, const sharedVariable<TT>& sv2) {
	TT temp = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = sv1+temp;
	return s;
}

template<class TT>
sharedVariable<TT> operator-(const sharedVariable<TT>& sv1, const TT& sv2) {
	TT temp = sv1;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp-sv2;
	return s;
}


template<class TT>
sharedVariable<TT> operator-(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
	TT temp1 = sv1;
   TT temp2 = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp1-temp2;
	return s;
}

template<class TT>
sharedVariable<TT> operator-(const TT& sv1, const sharedVariable<TT>& sv2) {
	TT temp = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = sv1-temp;
	return s;
}

template<class TT>
sharedVariable<TT> operator*(const sharedVariable<TT>& sv1, const TT& sv2) {
	TT temp = sv1;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp*sv2;
	return s;
}


template<class TT>
sharedVariable<TT> operator*(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
	TT temp1 = sv1;
    TT temp2 = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp1*temp2;
	return s;
}

template<class TT>
sharedVariable<TT> operator*(const TT& sv1, const sharedVariable<TT>& sv2) {
	TT temp = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = sv1*temp;
	return s;
}

template<class TT>
sharedVariable<TT> operator/(const sharedVariable<TT>& sv1, const TT& sv2) {
	TT temp = sv1;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp/sv2;
	return s;
}


template<class TT>
sharedVariable<TT> operator/(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
	TT temp1 = sv1;
   TT temp2 = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp1/temp2;
	return s;
}

template<class TT>
sharedVariable<TT> operator/(const TT& sv1, const sharedVariable<TT>& sv2) {
	TT temp = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = sv1/temp;
	return s;
}

template<class TT>
sharedVariable<TT> operator%(const sharedVariable<TT>& sv1, const TT& sv2) {
	TT temp = sv1;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp%sv2;
	return s;
}


template<class TT>
sharedVariable<TT> operator%(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
	TT temp1 = sv1;
   TT temp2 = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp1%temp2;
	return s;
}

template<class TT>
sharedVariable<TT> operator%(const TT& sv1, const sharedVariable<TT>& sv2) {
	TT temp = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = sv1%temp;
	return s;
}


template<class TT>
sharedVariable<TT> operator&(const sharedVariable<TT>& sv1, const TT& sv2) {
	TT temp = sv1;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp&sv2;
	return s;
}

template<class TT>
sharedVariable<TT> operator&(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
	TT temp1 = sv1;
	TT temp2 = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp1&temp2;
	return s;
}

template<class TT>
sharedVariable<TT> operator&(const TT& sv1, const sharedVariable<TT>& sv2) {
	TT temp = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = sv1&temp;
	return s;
}


template<class TT>
sharedVariable<TT> operator|(const sharedVariable<TT>& sv1, const TT& sv2) {
	TT temp = sv1;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp|sv2;
	return s;
}

template<class TT>
sharedVariable<TT> operator|(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
	TT temp1 = sv1;
	TT temp2 = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp1|temp2;
	return s;
}

template<class TT>
sharedVariable<TT> operator|(const TT& sv1, const sharedVariable<TT>& sv2) {
	TT temp = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = sv1|temp;
	return s;
}


template<class TT>
sharedVariable<TT> operator^(const sharedVariable<TT>& sv1, const TT& sv2) {
	TT temp = sv1;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp^sv2;
	return s;
}

template<class TT>
sharedVariable<TT> operator^(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
	TT temp1 = sv1;
	TT temp2 = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp1^temp2;
	return s;
}

template<class TT>
sharedVariable<TT> operator^(const TT& sv1, const sharedVariable<TT>& sv2) {
	TT temp = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = sv1^temp;
	return s;
}

template<class TT>
sharedVariable<TT> operator<<(const sharedVariable<TT>& sv1, const TT& sv2) {
	TT temp = sv1;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp<<sv2;
	return s;
}

template<class TT>
sharedVariable<TT> operator<<(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
	TT temp1 = sv1;
	TT temp2 = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp1<<temp2;
	return s;
}

template<class TT>
sharedVariable<TT> operator<<(const TT& sv1, const sharedVariable<TT>& sv2) {
	TT temp = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = sv1<<temp;
	return s;
}

template<class TT>
sharedVariable<TT> operator>>(const sharedVariable<TT>& sv1, const TT& sv2) {
	TT temp = sv1;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp>>sv2;
	return s;
}

template<class TT>
sharedVariable<TT> operator>>(const sharedVariable<TT>& sv1, const sharedVariable<TT>& sv2) {
	TT temp1 = sv1;
	TT temp2 = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = temp1>>temp2;
	return s;
}

template<class TT>
sharedVariable<TT> operator>>(const TT& sv1, const sharedVariable<TT>& sv2) {
	TT temp = sv2;
	sharedVariable<TT> s(sharedVariable<TT>::analysisOff,sharedVariable<TT>::traceOff);
	s = sv1>>temp;
	return s;
}
