public final class ReadersWritersSemaphores {
   public static void main (String args[]) {
   	RWController c = new RWController(); 
		Reader r1 = new Reader(c,1);	Reader r2 = new Reader(c,2); Reader r3 = new Reader(c,3);
		Writer w1 = new Writer(c,1);	Writer w2 = new Writer(c,2);  
		r1.start(); w1.start(); r2.start(); w2.start(); r3.start();
		try {
	    r1.join(); r2.join(); r3.join();
	    w1.join(); w2.join();
		} 
		catch (InterruptedException e) {}
	}
}

final class Reader extends Thread {
	private RWController c;
	private int num;
	Reader (RWController c, int num) {this.c = c; this.num = num; }
	public void run () {
		System.out.println("Reader"+num+" running");
		c.read(num);
		System.out.println("Reader"+num+" finished");
	}
}

final class Writer extends Thread {
	private RWController c;
	private int num;
	Writer (RWController c, int num) {this.c = c; this.num = num; }
	public void run () {		
		System.out.println("Writer"+num+" running");
		c.write(num);
		System.out.println("Writer"+num+" finished");
	}
}

final class RWController{
// R<W.2. This strategy allows concurrent reading and generally gives writers a 
// higher priority than readers. Readers have priority in the following situation: 
// when a writer requests to write, if it is a lead writer (i.e., no other writer 
// is writing or waiting), it waits until all readers that arrived earlier 
// have finished reading.

	int activeReaders = 0;					// number of active readers
	int waitingOrWritingWriters = 0;		// number of writers waiting or writing
	mutexLock mutex_r;						// exclusion for activeReaders 
	// exclusion for waitingOrWritingWriters
	mutexLock mutex_w;	
	binarySemaphore writers_que;			// waiting writers
	// condition queue for waiting readers and the first waiting writer
	binarySemaphore readers_w_que;	
	int x=0;       								// x is shared data

	public RWController() {
		mutex_r = new mutexLock();
		mutex_w = new mutexLock();	
		writers_que = new binarySemaphore(1);
		readers_w_que = new binarySemaphore(1);
	}

	public void read(int num) {
	
		readers_w_que.P();	// serve waiting readers and lead writer FCFS
		mutex_r.lock();
		++activeReaders;
		if (activeReaders == 1) 
			writers_que.P();	// block writers reads are occurring
		mutex_r.unlock();
		readers_w_que.V();	// signal the next waiting reader or a lead writer

		//System.out.println("Reader" + num + " reading");
		//try{Thread.sleep(10);}catch(InterruptedException e){}; 

		mutex_r.lock();
		--activeReaders;
		if (activeReaders == 0) 
			writers_que.V();	// allow writing
		mutex_r.unlock();		
	}				

	public void write(int num) {
		// if a lead writer is waiting in readers_w_que, this blocks other writers
		mutex_w.lock();	
		++waitingOrWritingWriters;
		if (waitingOrWritingWriters == 1)	// true if this is a lead writer
			readers_w_que.P();	// block lead writer if there are waiting readers
		mutex_w.unlock();
		writers_que.P();	// block if a writer is writing or a reader is reading

		//System.out.println("Writer" + num + " writing");
		//try{Thread.sleep(10);}catch(InterruptedException e){};

		writers_que.V();	// signal writing is over; wakes up waiting writer (if any)
		mutex_w.lock();
		-- waitingOrWritingWriters;
		if (waitingOrWritingWriters == 0)		// no writers are waiting or writing	
			readers_w_que.V();				// so allow reading
		mutex_w.unlock();
	}
}


