#include <windows.h>
#include <process.h>

HANDLE hMutex = NULL;								// Global Mutex

void PrintError(LPTSTR lpszFunction,LPSTR fileName, int lineNumber) { 
	TCHAR szBuf[256]; LPSTR lpErrorBuf;
	DWORD errorCode = GetLastError(); 
	FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | 
	FORMAT_MESSAGE_FROM_SYSTEM, NULL, errorCode, 
	MAKELANGID(LANG_NEUTRAL, 
	SUBLANG_DEFAULT), (LPTSTR) &lpErrorBuf, 0, NULL );
	wsprintf(szBuf, "%s failed at line %d in %s with error %d: %s", lpszFunction, 
	lineNumber, fileName, errorCode, lpErrorBuf); 
 	DWORD numWritten;
	WriteFile(GetStdHandle(STD_ERROR_HANDLE),szBuf,strlen(szBuf),
	&numWritten,FALSE);
	LocalFree(lpErrorBuf);
	exit(errorCode); 
}

unsigned WINAPI Thread1(LPVOID lpvThreadParm) {
	// Request ownership of mutex. 
	DWORD rc =::WaitForSingleObject(
	hMutex, 	// handle to the Mutex
	INFINITE); 	// wait forever (no timeout)
	switch(rc) {
	case 	WAIT_OBJECT_0: 		// wait completed successfully
	break; 	// received ownership
	case 	WAIT_FAILED:	// wait failed
	// received ownership but the program's state is unknown 
	case 	WAIT_ABANDONED:	
	case 	WAIT_TIMEOUT: // timeouts impossible since INFINITE used		
	PrintError("WaitForSingleObject failed at ",__FILE__,__LINE__); 		// see Listing 1.3 for PrintError().
	break;
	}

	// Release ownership of mutex
	rc = ::ReleaseMutex(hMutex);
	if (!rc) PrintError("ReleaseMutex failed at ",__FILE__,__LINE__);

	return 0;
}

unsigned WINAPI Thread2(LPVOID lpvThreadParm) { 
	// Request ownership of mutex. 
	DWORD rc =::WaitForSingleObject(
	hMutex, 	// handle to the Mutex
	INFINITE); 	// wait forever (no timeout)
	switch(rc) {
	case 	WAIT_OBJECT_0: 		// wait completed successfully
	break; 	// received ownership
	case 	WAIT_FAILED:	// wait failed
	// received ownership but the program's state is unknown 
	case 	WAIT_ABANDONED:	
	case 	WAIT_TIMEOUT: // timeouts impossible since INFINITE used		
	PrintError("WaitForSingleObject failed at ",__FILE__,__LINE__); 		// see Listing 1.3 for PrintError().
	break;
	}

	// Release ownership of mutex
	rc = ::ReleaseMutex(hMutex);
	if (!rc) PrintError("ReleaseMutex failed at ",__FILE__,__LINE__);

	return 0;
}

int main() {
HANDLE threadArray[2];
	unsigned threadID; 
	hMutex = CreateMutex(
	NULL,	// no security attributes
	FALSE,	// this mutex is not initially owned by the creating thread
	NULL);	// unnamed mutex that will not be shared across processes
	threadArray[0]= (HANDLE)_beginthreadex(NULL,0, Thread1, NULL,0, 
	&threadID );
	threadArray[1]= (HANDLE)_beginthreadex(NULL,0, Thread2, NULL,0,
	&threadID );
	WaitForMultipleObjects(2,threadArray,TRUE,INFINITE);
	CloseHandle(threadArray[0]);	// release references when finished with them
	CloseHandle(threadArray[1]);
	CloseHandle(hMutex);
	return 0;
}


//Listing 3.21 Using Win32 Mutex objects.
