#include "countingSemaphore.h"
#include <iostream>

countingSemaphore::countingSemaphore(int initialPermits)  : semaphore(initialPermits) {}

void countingSemaphore::P() {
	lock();	// lock is inherited from class semaphore
	--permits;
	if (permits>=0) {unlock(); return; }
	// each thread blocks on its own semaphore
	win32Semaphore* s =  pool.acquire();	
	waitingP.push(s); 	// append the semaphore
	unlock();
	s->P();	// block on the semaphore
	pool.release(s);
} 

void countingSemaphore::V() {
// each thread blocks on its own semaphore
	lock(); // lock semaphore
	++permits;
	if (permits>0) { unlock(); return; }
	win32Semaphore* oldest = waitingP.front();
	waitingP.pop();
	oldest->V();
	unlock(); // end synchronized(this) to avoid doing s.P() while 
}	// holding the lock on this semaphore

bool countingSemaphore::doP() {
// Called by VP() operation; checks permits and returns true if P() should block; 
// false otherwise.
	--permits;
	if (permits>=0) 
	return false;	// P() does not block
	else
	return true;	// P() does block
}

void countingSemaphore::VP(semaphore* vSem) {
// Execute {vSem->V(); this->P();} without any intervening  P() or V() operations 
// on this or vSem.	 Return 0 if this operation fails.
// Lock the semaphores in ascending order of IDs to prevent circular deadlock (i.e. 
// T1 holds the lock of this and waits for vSem's lock while T2 holds vSem's lock 
// and waits for the lock of this.)
	semaphore* first = this;
	semaphore* second = vSem;
	if (this->getSemaphoreID() > vSem->getSemaphoreID()) { 
	first = vSem;
	second = this;
	}
	first->lock(); 
	second->lock();
	//vSem.V() must not block
	if (vSem->permits==1 && vSem->isBinarySemaphore()) { 
	// method isBinarySemaphore() is inherited from class semaphore
	std::cout << "VP failed at line " << __LINE__ << " in " << __FILE__ 
	<< " with error: V operation will block." << std::endl;
	exit(1);
	}
	// perform vSem.V()
	vSem->V(); //  okay to already hold vSem's lock (which is first or second)
	//       since it is a recursive lock
	// perform this->P()
	bool blockingP = doP();
	if (!blockingP) {
	second->unlock();
	first->unlock();
	}
	// each thread blocks on own semaphore
	win32Semaphore* s =  pool.acquire();
	waitingP.push(s); 	// otherwise append blocked thread
	second->unlock();	// unlock semaphores before blocking
	first->unlock();
	s->P();	// s is already in waitingP so FCFS is enforced
	pool.release(s);
} 


//Listing 3.24 C++/Win32 class countingSemaphore.

