#include "countingSemaphore.h"
#include "mutexLock.h"
#include <memory>
#include "thread.h"
#include <iostream>

const int capacity = 3;
class Buffer {
private:
	int buffer[capacity];
	int count, in, out;
public:
	Buffer() : in(0), out(0), count(0) { }
	int size() { return count;}
	int withdraw () {
	int value = 0;
	value = buffer[out];		// out is shared by consumers
	out = (out + 1) % capacity;
	count--;
	return value;
	}
	void deposit (int value) {
	buffer[in] = value;		// in is shared by producers
	in = (in + 1) % capacity;
	count++;
	}
};
Buffer sharedBuffer;		// 3-slot buffer
mutexLock mutexD, mutexW;
countingSemaphore emptySlots(capacity);
countingSemaphore fullSlots(0);
class Producer : public Thread {
public:
	virtual void* run () {
	int i;
	std::cout << "producer running" << std::endl;
	for (i=0; i<2; i++) {
	emptySlots.P();
	mutexD.lock();
	sharedBuffer.deposit(i);
	std::cout << "Produced: " << i << std::endl;
	mutexD.unlock();
	fullSlots.V();
	}
	return 0;
	}
};
class Consumer : public Thread {
public:  
	virtual void* run () {
 	int result;
	std::cout << "consumer running" << std::endl;
	for (int i=0; i<2; i++) {
	fullSlots.P();
	mutexW.lock();
	result = sharedBuffer.withdraw();
	mutexW.unlock();
	std::cout << "Consumed: " << result << std::endl;
	emptySlots.V();
	}
	return 0;
	}
};


int main() {
	std::auto_ptr<Producer> p1(new Producer);
	std::auto_ptr<Producer> p2(new Producer);
	std::auto_ptr<Consumer> c1(new Consumer);
	std::auto_ptr<Consumer> c2(new Consumer);
	c1->start();p1->start(); p2->start();c2->start();
	p1->join(); p2->join(); c1->join(); c2->join();
	return(0);
}

//Listing 3.26 Win32 bounded buffer using countingSemaphores and mutexLocks.
