#include "semaphore.h"
#include <queue>
#include "countingSemaphore.h"
#include "win32Semaphore.h"
#include <typeinfo>
#include <string.h>
#include <assert.h>
#include <iostream>

countingSemaphore::countingSemaphore(int initialPermits)
 : semaphore(initialPermits) {}

countingSemaphore::~countingSemaphore() {}		

void countingSemaphore::P() {
	// each thread blocks on its own conditionVar
	lock();
	--permits;
	if (permits>=0) {
	   unlock();
	   return;
	}
    waitRecord rec;
    rec.conditionVar = pool.acquire();
    waitingP.push(rec); // otherwise append blocked thread
	unlock();
	rec.conditionVar->P();
	// never block during replay
	pool.release(rec.conditionVar);
} // end P()

void countingSemaphore::V() {
	lock(); // lock semaphore
	++permits;
	if (permits>0) {
		unlock();
		return;
	}
    win32Semaphore* oldestConditionVar = (waitingP.front()).conditionVar;
	waitingP.pop();
	// never block during replay
	oldestConditionVar->V();
    unlock(); // end synchronized (this) to avoid conditionVar.wait() while holding lock on this
} // end V()


bool countingSemaphore::doP() {
// Called by VP() operation; checks permit and unblocks a thread waiting
// in V() if necessary. Returns true if P() should block; false otherwise.
   --permits;
	if (permits>=0)
		return false;	// P does not block
	else
		return true;	// P() does block
}

void countingSemaphore::VP(semaphore* vSem) {
// execute {vSem.V(); this.P();} without any intervening 
// P() or V() operations on this or vSem.	

// lock semaphores in ascending order of IDs to prevent circular deadlock (i.e. T1 holds
// this's lock and waits for vSem's lock while T2 holds vSem's lock and waits for this's lock.)
   semaphore* first = this;
   semaphore* second = vSem;
   if (this->getSemaphoreID() > vSem->getSemaphoreID()) { 
     first = vSem;
     second = this;
   }
   first->lock(); 
   second->lock();
   //vSem.V() must not block
   //if (!strcmp(typeid(*vSem).name(),"binarySemaphore") && vSem->permits==1)
	if (vSem->permits==1 && vSem->isBinarySemaphore())
	   std::cout << "VP failed at line " << __LINE__ << " in " << __FILE__ 
				  << " with error: V operation will block." << std::endl;
	   exit(1);
   // perform vSem.V()
   vSem->V(); // it's okay that we already hold vSem's lock
   // perform this.P()
   bool blockingP = doP();
   if (!blockingP) {
     second->unlock();
     first->unlock();
     return;
   }
   waitRecord rec;
   rec.conditionVar = pool.acquire();
   waitingP.push(rec); // otherwise append blocked thread
   second->unlock();
   first->unlock();
   rec.conditionVar->P();
   pool.release(rec.conditionVar);
} // end VP()
