// this mutexLock suppoted race analysis but not tracing/replay
#include "mutexlock.h"
#include <windows.h>
#include <iostream>
#include <assert.h>


mutexLock::mutexLock(): owner(0), count(0), mutex(1) {}


mutexLock::~mutexLock() {}

void mutexLock::lock() {
	mutex.P();
	if (count == 0) {
		count = 1;
		owner = GetCurrentThreadId();
	}
	else if (owner == GetCurrentThreadId()) {
		++count;
	}
	else  {
		win32Semaphore* s = pool.acquire();	// a pool of semaphores
		waitingLock.push(s);	// otherwise append blocked thread
		mutex.V();
		s->P();					// returns holding mutex since unlock() didn't release it
		count = 1;			
		owner = GetCurrentThreadId();
		pool.release(s);
	}
	mutex.V();
}

void mutexLock::unlock() {
	mutex.P();
	if (owner != 0) {
		if (owner == GetCurrentThreadId()) {
				--count;
				if (count == 0) {
					owner = 0;
					if (!waitingLock.empty()) { // any threads waiting for lock?
						win32Semaphore* oldest = waitingLock.front();
						waitingLock.pop();
						oldest->V();
						// do not release mutex so oldest will complete P() next
						return;
					}
					else {
						mutex.V(); return;
					}
				}
				else {
					mutex.V(); return;
				}
		}
		else {
			std::cout << "unlock failed at line " << __LINE__ << " in " << __FILE__ 
					  << " with error: Calling thread is not the owner." << std::endl;
			exit(1);
		}

	}
	else {
		std::cout << "unlock failed at line " << __LINE__ << " in " << __FILE__ 
				  << " with error: Calling thread is not the owner." << std::endl;
		exit(1);
	}
}



