#include <pthread.h>
#include <iostream>
#include <errno.h>

pthread_mutex_t mutex;

void* Thread1(void* arg) {
	pthread_mutex_lock(&mutex); 
	std::cout << "thread 1 in critical section" << std::endl;
	pthread_mutex_unlock(&mutex);
	return NULL;
}

void* Thread2(void* arg) {
	pthread_mutex_lock(&mutex); 
	std::cout << "thread 2 in critical section" << std::endl;
	pthread_mutex_unlock(&mutex);
	return NULL;
}

int main() {
	pthread_t threadArray[2];			// array of thread IDs
	int status;
	pthread_attr_t threadAttribute;

	pthread_mutexattr_t mutexAttribute;
	status = pthread_mutexattr_init (&mutexAttribute);

	if (status !=0) { 
		std::cout << "pthread_mutexattr_init failed at " << __FILE__ << ":" << __LINE__ << " - " << strerror(status) << 
      std::endl; exit(status);
	}
	status = pthread_mutexattr_settype(&mutexAttribute,PTHREAD_MUTEX_RECURSIVE);
	if (status !=0) { 
		std::cout << "pthread_mutexattr_settype failed at " << __FILE__ << ":" << __LINE__ << " - " << strerror(status) << 
      std::endl; exit(status);
	}
	status = pthread_mutex_init(&mutex,&mutexAttribute);
	if (status !=0) { 
		std::cout << "pthread_mutex_init failed at " << __FILE__ << ":" << __LINE__ << " - " << strerror(status) << 
      std::endl; exit(status);
	}

	status = pthread_attr_init(&threadAttribute);		// initialize the attribute object
	if (status != 0) {std::cout << "pthread_attr_init failed at " << __FILE__ << ":" << __LINE__ << std::endl; exit(status); }		
	status = pthread_attr_setscope(&threadAttribute, PTHREAD_SCOPE_SYSTEM);
	if (status != 0) {std::cout << "pthread_attr_init failed at " << __FILE__ << ":" << __LINE__ << std::endl; exit(status); }

	// Create two threads and store their IDs in array threadArray
	status = pthread_create(&threadArray[0], &threadAttribute, Thread1, (void*) 1L);
	if (status != 0) { 
		std::cout << "pthread_create failed at " << __FILE__ << ":" << __LINE__ << " - " << strerror(status) << 
      std::endl; exit(status);
	}
	status = pthread_create(&threadArray[1], &threadAttribute, Thread2, (void*) 2L);
	if (status != 0) { 
		std::cout << "pthread_create failed at " << __FILE__ << ":" << __LINE__ << " - " << strerror(status) << 
      std::endl; exit(status);
	}
	status = pthread_attr_destroy(&threadAttribute);
	if (status != 0) {std::cout << "pthread_attr_destroy failed at " << __FILE__ << ":" << __LINE__ << std::endl; exit(status);  }

	// Wait for threads to finish
	status = pthread_join(threadArray[0],NULL);
	if (status != 0) {
		std::cout << "pthread_join failed at " << __FILE__ << ":" << __LINE__ << " - " << strerror(status) << 
      std::endl; exit(status);
	}	
	status = pthread_join(threadArray[1],NULL);
	if (status != 0) {
		std::cout << "pthread_join failed at " << __FILE__ << ":" << __LINE__ << " - " << strerror(status) << 
      std::endl; exit(status);
	}

	status = pthread_mutexattr_destroy(&mutexAttribute);
	if (status !=0) { 
		std::cout << "pthread_mutexattr_destroy failed e at " << __FILE__ << ":" << __LINE__ << " - " << strerror(status) << 
      std::endl; exit(status);
	}

	// Destroy mutex
	status = pthread_mutex_destroy(&mutex);
	if (status !=0) { 
		std::cout << "pthread_mutex_destroy failed  at " << __FILE__ << ":" << __LINE__ << " - " << strerror(status) << 
      std::endl; exit(status);
	}
}

