public final class DiningPhilosophers {

	public static void main (String args[]) {
		diningPhilosopherController d = new diningPhilosopherController();
		Philosopher p0 = new Philosopher (d, 0);
		Philosopher p1 = new Philosopher (d, 1);
		Philosopher p2 = new Philosopher (d, 2);
		Philosopher p3 = new Philosopher (d, 3);
		Philosopher p4 = new Philosopher (d, 4);
		p0.start(); p1.start(); p2.start();	p3.start();	p4.start();
	}
}

final class Philosopher extends Thread {

	private diningPhilosopherController d;
	private int num;

	Philosopher (diningPhilosopherController d, int num) {
 		this.d = d;
		this.num = num;
 	}

	public void run () {
			d.pickUp(num);
			System.out.println ("Philosopher # " + num + " Eating ");
			d.putDown(num);
	}
}

class diningPhilosopherController  {
	private final int n = 5; 				// number of philosophers
	private final int thinking = 0; private final int hungry = 1; private final int eating = 2;
	private int state[] = new int[n];
	private binarySemaphore mutex = new binarySemaphore(1);		// provides mutual exclusion for accessing state[].
	// philosopher i blocks herself on self[i] when she is hungry but unable to eat
	private binarySemaphore self[] = new binarySemaphore[n]; 

   private final int left(int i) { return (n + i - 1) % n; }

   private final int right(int i) { return (i + 1) % n; }
	
	diningPhilosopherController() {
		for (int i = 0; i < n; i++) state[i] = thinking;
		for (int j = 0; j < n; j++) self[j] = new binarySemaphore(0);
	}
	public void pickUp(int i) {
		mutex.P();
		state[i] = hungry;
		test(i);					// performs self[i].V() if philosopher i can eat
		mutex.V();
		self[i].P();			// self[i].P() will not block if self[i].V() was performed during call to test(i)
	}

	public void  putDown(int i) {
 		mutex.P();
		state[i] = thinking;
		test(left(i));		// unblock left neighbor if the neighbor  is hungry and can now eat
		test(right(i));  	// unblock right neighbor if the neighbor is hungry and can now eat
		mutex.V();

	}
	
	void test(int k) {
		// philosopheri calls test(i) to check whether she can eat
		// philosopheri calls test(left(k)) when she is finished eating to unblock a hungry left neighbor 
		// philosopheri calls test(right(k)) when she is finished eating to unblock a hungry right neighbor 
		if ((state[k] == hungry) && (state[left(k)] != eating)
			&& (state[right(k)] != eating)) { 
			state[k] = eating; 
			self[k].V(); 	// unblock neighbor, or guarantee that philosopherk will not block on self[k].P().
		}
	}
	
}


