
#pragma warning(disable:4786) 
#include <iostream>
#include <windows.h>
#include "sharedVariable.h"
#include "TDThread.h"
#include "mutexLock.h"
#include "binarySemaphore.h"
#include "countingSemaphore.h"

const int capacity = 3;

class Buffer {
private:
	sharedVariable<int> buffer[capacity];
	sharedVariable<int> count;
	sharedVariable<int> in, out;
public:
	Buffer() : in(0), out(0), count(0) { 
		for (int i=0; i<capacity; i++) {
			buffer[i].setTrace(sharedVariable<int>::traceOff);
			buffer[i].setAnalysis(sharedVariable<int>::analysisOff);
		}
	}

	int size() { return count;}

	int withdraw () {
		int value = 0;
		value = buffer[out];
		out = (out + 1) % capacity;
		count--;
		return value;
	}

	void deposit (int value) {
		buffer[in] = value;
		in = (in + 1) % capacity;
		count++;
	}

};


Buffer sharedBuffer;	 // 3-slot buffer
mutexLock mutexD;
mutexLock mutexW;
//binarySemaphore mutexD(1), mutexW(1);	// Either a mutexLock or a binarySemaphore can be used. 
										// But the lockset algorithm is not implemented for semaphores,
										//    so spurious races will be detected.
countingSemaphore emptySlots(capacity);
countingSemaphore fullSlots(0);

class Producer : public TDThread {
public:
	Producer(std::string name_) : TDThread(name_), name(name_) {}
private:
	std::string name;
virtual void* run () {
 	int i;
	std::cout << name << " running" << std::endl;
	for (i=0; i<2; i++) {
		emptySlots.P();
		mutexD.lock();
		sharedBuffer.deposit(i);
		std::cout << "Produced: " << i << std::endl;
		mutexD.unlock();
		fullSlots.V();
	}
	return 0;
}
};

class Consumer : public TDThread {
public:
	Consumer(std::string name_) : TDThread(name_), name(name_) {}
private:  
	std::string name;
virtual void* run () {
 	int result;
	std::cout << name << " running" << std::endl;
	for (int i=0; i<2; i++) {
		fullSlots.P();
		mutexW.lock();
		result = sharedBuffer.withdraw();
		mutexW.unlock();
		std::cout << "Consumed: " << result << std::endl;
		emptySlots.V();
	}
	return 0;
}
};

int main() {
	Producer* p1 = new Producer("Producer1");
	Producer* p2 = new Producer("Producer2");
	Producer* p3 = new Producer("Producer3");
	Consumer* c1 = new Consumer("Consumer1");
	Consumer* c2 = new Consumer("Consumer2");
	Consumer* c3 = new Consumer("Consumer3");
	p1->start(); p2->start(); p3->start();
	c1->start(); c2->start(); c3->start();
	p1->join(); p2->join(); p3->join();
	c1->join(); c2->join(); c3->join();
	delete p1; delete p2; delete p3; delete c1; delete c2; delete c3;
	return(0);
}
