
#pragma warning(disable:4786) 
#include "TDThread.h"
#include "runnable.h"
#include "ThreadIDGenerator.h"
#include "threadData.h"
#include <string.h>
#include <string>
#include <assert.h>
#include <windows.h>
#include <process.h>
#include "traceOrReplay.h"
#include "deadlockWatchdog.h"

int TDThread::mode=0;
int TDThread::raceAnalysis=0;
bool TDThread::randomDelay=false;
int TDThread::numControllers=0;
int TDThread::strategy=0;
bool TDThread::detectDeadlock=false;
int TDThread::numRunningThreads=0;
int TDThread::numBlockedThreads=0;

criticalSectionLock TDThread::instanceLock;


TDThread::TDThread(std::auto_ptr<Runnable> runnable_, const std::string tName) 
:  threadName(tName), runnable(runnable_)/*, versionNumber(1),*/ /*control(NULL)*/ {

	completionEvent = CreateEvent( 
     NULL,   		// no security attributes
     1,  	 		// manual reset event
     0,   			// initially nonsignaled
     NULL);  		// unnamed event

	char* env = getenv("MODE");
	if (env!=NULL) {
		if (!(strcmp(env,"TRACE")))
			mode = TRACE;
		else if (!(strcmp(env,"REPLAY")))
			mode = REPLAY;
		else if (!(strcmp(env,"TEST")))
			mode = TEST;
		else if (!(strcmp(env,"SPECTEST")))
			mode = SPECTEST;
		else if (!(strcmp(env,"RT")))
			mode = RT;
		else if (!(strcmp(env,"NONE")))
			mode = NONE;
		else mode = NONE;
	}
	else mode = NONE;
	env = getenv("RACEANALYSIS");
	if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			raceAnalysis = true;
		else raceAnalysis = false;
	}
	else raceAnalysis = false;

	env = getenv("DEADLOCKDETECTION");
	if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			detectDeadlock = true;
		else detectDeadlock = false;
	}
	else detectDeadlock = false;

	if (mode || raceAnalysis) {
		if (threadName != "")
			ID = generateIDFromUserName(threadName);
		else {
			ID = generateID();
			threadName = ThreadIDGenerator::instance().getName(ID);
		}
		// Chapter 2 TDThread is simpler and has some functions disabled
		//if (mode) {
		//	initVectorTS();
		//	constructController();
		//}
	}
	
	hThread = (HANDLE)_beginthreadex(NULL,0,TDThread::startThreadRunnable,
		(LPVOID)this, CREATE_SUSPENDED,&winThreadID );
 
	if (!hThread)
		PrintError("_beginthreadex failed at ",__FILE__,__LINE__);
}

TDThread::TDThread(const std::string tName)
: threadName(tName),  runnable(NULL) {
	completionEvent = CreateEvent( 
     NULL,   		// no security attributes
     1,  	 		// manual reset event
     0,   	      // initially nonsignaled
     NULL);  		// unnamed event

	char* env = getenv("MODE");
	if (env!=NULL) {
		if (!(strcmp(env,"TRACE")))
			mode = TRACE;
		else if (!(strcmp(env,"REPLAY")))
			mode = REPLAY;
		else if (!(strcmp(env,"TEST")))
			mode = TEST;
		else if (!(strcmp(env,"SPECTEST")))
			mode = SPECTEST;
		else if (!(strcmp(env,"RT")))
			mode = RT;
		else if (!(strcmp(env,"NONE")))
			mode = NONE;
		else mode = NONE;
	}
	else mode = NONE;
	env = getenv("RACEANALYSIS");
	if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			raceAnalysis = true;
		else raceAnalysis = false;
	}
	else raceAnalysis = false;

	env = getenv("DEADLOCKDETECTION");
	if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			detectDeadlock = true;
		else detectDeadlock = false;
	}
	else
		detectDeadlock = false;

	if (mode || raceAnalysis) {	
		if (threadName != "") {
			ID = generateIDFromUserName(threadName);
		}
		else {
			ID = generateID();
			threadName = ThreadIDGenerator::instance().getName(ID);
		}
	}
	
	hThread = (HANDLE)_beginthreadex(NULL,0,TDThread::startThread,
		(LPVOID)this, CREATE_SUSPENDED, &winThreadID );
						 
	if (!hThread)
		PrintError("_beginthreadex failed at ",__FILE__,__LINE__);

}

//Thread::Thread(const Thread& aThread){}
//const Thread& Thread::operator= (const Thread& aThread) {return *this;}

TDThread::~TDThread() {
	DWORD rc = CloseHandle(completionEvent);
	if (!rc)
		PrintError("CloseHandle failed at ",__FILE__,__LINE__);
	if (winThreadID != GetCurrentThreadId()) {
      rc = CloseHandle(hThread);
  	  if (!rc)
		PrintError("CloseHandle failed at ",__FILE__,__LINE__);
	}
}

int TDThread::getID() {return ID;}


std::string TDThread::getThreadName() { return threadName; }

/*
threadBasedMsgTracingAndReplay* TDThread::instance(int mode, 
		int numControllers, int strategy, std::string traceFileName) { 
	instanceLock.lock();
	if (m == NULL) {
      	m = new threadBasedMsgTracingAndReplay(mode,numControllers,strategy, traceFileName);
		m->start();
	}
    instanceLock.unlock();
  	return m;
  }
*/
/*
void TDThread::constructController() {
	 char* env = getenv("CONTROLLERS");
	 if (env!=NULL) {
		if (!(strcmp(env,"SINGLE")))
			numControllers = SINGLE;
		else numControllers = MULTIPLE;
	 }
	env = getenv("STRATEGY");
	if (env!=NULL) {
		if (!(strcmp(env,"OBJECT")))
			strategy = OBJECT;
		else strategy = THREAD;
	}
	if (numControllers == MULTIPLE)
		traceFileName = threadName;
	else
		traceFileName = "channel";
	 if (strategy == THREAD) {
			if (mode == TRACE || mode == REPLAY || mode == TEST || mode == RT) {
				if (numControllers == SINGLE)
					if (mode == RT) 
						control = (Controller*) msgTracingAndReplay::getInstance(mode,numControllers,strategy,traceFileName);
					else
						control = instance(mode,numControllers,strategy,traceFileName);
				else {
					control = new threadBasedMsgTracingAndReplay(mode,numControllers,strategy,traceFileName);
					control->start();
				}
			}
		 }
	}
*/			
//Controller* TDThread::getController() {
//	if (mode == RT) 
//		return /*(msgTracingAndReplay*)*/control;
//	else
//		return /*(threadBasedMsgTracingAndReplay*)*/control;
//	}


//void TDThread::initVectorTS() {}

void TDThread::start() {
	assert(hThread != NULL); 
	DWORD rc = ResumeThread(hThread);
 	if (!rc)
		PrintError("ResumeThread failed at ",__FILE__,__LINE__);
	if (mode == TRACE && randomDelay) {
		Sleep(rand()%delayAmount); // delay parent after starting child
								   // default delayAmount is 750
	}
}


void* TDThread::join() {
	DWORD rc = WaitForSingleObject( 
        	completionEvent,		// handle to event
        	INFINITE);				// no time-out 
	if (!(rc==WAIT_OBJECT_0))
		PrintError("WaitForSingleObject failed at ",__FILE__,__LINE__);
	return result;
}

void TDThread::setCompleted() {	
	DWORD rc = SetEvent(completionEvent); // release all threads and leave event signalled
 	if (!rc)
		PrintError("SetEvent failed at ",__FILE__,__LINE__);
}


int TDThread::generateID() {
	const std::string thisClassName = "Thread";
	
	data d = threadData::instance().getThreadData(GetCurrentThreadId());// return child or main
	threadName = d.threadName;

	// threadName is either a child thread or "main", as threadData sets main's name to "main"

	threadName = threadName + "_" + thisClassName;
		
	int myID = ThreadIDGenerator::instance().getID(threadName);
	return myID;
}

int TDThread::generateIDFromUserName(std::string userName) {
	return ThreadIDGenerator::instance().getIDFromUserName(userName);
}


unsigned int WINAPI TDThread::startThreadRunnable(LPVOID pVoid){
			TDThread* runnableThread = static_cast<TDThread*> (pVoid);
			if (runnableThread != NULL) {
				if (mode || raceAnalysis) {
				   data d(runnableThread->ID,runnableThread->threadName);
				   threadData::instance().addThreadData(GetCurrentThreadId(),d);
				}
				if (mode == TRACE && randomDelay) {
					srand(GetCurrentThreadId()); // each thread must call srand()
					Sleep(rand()%delayAmount);	 // default delayAmount is 750
				}
				if (detectDeadlock && (mode == TRACE || mode == RT)) {
					incNumRunningThreads();
					deadlockWatchdog::startInstance();
					deadlockWatchdog::changeStatus(GetCurrentThreadId(),"running");
				}
				runnableThread->runnable->run();
				runnableThread->setCompleted();
				if (detectDeadlock && (mode == TRACE || mode == RT)) {
					decNumRunningThreads();
					deadlockWatchdog::removeThread(GetCurrentThreadId());
				}
			}
			return reinterpret_cast<unsigned>(runnableThread->result);
}	
 
unsigned int WINAPI TDThread::startThread(LPVOID pVoid) {
			TDThread* aThread = static_cast<TDThread*> (pVoid);
			if (aThread != NULL) {
				if (mode || raceAnalysis) {
				   data d(aThread->ID,aThread->threadName);
				   threadData::instance().addThreadData(GetCurrentThreadId(),d);
				}
				if (mode == TRACE && randomDelay) {
					srand(GetCurrentThreadId()); // each thread must call srand()
					Sleep(rand()%delayAmount);	 // default delayAmount is 750
				}
				if (detectDeadlock && (mode == TRACE || mode == RT)) {
					incNumRunningThreads();
					deadlockWatchdog::startInstance();
					deadlockWatchdog::changeStatus(GetCurrentThreadId(),"running");
				}
				aThread->run();
				aThread->setCompleted();
				if (detectDeadlock && (mode == TRACE || mode == RT)) {
					decNumRunningThreads();
					deadlockWatchdog::removeThread(GetCurrentThreadId());
				}
			}
			return reinterpret_cast<unsigned>(aThread->result);			
}

void TDThread::PrintError(LPTSTR lpszFunction,LPSTR fileName, int lineNumber)  { 
	TCHAR szBuf[256]; 
	LPSTR lpErrorBuf;
	DWORD errorCode = GetLastError(); 
	FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | 
	     FORMAT_MESSAGE_FROM_SYSTEM, NULL,  errorCode, MAKELANGID(LANG_NEUTRAL, 		     SUBLANG_DEFAULT),  (LPTSTR) &lpErrorBuf, 0, NULL );
	wsprintf(szBuf, "%s failed at line %d in %s with error %d: %s", lpszFunction, lineNumber, fileName, 
  	     errorCode, lpErrorBuf); 
 	DWORD numWritten;
	WriteFile(GetStdHandle(STD_ERROR_HANDLE),szBuf,strlen(szBuf),&numWritten,FALSE);
	LocalFree(lpErrorBuf);
	exit(errorCode); 
}

void TDThread::incNumRunningThreads() {instanceLock.lock(); numRunningThreads++; instanceLock.unlock();}
void TDThread::decNumRunningThreads() {instanceLock.lock();numRunningThreads--;instanceLock.unlock();}
void TDThread::incNumBlockedThreads() {instanceLock.lock();numBlockedThreads++;instanceLock.unlock();}
void TDThread::decNumBlockedThreads() {instanceLock.lock();numBlockedThreads--;instanceLock.unlock();}
int TDThread::getNumRunningThreads() {instanceLock.lock(); int x = numRunningThreads; instanceLock.unlock(); return x;}
int TDThread::getNumBlockedThreads() {instanceLock.lock(); int x = numBlockedThreads; instanceLock.unlock(); return x;}
void TDThread::resetNumBlockedRunningThreads() {instanceLock.lock();numRunningThreads=0;numBlockedThreads=0;instanceLock.unlock();}
void TDThread::setNumBlockedThreads(int value) {instanceLock.lock();numBlockedThreads=value;instanceLock.unlock();}


