#ifndef SHAREDVARIABLE_H
#define SHAREDVARIABLE_H
#pragma warning(disable:4786) 
#include <windows.h>
#include "assert.h"
#include <iostream>
#include "win32critical_section.h"
#include "lockdata.h"
#include "mutexlocker.h"
#include "threaddata.h"
#include <set>
#include <vector>
#include <list>
#include "traceOrReplay.h"
#include "sharedvariablecontroller.h"
#include "sharedvariableidgenerator.h"


template <class T>
class sharedVariable {
public:
	enum analysisOnOff {analysisOff, analysisOn};
	enum tracingOnOff {traceOff, traceOn};

	/* constructors and destructor */
    sharedVariable();
	sharedVariable(analysisOnOff aOnOff, tracingOnOff tOnOff);
	explicit sharedVariable(const T& v);
	sharedVariable(const T& v, analysisOnOff on, tracingOnOff onOff);
	sharedVariable(const sharedVariable<T>& sv);
	~sharedVariable();

	/* casting and assignment operators */
	operator T() const;
	const sharedVariable<T>& operator=(const sharedVariable<T>& sv);
	const sharedVariable<T>& operator=(const T& v);

	/* member funnctions: arithmetic binary assignment */
	const sharedVariable<T>& operator+=(const sharedVariable<T>& sv);
	const sharedVariable<T>& operator-=(const sharedVariable<T>& sv);
	const sharedVariable<T>& operator*=(const sharedVariable<T>& sv);
	const sharedVariable<T>& operator/=(const sharedVariable<T>& sv);
	const sharedVariable<T>& operator%=(const sharedVariable<T>& sv);

	/* member functions: bit-wise binary assignment operators */
	const sharedVariable<T>& operator&=(const sharedVariable<T>& sv);
	const sharedVariable<T>& operator|=(const sharedVariable<T>& sv);
	const sharedVariable<T>& operator^=(const sharedVariable<T>& sv);
	const sharedVariable<T>& operator>>=(const sharedVariable<T>& sv);
	const sharedVariable<T>& operator<<=(const sharedVariable<T>& sv);


	friend std::ostream& operator<<(std::ostream& out, const sharedVariable<T>& sv);
	/* boolean relational operators */
	friend bool operator==(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend bool operator==(const sharedVariable<T>& sv1, const T& sv2);
	friend bool operator==(const T& sv1, const sharedVariable<T>& sv2);
	friend bool operator!=(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend bool operator!=(const sharedVariable<T>& sv1, const T& sv2);
	friend bool operator!=(const T& sv1, const sharedVariable<T>& sv2);
	friend bool operator&&(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend bool operator&&(const sharedVariable<T>& sv1, const bool& sv2);
	friend bool operator&&(const sharedVariable<T>& sv1, const T& sv2);
	friend bool operator&&(const T& sv1, const sharedVariable<T>& sv2);
	friend bool operator||(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend bool operator||(const sharedVariable<T>& sv1, const bool& sv2);
	friend bool operator||(const sharedVariable<T>& sv1, const T& sv2);
	friend bool operator||(const T& sv1, const sharedVariable<T>& sv2);
	friend bool operator<(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend bool operator<(const sharedVariable<T>& sv1, const T& sv2);
	friend bool operator<(const T& sv1, const sharedVariable<T>& sv2);
	friend bool operator>(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend bool operator>(const sharedVariable<T>& sv1, const T& sv2);
	friend bool operator>(const T& sv1, const sharedVariable<T>& sv2);
	friend bool operator<=(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend bool operator<=(const sharedVariable<T>& sv1, const T& sv2);
	friend bool operator<=(const T& sv1, const sharedVariable<T>& sv2);
	friend bool operator>=(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend bool operator>=(const sharedVariable<T>& sv1, const T& sv2);
	friend bool operator>=(const T& sv1, const sharedVariable<T>& sv2);

	friend sharedVariable<T> operator+(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator+(const T& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator+(const sharedVariable<T>& sv1, const T& sv2);
	
	friend sharedVariable<T> operator-(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator-(const T& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator-(const sharedVariable<T>& sv1, const T& sv2);
	
	friend sharedVariable<T> operator*(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator*(const T& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator*(const sharedVariable<T>& sv1, const T& sv2);
	
	friend sharedVariable<T> operator/(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator/(const T& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator/(const sharedVariable<T>& sv1, const T& sv2);

	friend sharedVariable<T> operator%(const sharedVariable<T>& sv1, const T& sv2);
	friend sharedVariable<T> operator%(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator%(const T& sv1, const sharedVariable<T>& sv2);

	/* bitwise binary operators */
	friend sharedVariable<T> operator&(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator&(const T& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator&(const sharedVariable<T>& sv1, const T& sv2);
	
	friend sharedVariable<T> operator|(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator|(const T& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator|(const sharedVariable<T>& sv1, const T& sv2);

	friend sharedVariable<T> operator^(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator^(const T& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator^(const sharedVariable<T>& sv1, const T& sv2);

	friend sharedVariable<T> operator>>(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator>>(const T& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator>>(const sharedVariable<T>& sv1, const T& sv2);

	friend sharedVariable<T> operator<<(const sharedVariable<T>& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator<<(const T& sv1, const sharedVariable<T>& sv2);
	friend sharedVariable<T> operator<<(const sharedVariable<T>& sv1, const T& sv2);

	/* member functions: postfix and prefix operators */
	const sharedVariable<T>& operator++();
	const sharedVariable<T> operator++(int);
	const sharedVariable<T>& operator--();
	const sharedVariable<T> operator--(int);

	/* member functions: unary operators */
	const sharedVariable<T> operator+();
	const sharedVariable<T> operator-();
	const sharedVariable<T> operator~();
	bool operator!();

	void setAnalysis(analysisOnOff on);
	bool getAnalysisOn();
	void setTrace(tracingOnOff on);
	bool getTraceOn();

private:
	bool computeLockSet() const;
	bool raceDetected() const;
	void displayLockset(std::string msg) const;
	void startRead(const sharedVariable<T>& sv) const;
	void endRead(const sharedVariable<T>& sv) const;
	void startWrite(const sharedVariable<T>& sv) const;
	void endWrite(const sharedVariable<T>& sv) const;
	T value;
	// one lockSet for each shared varible
	mutable std::list<int> lockSet;
	mutable bool first; // first is true if first call to computeLockset
	mutable bool writtenAtLeastOnce; // true if this var written to at least once
	mutable bool analysisIsOn; 	// on if checking for races on this sharedVariable
	mutable bool tracingIsOn;     // on if tracing is on for this sharedVariable
	mutable win32Critical_Section dataLock;   // one lock for each sharedVariable object
	sharedVariableController* SVcontrol;   // pointer since need to generate argument for constructor
	std::string generateID();
	int mode;			// set by reading environment variable MODE
	bool raceAnalysis;			// set by reading environment variable RACEANALYSIS
	/* These functions are disallowed */
	const sharedVariable<T>& operator&();
	const sharedVariable<T>& operator*();
};

#include "sharedvariable.inl"

#endif






