#include "deadlockWatchDog.h"
#include <assert.h>
#include "traceorreplay.h"
#include "TDThread.h"
#include "threadData.h"

criticalSectionLock deadlockWatchdog::classLock;
criticalSectionLock deadlockWatchdog::mutex;
deadlockWatchdog* deadlockWatchdog::instance = NULL;
std::map<int,std::string> deadlockWatchdog::status;
int deadlockWatchdog::mode = 0;
bool deadlockWatchdog::detectDeadlock = false;

void deadlockWatchdog::changeStatus(int ID, std::string statusMessage) {
	classLock.lock();
	std::pair<std::map<int, std::string>::iterator,bool> p = 
	status.insert(std::make_pair(ID,statusMessage));
	if (!p.second) // system reuses thread IDs so be prepared to reuse winThreadID
		p.first->second = statusMessage;
	//std::cout << "changed status of " << ID << " to " << statusMessage << std::endl;
	classLock.unlock();
}
void deadlockWatchdog::removeThread(int ID) {
	classLock.lock();
	int num = status.erase(ID);
	assert(num==1); // data must be present
	classLock.unlock();
}

deadlockWatchdog* deadlockWatchdog::startInstance() { 
	classLock.lock();
   	if (instance == NULL) {
   		instance = new deadlockWatchdog();
		instance->start();
		if (detectDeadlock)
			std::cout << "Monitoring for Deadlock." << std::endl;
	}
	classLock.unlock();
	return instance;

}//end startInstance()
deadlockWatchdog::deadlockWatchdog() : stopped(false) {
	char* env = getenv("MODE");
	if (env!=NULL) {
		if (!(strcmp(env,"TRACE")))
			mode = TRACE;
		else if (!(strcmp(env,"REPLAY")))
			mode = REPLAY;
		else if (!(strcmp(env,"TEST")))
			mode = TEST;
		else if (!(strcmp(env,"SPECTEST")))
			mode = SPECTEST;
		else if (!(strcmp(env,"RT")))
			mode = RT;
		else if (!(strcmp(env,"NONE")))
			mode = NONE;
		else mode = NONE;
	}
	else mode = NONE;

	env = getenv("DEADLOCKDETECTION");
	if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			detectDeadlock = true;
		else detectDeadlock = false;
	}
	else detectDeadlock = false;
}

void deadlockWatchdog::notifyDeadlock() {
	// if RT controller detects a deadlock when -DdeadlockDetection=on, the RT
	// controller calls this method so that the watchdog will detect the
	// deadlock. This is necessary since, in one case, a thread can become			
	// blocked on a call to the RT controller, and this block will not be
	// visible to this watchdog thread, i.e., this watchdog will consider
	// the blocked thread to still be running
	TDThread::setNumBlockedThreads(TDThread::getNumRunningThreads());		
}

void deadlockWatchdog::stopWatchdog() {
	TDThread::resetNumBlockedRunningThreads();  
	instance->stopped = true;
	//BOOL TerminateThread(  //If the function succeeds, the return value is nonzero.
	//	HANDLE hThread,
	//	DWORD dwExitCode
	//); // don't forget to closeHandle

	//instance.interrupt();

	instance = NULL;    // old WatchDog may wakeup and do one last check and
}						// detect same deadlock as new watchDog. Below,
						// we lock detection code so only one of them can do it
						// that is, first to detect deadlock executes System.exit()

void* deadlockWatchdog::run() {
	
	bool deadlockDetected = false;
	//std::cout << "Watchdog running." << std::endl;
	while (!deadlockDetected && !stopped) {
		Sleep(500);
		if (deadlockDetected || stopped) return 0;
		mutex.lock(); // in case there are two watchDog's running 
		//std::cout << "running:" << TDThread::getNumRunningThreads() << std::endl;
		//std::cout << "blocked:" << TDThread::getNumBlockedThreads() << std::endl;			
		if (TDThread::getNumRunningThreads() == TDThread::getNumBlockedThreads() &&
			TDThread::getNumRunningThreads()>0 ) {
				std::cout << "Deadlock detected:" << std::endl;
					deadlockDetected = true;
					std::map< int, std::string>::iterator pos;
					for (pos = status.begin(); pos != status.end(); pos++) {
						std::cout << "- Thread " << ((threadData::instance().getThreadData(pos->first)).threadName) << " " << pos->second << std::endl;
					}
					exit(1);
			}
			else if (mode == TRACE && TDThread::getNumRunningThreads()==0)
				stopWatchdog();
		mutex.unlock();
	}
	
	return 0;
}
