
#ifndef MUTEXLOCK_H
#define MUTEXLOCK_H

#include <windows.h>
#include "criticalSectionLock.h"
#include "lockData.h"
#include "mutexLocker.h"
#include "traceOrReplay.h"
#include "semaphorePool.h"
#include "win32Semaphore.h"
#include "semaphoreAndLockTracingAndReplay.h"
#include <queue>

class mutexLock {
// simple class for lock/unlock on a mutex lock
private:
	semaphorePool pool; // resource pool of semaphores
	int getMutexID() const { return mutexID;}
	int mutexID;	         // each mutexLock has an integer ID 
	bool analysisIsOn;		 // true if race analysis is requested
	bool tracingIsOn;		 // true if tracing/replay is requested
	bool raceAnalysis;		// true if env var RACEANALYSIS set to ON
	int numControllers;
	int mode;		// true if env var MODE is TRACE or REPLAY
	bool randomDelay;
	bool detectDeadlock;
	std::string generateID();
	std::string mutexName;
	static semaphoreAndLockTracingAndReplay* instance(int mode, 
			int numControllers, std::string traceFileName);
	semaphoreAndLockTracingAndReplay* control;
    static criticalSectionLock idLock;   // static lock for acessing ID;
	static int ID;        // mutexLock IDs start with 1; 
	static int getID() {   // called by constructor when mutexLock is created
		mutexLocker<criticalSectionLock> mutex(idLock); // lock (unlock) idLock in contructor (destructor) 
		return ID++;
	}
	std::queue<win32Semaphore*> waitingLock; // queue of threads blocked on lock()

	unsigned long owner;
	int count;
	win32Semaphore mutex;
public:
  void lock();
  int unlock();
  mutexLock();
  ~mutexLock();
  enum analysisOnOff {analysisOff, analysisOn};
  enum tracingOnOff {traceOff, traceOn};
  mutexLock(analysisOnOff aOnOff, tracingOnOff tOnOff);
};

#endif


