#include <iostream>
#include <pthread.h>
#include "thread.h"

class boundedBuffer  {
private:
	int fullSlots;	// # of full slots in the buffer
	int capacity;
	int* buffer;
	int in, out;
	pthread_cond_t notFull;	// Producers wait on notFull
	pthread_cond_t notEmpty;	// Consumers wait on notEmpty
	pthread_mutex_t mutex;	// exclusion for deposit() and withdraw()
public:
	boundedBuffer(int capacity_) : capacity(capacity_), fullSlots(0), in(0), 
			out(0), buffer(new int[capacity_]) { 	
		pthread_cond_init(&notFull,NULL); 	pthread_cond_init(&notEmpty,NULL);
		pthread_mutex_init(&mutex,NULL);	
	}
	~boundedBuffer() {
		delete [] buffer;
		pthread_cond_destroy(&notFull); pthread_cond_destroy(&notEmpty);
		pthread_mutex_destroy(&mutex);
	}
	void deposit(int value) {
		pthread_mutex_lock(&mutex);
		while (fullSlots == capacity)
			pthread_cond_wait(&notFull,&mutex);
		buffer[in] = value;
		in = (in + 1) % capacity;
		++fullSlots;
		pthread_mutex_unlock(&mutex);
		pthread_cond_signal(&notEmpty);
	}
	int withdraw() {
		pthread_mutex_lock(&mutex);
		int value;
		while (fullSlots == 0) 
			pthread_cond_wait(&notEmpty,&mutex);
		value = buffer[out];
		out = (out + 1) % capacity;
		--fullSlots;
		pthread_mutex_unlock(&mutex);
		pthread_cond_signal(&notFull);
		return value;
	}
};

class Producer : public Thread {
private:
	boundedBuffer& b;
	int num;
public:
	Producer (boundedBuffer* b_, int num_) : b(*b_),  num(num_) { }
	virtual void* run () {
		std::cout << "Producer Running" << std::endl;
		for (int i = 0; i < 3; i++) {
			b.deposit(i);
			std::cout << "Producer # " << num << " deposited " << i << std::endl;
		}
		return NULL;
	}
};

class Consumer : public Thread {
private:
	boundedBuffer& b;
	int num;
public:
	Consumer (boundedBuffer* b_, int num_) : b(*b_), num(num_) { }
	virtual void* run () {
		std::cout << "Consumer Running" << std::endl;
		int value = 0;
		for (int i = 0; i < 3; i++) {
			value = b.withdraw();
			std::cout << "Consumer # " << num << " withdrew " << 
			value << std::endl;
		}
	return NULL;
	}
};

int main ( ) {
	boundedBuffer* b1 = new boundedBuffer(3);
	Producer p1(b1, 1); Consumer c1(b1, 1);
	p1.start(); c1.start();
	p1.join(); c1.join();  
	delete b1;
	return 0;
}

//Listing 4.15 C++/Pthreads class boundedBuffer.

