import java.util.concurrent.locks.*;
final class boundedBuffer {
	private int fullSlots=0; private int capacity = 0; private int in = 0, out = 0;
	private int[] buffer = null;
	private ReentrantLock mutex;
	private Condition notFull;
	private Condition notEmpty;

	public boundedBuffer(int bufferCapacity) {
		capacity = bufferCapacity; buffer = new int[capacity];
		mutex = new ReentrantLock();
		// notFull and notEmpty are both attached to mutex
		notFull = mutex.newCondition(); notEmpty = mutex.newCondition();
	}
	public void deposit (int value) throws InterruptedException {
		mutex.lock();
		try {		
			while (fullSlots == capacity)
				notFull.await();
			buffer[in] = value;
			in = (in + 1) % capacity;
			notEmpty.signal();
		} finally {mutex.unlock();}
	}
	public synchronized int withdraw () throws InterruptedException {
		mutex.lock();
		try {
			int value = 0;
			while (fullSlots == 0)
				notEmpty.await();
			value = buffer[out];
			out = (out + 1) % capacity;
			notFull.signal();
			return value;
		} finally {mutex.unlock();}
	}
}
//Listing 4.16 Java class boundedBuffer using Condition objects.

