import java.util.*;

public final class countingSemaphore extends semaphore {
// a FIFO counting semaphore

	private ArrayList waitingP = new ArrayList(); // queue of threads blocked on P

	public countingSemaphore(int initialPermits) {super(initialPermits);}

	public void P() {
		Object conditionVar = new Object(); // each thread blocks on its own conditionVar object
		synchronized (conditionVar) {
			synchronized (this) { // lock semaphore
				permits--;
				if (permits>=0)  // then no need to block thread
					return;
				waitingP.add(conditionVar); // otherwise append blocked thread
			} // end synchronized (this) to avoid conditionVar.wait() while holding lock on this
			try {conditionVar.wait();} catch (InterruptedException ex) {}
		} // end synchronized (conditionVar)
	} // end P

	public synchronized void V() {
		permits++;
		if (permits>0) // then no thread to signal
			return;
		if (waitingP.size() > 0) { // this should always be true since P()'s have blocked
			Object oldest = waitingP.get(0);
			waitingP.remove(0);
			synchronized(oldest) {oldest.notify();}
		}	
	} // end

	private boolean doP() {
	// called by VP() operation, decrements permits and returns
   // true if P() should block; false otherwise.
		permits--;
		if (permits >= 0)
			return false;	// no need to block thread
		else
			return true;
	}

	public int getSemaphoreID() { return semaphoreID;}

	public void VP(semaphore vSem) {
	// execute {vSem.V(); this.P();} without any intervening P() or V() operations on this or vSem.	

		// lock semaphores in ascending order of IDs to prevent circular deadlock (i.e. T1 holds
		// this's lock and waits for vSem's lock while T2 holds vSem's lock and waits for this's lock.)
		semaphore first = this;
 		semaphore second = vSem;
		if (this.getID() > vSem.getID()) { 
			first = vSem;
			second = this;
		}
		Object conditionVar = new Object(); // each thread blocks on its own conditionVar
		synchronized(conditionVar) {
			synchronized(first) { 
				synchronized(second) { 

					//vSem.V() must not block
					if (vSem instanceof binarySemaphore && vSem.permits == 1)
						throw new IllegalArgumentException("V() will block");

					// perform vSem.V()
					vSem.V(); 		// it's okay that we already hold vSem's lock
					// perform this.P()
					boolean blockingP = doP();
					if (!blockingP)
						return;
					waitingP.add(conditionVar); // append blocked thread
				} // end synchronized(second)
			} // end synchronized(first)
			try {conditionVar.wait();} catch (InterruptedException ex) {}
		} // end synchronized (conditionVar)
	} // end VP()


}// end  countingSemaphore

