#include <iostream>
#include <windows.h>
#include "monitorSC.h"
#include "conditionVariable.h"
#include "thread.h"

class Buffer : private monitorSC {
private:
int fullSlots; 			// # of full slots in the buffer
int capacity;
int* buffer;
int in, out;
conditionVariable notFull;
conditionVariable notEmpty;

public:
	Buffer(int capacity_) : fullSlots(0), capacity(capacity_), in(0), out(0),
	notFull(this), notEmpty(this){ 				
		buffer = new int[capacity];	
	}

	void deposit(int value) {
		enterMonitor();
		while (fullSlots == capacity) 
			notFull.waitC();
		buffer[in] = value;
		in = (in + 1) % capacity;
		++fullSlots;
		notEmpty.signalC();
		exitMonitor();
	}

	int withdraw() {
		enterMonitor();
		int value;
		while (fullSlots == 0) 
			notEmpty.waitC();
		value = buffer[out];
      	out = (out + 1) % capacity;
		--fullSlots;
		notFull.signalC();
		exitMonitor();
		return value;
	}
};


class Producer : public Thread {
private:
	Buffer& b;
	int num;
public:
	Producer (Buffer* b_, int num_) : b(*b_),  num(num_) { }
	void* run () {
		std::cout << "Producer Running" << std::endl;
		for (int i = 0; i < 3; i++) {
			b.deposit(i);
			std::cout << "Producer # " << num << " deposited " << i << std::endl;
		}
		return 0;
	}
};

class Consumer : public Thread {
private:
	Buffer& b;
	int num;
public:
	Consumer (Buffer* b_, int num_) : b(*b_), num(num_) { }
	void* run () {
		std::cout << "Consumer Running" << std::endl;
 		int value = 0;
		for (int i = 0; i < 3; i++) {
			value = b.withdraw();
			std::cout << "Consumer # " << num << " withdrew " << value << std::endl;
		}
		return 0;
	}
};


int main ( ) {
	Buffer* c = new Buffer(3); 
	Producer* p1 = new Producer (c, 1); Consumer* c1 = new Consumer (c, 1);
	p1->start(); c1->start();
	c1->join(); 
	p1->join();
 	return 0;
}

