
#include "win32Semaphore.h"
#include <iostream>

void PrintError(LPTSTR lpszFunction,LPSTR fileName, int lineNumber) { 
	TCHAR szBuf[256]; LPSTR lpErrorBuf;
	DWORD errorCode = GetLastError(); 
	FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | 
	FORMAT_MESSAGE_FROM_SYSTEM, NULL, errorCode, 
	MAKELANGID(LANG_NEUTRAL, 
	SUBLANG_DEFAULT), (LPTSTR) &lpErrorBuf, 0, NULL );
	wsprintf(szBuf, "%s failed at line %d in %s with error %d: %s", lpszFunction, 
	lineNumber, fileName, errorCode, lpErrorBuf); 
 	DWORD numWritten;
	WriteFile(GetStdHandle(STD_ERROR_HANDLE),szBuf,strlen(szBuf),
	&numWritten,FALSE);
	LocalFree(lpErrorBuf);
	exit(errorCode); 
}


win32Semaphore :: win32Semaphore(int initial) : initialValue(initial), maxValue(maxDefault) {
	hSemaphore = CreateSemaphore( 
	NULL,	// no security attributes
	initial,	// initial count
	maxValue,	// maximum count
	NULL);	// unnamed semaphore
	if (!hSemaphore)
	PrintError("CreateSemaphore",__FILE__,__LINE__);
}
win32Semaphore :: win32Semaphore(int initial, int max) : initialValue(initial), 
	maxValue(max) {
	hSemaphore = CreateSemaphore( 
	NULL,	// no security attributes
	initial,	// initial count
	maxValue,	// maximum count
	NULL);	// unnamed semaphore
	if (!hSemaphore)
	PrintError("CreateSemaphore",__FILE__,__LINE__);
}
win32Semaphore :: ~win32Semaphore() {
	DWORD rc = CloseHandle(hSemaphore);
	if (!rc)
	PrintError("CloseHandle",__FILE__,__LINE__);
}
void win32Semaphore :: P() {	
	DWORD rc = WaitForSingleObject( 
		hSemaphore,	// handle to semaphore
		INFINITE);	// no time-out 
	if (!rc==WAIT_OBJECT_0) {
		std::cout << "!rc is " << !rc << std::endl;
	 PrintError("WaitForSingleObject",__FILE__,__LINE__);
	}
}
DWORD win32Semaphore :: P(long timeout) {
	DWORD rc = WaitForSingleObject( 
	hSemaphore,	// handle to semaphore
	timeout);	// no time-out 
	if (!(rc==WAIT_OBJECT_0)||(rc==WAIT_TIMEOUT))
	PrintError("WaitForSingleObject failed at ",__FILE__,__LINE__);
	return rc;
}
void win32Semaphore :: V() {
	DWORD rc = ReleaseSemaphore( 
	hSemaphore,	// handle to semaphore
	1,	// increase count by one
	NULL);	// not interested in previous count
	if (!rc)
	PrintError("ReleaseSemaphore failed at ",__FILE__,__LINE__);
}


//Listing 3.22 Class win32Semaphore.
