#pragma warning(disable:4786) 
#include "semaphore.h"
#include <queue>
#include "win32Semaphore.h"
#include "binarySemaphore.h"
#include <assert.h>
#include <typeinfo>
#include <string.h>
#include <iostream>


binarySemaphore::binarySemaphore(int initialPermits)
 : semaphore(initialPermits) {}

binarySemaphore::~binarySemaphore() {}


void binarySemaphore::P() {
   lock();
   if (permits==1)  { // then no need to block thread
     if (waitingV.size() > 0) {  // signal thread blocked in V()
       win32Semaphore* oldestConditionVar = (waitingV.front()).conditionVar;
       waitingV.pop();
       oldestConditionVar->V();
     }
     else {
       permits=0;
	 }
	 unlock();
     return;
   }
   waitRecord rec;
   rec.conditionVar = pool.acquire();
   waitingP.push(rec); // otherwise append blocked thread
   unlock();
   rec.conditionVar->P();
   pool.release(rec.conditionVar);
} // end P()

void binarySemaphore::V() {
   lock(); // lock semaphore
   if (permits==0) { // then no need to block thread
     if (waitingP.size() > 0) {  // signal thread blocked in P()
       win32Semaphore* oldestConditionVar = (waitingP.front()).conditionVar;
       waitingP.pop();
       oldestConditionVar->V();
     }
     else {
       permits=1;
	 }
	 unlock();
     return;
   }
   waitRecord rec;
   rec.conditionVar = pool.acquire();
   waitingV.push(rec); // otherwise append blocked thread
   unlock(); // end synchronized (this) to avoid conditionVar.wait() while holding lock on this
   rec.conditionVar->P();
   pool.release(rec.conditionVar);
} // end V()


bool binarySemaphore::doP() {
// Called by VP() operation; checks permit and unblocks a thread waiting
// in V() if necessary. Returns true if P() should block; false otherwise.
   if (permits==1)  { // P() does not block
     if (waitingV.size() > 0) {  // signal thread blocked in V()
       win32Semaphore* oldestConditionVar = (waitingV.front()).conditionVar;
       waitingV.pop();
       oldestConditionVar->V();
     }
     else
       permits=0;
     return false;
    } // P() does not block
    else
       return true;	// P() does block
}

void binarySemaphore::VP(semaphore* vSem) {
// execute {vSem.V(); this.P();} without any intervening 
// P() or V() operations on this or vSem.	

// lock semaphores in ascending order of IDs to prevent circular deadlock (i.e. T1 holds
// this's lock and waits for vSem's lock while T2 holds vSem's lock and waits for this's lock.)
   semaphore* first = this;
   semaphore* second = vSem;
   if (this->getSemaphoreID() > vSem->getSemaphoreID()) { 
     first = vSem;
     second = this;
   }
   first->lock(); 
   second->lock();
   //vSem.V() must not block
   //if (!strcmp(typeid(*vSem).name(),"binarySemaphore") && vSem->permits==1)
	if (vSem->permits==1 && vSem->isBinarySemaphore())
	   std::cout << "illegal VP operation" << std::endl;
   // perform vSem.V()
   vSem->V(); // it's okay that we already hold vSem's lock
   // perform this.P()
   bool blockingP = doP();
   if (!blockingP) {
     second->unlock();
     first->unlock();
     return;
   }
   waitRecord rec;
   rec.conditionVar = pool.acquire();
   waitingP.push(rec); // otherwise append blocked thread
   second->unlock();
   first->unlock();
   rec.conditionVar->P();
   pool.release(rec.conditionVar);
} // end VP()

