#pragma warning(disable:4786) 
#include "thread.h"
#include "runnable.h"
#include <string.h>
#include <string>
#include <assert.h>
#include <windows.h>
#include <process.h>

typedef unsigned (WINAPI *PBEGINTHREADEX_THREADFUNC) (
    LPVOID lpThreadParameter
		);
typedef unsigned *PBEGINTHREADEX_THREADID;

Thread::Thread(Runnable* runnable_, const std::string tName, size_t stackSize_) 
: stackSize(stackSize_), threadName(tName), runnable(runnable_) {

	completionEvent = CreateEvent( 
     NULL,   		// no security attributes
     1,  	 		// manual reset event
     0,   	      // initially nonsignaled
     NULL);  		// unnamed event
	
  hThread = (HANDLE)_beginthreadex(NULL,stackSize,
	  (PBEGINTHREADEX_THREADFUNC) Thread::startThreadRunnable,
		(LPVOID)this, CREATE_SUSPENDED, 
		(PBEGINTHREADEX_THREADID) &winThreadID );
 
	assert(hThread != NULL);

}

Thread::Thread(const std::string tName, size_t stackSize_)
: stackSize(stackSize_), threadName(tName),  runnable(NULL) {

	completionEvent = CreateEvent( 
     NULL,   		// no security attributes
     1,  	 		// manual reset event
     0,   	        // initially nonsignaled
     NULL);  		// unnamed semaphore
	  
	hThread = (HANDLE)_beginthreadex(NULL,stackSize,
	  (PBEGINTHREADEX_THREADFUNC)Thread::startThread,
		(LPVOID)this, CREATE_SUSPENDED, 
		(PBEGINTHREADEX_THREADID) &winThreadID );
						 
	assert(hThread != NULL);
	 
}

//Thread::Thread(const Thread& aThread){}

//const Thread& Thread::operator= (const Thread& aThread) {return *this;}

Thread::~Thread() {
	CloseHandle(completionEvent);
	if (winThreadID != GetCurrentThreadId())
      CloseHandle(hThread);
	if (runnable != NULL)
	  delete runnable;
}
 
void Thread::start() {
	assert(hThread != NULL); 
	DWORD ret = ResumeThread(hThread);
	assert(ret == 1);
}

unsigned long Thread::getThreadID() {
    return winThreadID;
}

void Thread::exit(int code) {
    assert(winThreadID != GetCurrentThreadId());
    TerminateThread(hThread, code);
    delete this;
}

unsigned long Thread::currentThreadID() {
    return GetCurrentThreadId();
}

void Thread::exitCurrentThread(int code) {
    ExitThread(code);
}

void Thread::join() {
	DWORD dwWaitResult; 
	dwWaitResult = WaitForSingleObject( 
        	completionEvent,		// handle to event
        	INFINITE);          // no time-out 
}

void Thread::setCompleted() {
	if (!SetEvent(completionEvent)) // release all threads and leave event signalled
   	{ ; }    // deal with the error.
}

const std::string Thread::getThreadName() {
    return threadName;
}


DWORD WINAPI Thread::startThreadRunnable(LPVOID pVoid){
			Thread* runnableThread = static_cast<Thread*> (pVoid);
			if (runnableThread != NULL) {
			  {
				runnableThread->runnable->run();
				runnableThread->setCompleted();
			  }
			}
			return 0;
}	
 
DWORD WINAPI Thread::startThread(LPVOID pVoid) {
			Thread* aThread = static_cast<Thread*> (pVoid);
			if (aThread != NULL) {
			  {
				aThread->run();
				aThread->setCompleted();
			  }
			}
			return 0;
}

