
#include <iostream>
#include <windows.h>
#include "TDThread.h"
#include "monitorSU.h"
#include <string>
#include <stdlib.h>

class boundedBuffer : private monitorSU {
// An SC bounded buffer monitor
private:
	int fullSlots; 			// # of full slots in the buffer
	int capacity;
	int* buffer;
	int in, out;
	conditionVariable notFull;
	conditionVariable notEmpty;
public:
#pragma warning(disable:4355) 
    // using "this" here generates a warning in VC++
	boundedBuffer(int capacity_, std::string ID) : monitorSU("boundedBuffer" + ID), fullSlots(0), 
	  capacity(capacity_), in(0), out(0), notFull(this,"notFull"+ID), notEmpty(this,"notEmpty"+ID) { 	
		buffer = new int[capacity];	
	}

	~boundedBuffer() {
		delete [] buffer;
	}

	void deposit(int value) {
		enterMonitor("deposit");
		if (fullSlots == capacity) 
			notFull.waitC();
		buffer[in] = value;
		in = (in + 1) % capacity;
		++fullSlots;
		exerciseEvent("deposit");
		notEmpty.signalC_and_exitMonitor();
		//exitMonitor();
	}

	int withdraw() {
		enterMonitor("withdraw");
		int value;
		if (fullSlots == 0) 
			notEmpty.waitC();
		value = buffer[out];
      	out = (out + 1) % capacity;
		--fullSlots;
		exerciseEvent("withdraw");
		notFull.signalC_and_exitMonitor();
		//exitMonitor();
		return value;
	}
};


class Producer : public  TDThread {
private:
	boundedBuffer& b;
	std::string num;
public:
	Producer (boundedBuffer* b_, std::string num_) : TDThread("Producer"+num_), b(*b_),  num(num_) { }
	void* run () {
		std::cout << "Producer Running" << std::endl;
		for (int i = 0; i < 3; i++) {
			b.deposit(i);
			std::cout << "Producer # " << num << " deposited " << i << std::endl;
		}
	return 0;
	}
};

class Consumer : public TDThread {
private:
	boundedBuffer& b;
	std::string num;
public:
	Consumer (boundedBuffer* b_, std::string num_) : TDThread("Consumer"+num_), b(*b_), num(num_) { }
	void* run () {
		std::cout << "Consumer Running" << std::endl;
 		int value = 0;
		for (int i = 0; i < 3; i++) {
			value = b.withdraw();
			std::cout << "Consumer # " << num << " withdrew " << value << std::endl;
		}
		return 0;
	}
};

int main ( ) {
	const std::string buffer_1_ID = "1", buffer_2_ID = "2";
	const std::string producer_1_ID = "1", producer_2_ID = "2";
	const std::string consumer_1_ID = "1", consumer_2_ID = "2";
	boundedBuffer* b1 = new boundedBuffer(2,buffer_1_ID);
	Producer* p1 = new Producer (b1, producer_1_ID); Consumer* c1 = new Consumer (b1, consumer_1_ID);
	p1->start(); c1->start();
	c1->join(); p1->join(); 
	delete b1; 
	delete p1; delete c1; 
 	return 0;
}
