#pragma warning(disable:4786) 
#include <stddef.h>
#include "criticalsectionlock.h"
#include "mutexlocker.h"
#include <map>
#include <string>
#include <fstream>
#include "monitoridgenerator.h"
#include <stdlib.h>
#include <assert.h>
 
    criticalSectionLock MonitorIDGenerator::instanceLock;

	 MonitorIDGenerator& MonitorIDGenerator::instance() {
	  mutexLocker<criticalSectionLock> mutex(instanceLock); 
	  static MonitorIDGenerator instance;
	  return instance;
	 }

	MonitorIDGenerator::MonitorIDGenerator() {
		outputMonitorIDs.open("MonitorID.txt");
		if (outputMonitorIDs.fail()) {
			std::cout << "Error opening file MonitorID.txt for output." << std::endl;
			exit(1);
		}
	}

	bool MonitorIDGenerator::containsName(std::string stringID) {
		//mutexLocker mutex(dataLock);
	  std::map<std::string, int>::iterator pos;
	  pos = names.find(stringID);
	  if (pos != names.end()) 
	    return true;
	  else
		return false;
	}

	void MonitorIDGenerator::putName(std::string name, int num) {
	// associates name with num
		//mutexLocker mutex(dataLock);
		std::pair<std::map<std::string,int>::iterator, bool> p =
							  names.insert(std::make_pair(name,num));
		if (!p.second) // name already in table so change its value
		  p.first->second = num;

		char buffer[20];
		itoa(num,buffer,10);
		std::string numString(buffer); // (nextNum);
		outputMonitorIDs << (name+numString) << std::endl;
	}

	void MonitorIDGenerator::putUserName(std::string name, int num) {
	// associates name with num
		//mutexLocker mutex(dataLock);
		std::pair<std::map<std::string,int>::iterator, bool> p =
							  names.insert(std::make_pair(name,num));
		if (!p.second) // name already in table so change its value
		  p.first->second = num;
		outputMonitorIDs << name << std::endl;
	}

	int MonitorIDGenerator::getNum(std::string name) {
	// gets num associated with name
		//mutexLocker mutex(dataLock);
		std::map<std::string, int>::iterator pos;
	  pos = names.find(name);
		assert(pos != names.end());
	  return(pos->second);
	}


	std::string MonitorIDGenerator::getID(std::string stringID) {
		mutexLocker<win32Critical_Section> mutex(dataLock);
		// check to see if this is first instance of this thread class
		// id has two tables:
		//	1. instance table of (className, #instances)
		// 2. identifier table of (className+instance#,ID)
		std::string nextName;
		if (!containsName(stringID)){
				// first thread of this class
				nextName = stringID + "1"; // next className+instance# for identifier table
				putName(stringID,1); // add to instance table
			}
		else {
				// at least one instance already exists
			int nextNum = getNum(stringID)+1;

			char buffer[20];
			itoa(nextNum,buffer,10);
			std::string nextNumString(buffer); // (nextNum);

			nextName = stringID + nextNumString; // next className+instance# for identifier table
			putName(stringID,nextNum); // add to instance table
		}
      return nextName;
	}

	std::string MonitorIDGenerator::getIDFromUserName(std::string userName) {
		mutexLocker<win32Critical_Section> mutex(dataLock);
		if (!containsName(userName)){
				// first thread of this class
				putName(userName, 1); // add to instance table
			}
		else {
			// at least one instance already exists
			std::cout << "Use of duplicate monitor name: " << userName << std::endl;
			exit(1);
			//throw InvalidMonitorName("Use of duplicate thread name: " + userName);
		}
      return userName;
	}

