#pragma warning(disable:4786) 
#include <stddef.h>
#include "criticalSectionLock.h"
#include "mutexLocker.h"
#include <map>
#include <string>
#include <fstream>
#include "MutexIDGenerator.h"
#include <stdlib.h>
#include <assert.h>
 
    criticalSectionLock MutexIDGenerator::instanceLock;

	MutexIDGenerator& MutexIDGenerator::instance() {
	  mutexLocker<criticalSectionLock> mutex(instanceLock); 
	  static MutexIDGenerator instance;
	  return instance;
	 }

	 MutexIDGenerator::MutexIDGenerator() {
		outputMutexIDs.open("MutexID.txt");
		if (outputMutexIDs.fail()) {
			std::cout << "Error opening file MutexID.txt for output." << std::endl;
			exit(1);
		}
	}

	bool MutexIDGenerator::containsName(std::string stringID) {
		//mutexLocker mutex(dataLock);
	  std::map<std::string, int>::iterator pos;
	  pos = names.find(stringID);
	  if (pos != names.end()) 
	    return true;
	  else
		return false;
	}

	void MutexIDGenerator::putName(std::string name, int num) {
	// associates name with num
		std::pair<std::map<std::string,int>::iterator, bool> p =
							  names.insert(std::make_pair(name,num));
		if (!p.second) // name already in table so change its value
		  p.first->second = num;

		char buffer[20];
		itoa(num,buffer,10);
		std::string numString(buffer); // (nextNum);
		outputMutexIDs << (name+numString) << std::endl;
	}

	void MutexIDGenerator::putUserName(std::string name, int num) {
	// associates name with num
		std::pair<std::map<std::string,int>::iterator, bool> p =
							  names.insert(std::make_pair(name,num));
		if (!p.second) // name already in table so change its value
		  p.first->second = num;
		outputMutexIDs << name << std::endl;
	}

	int MutexIDGenerator::getNum(std::string name) {
	// gets num associated with name
		std::map<std::string, int>::iterator pos;
	  pos = names.find(name);
		assert(pos != names.end());
	  return(pos->second);
	}


	std::string MutexIDGenerator::getID(std::string stringID) {
		mutexLocker<win32Critical_Section> mutex(dataLock);
		// check to see if this is first instance of this thread class
		// id has two tables:
		//	1. instance table of (className, #instances)
		// 2. identifier table of (className+instance#,ID)
		std::string nextName;
		if (!containsName(stringID)){
				// first thread of this class
				nextName = stringID + "1"; // next className+instance# for identifier table
				putName(stringID,1); // add to instance table
			}
		else {
				// at least one instance already exists
			int nextNum = getNum(stringID)+1;

			char buffer[20];
			itoa(nextNum,buffer,10);
			std::string nextNumString(buffer); // (nextNum);

			nextName = stringID + nextNumString; // next className+instance# for identifier table
			putName(stringID,nextNum); // add to instance table
		}
      return nextName;
	}

	std::string MutexIDGenerator::getIDFromUserName(std::string userName) {
		mutexLocker<win32Critical_Section> mutex(dataLock);
		if (!containsName(userName)){
				// first thread of this class
				putName(userName, 1); // add to instance table
			}
		else {
			// at least one instance already exists
			std::cout << "Use of duplicate Mutex name: " << userName << std::endl;
			exit(1);
		}
      return userName;
	}

