#ifndef TDTHREAD_H
#define TDTHREAD_H
#pragma warning(disable:4786) 

#include "runnable.h"
#include <windows.h>
#include <iostream>
#include <string>
#include <memory>
// Chapter 2 version of TDThread is simpler and has some functions disabled
//#include "vectorTimeStamp.h"
//#include "threadBasedMsgTracingAndReplay.h"
#include "criticalSectionLock.h"
//#include "Controller.h"

				
class TDThread {
public:
	TDThread(std::auto_ptr<Runnable> runnable_, const std::string tName = "");
	TDThread(const std::string tName = "");
    	virtual ~TDThread();
    	void start();	// threads are created in suspended state; this starts them
 	void* join();
	int getID();
	std::string getThreadName();
	static void incNumRunningThreads();
	static void decNumRunningThreads();
	static void incNumBlockedThreads();
	static void decNumBlockedThreads();
	static int getNumRunningThreads();
	static int getNumBlockedThreads();
	static void resetNumBlockedRunningThreads();
	static void setNumBlockedThreads(int value);
private:
	HANDLE hThread;
	unsigned int winThreadID; // Win32 thread ID
	int	ID; // TDThread ID
	std::auto_ptr<Runnable> runnable;
	TDThread(const TDThread&);
	const TDThread& operator= (const TDThread&);
	HANDLE completionEvent;
	void setCompleted();
	void* result;
	std::string threadName;
	int generateID();
	int generateIDFromUserName(std::string userName);
	static int mode;
	static int raceAnalysis;
	static int numControllers;
	static int strategy;
	static bool randomDelay;
	static bool detectDeadlock;
	//int versionNumber;			// now stored in threadData 
	//vectorTimeStamp vectorTS;
	std::string traceFileName;
	//static threadBasedMsgTracingAndReplay* m; // used for singleton controller
	static criticalSectionLock instanceLock;
	//Controller* control;
	//threadBasedMsgTracingAndReplay* TDThread::instance(int mode, 
	//	int numControllers, int strategy, std::string traceFileName);
	//Controller* getController();
	//void constructController();
	//void initVectorTS();	// not used; init done in threadData	
	static unsigned int WINAPI startThreadRunnable(LPVOID pVoid);
	virtual void* run() {return 0;}
	static unsigned int WINAPI startThread(LPVOID pVoid);
	void PrintError(LPTSTR lpszFunction,LPSTR fileName, int lineNumber);
	static int numRunningThreads;
	static int numBlockedThreads;
    };
#endif
