#pragma warning(disable:4786) 
#include "thread.h"
#include "runnable.h"
#include <assert.h>
#include <windows.h>
#include <process.h>
#include <memory>

Thread::Thread(std::auto_ptr<Runnable> runnable_) : runnable(runnable_) {
	if (runnable.get() == NULL)
		PrintError("Thread(std::auto_ptr<Runnable> runnable_) failed at ",__FILE__,__LINE__);
	completionEvent = CreateEvent( 
     NULL,   		// no security attributes
     1,  	 		// manual reset event
     0,   			// initially nonsignaled
     NULL);  		// unnamed event

	hThread = (HANDLE)_beginthreadex(NULL,0, Thread::startThreadRunnable,
		(LPVOID)this, CREATE_SUSPENDED, &winThreadID );
	if (!hThread)
		PrintError("_beginthreadex failed at ",__FILE__,__LINE__);
}

Thread::Thread() : runnable(NULL) {

	completionEvent = CreateEvent( 
     NULL,   		// no security attributes
     1,  	 		// manual reset event
     0,   	        // initially nonsignaled
     NULL);  		// unnamed semaphore
	  
	hThread = (HANDLE)_beginthreadex(NULL,0,Thread::startThread,
		(LPVOID)this, CREATE_SUSPENDED, &winThreadID );
	if (!hThread)
		PrintError("_beginthreadex failed at ",__FILE__,__LINE__);						 	 
}

//Thread::Thread(const Thread& aThread){}

//const Thread& Thread::operator= (const Thread& aThread) {return *this;}

Thread::~Thread() {
	DWORD rc = CloseHandle(completionEvent);
	if (!rc)
		PrintError("CloseHandle failed at ",__FILE__,__LINE__);
	if (winThreadID != GetCurrentThreadId()) {
      rc = CloseHandle(hThread);
  	  if (!rc)
		PrintError("CloseHandle failed at ",__FILE__,__LINE__);
	}
}
 
void Thread::start() {
	assert(hThread != NULL); 
	DWORD rc = ResumeThread(hThread);
 	if (!rc)
		PrintError("ResumeThread failed at ",__FILE__,__LINE__);
}

void* Thread::join() {
	DWORD rc = WaitForSingleObject( 
        	completionEvent,		// handle to event
        	INFINITE);              // no timeout
 	if (!(rc==WAIT_OBJECT_0))
		PrintError("WaitForSingleObject failed at ",__FILE__,__LINE__);
	return result;
}

void Thread::setCompleted() {
	DWORD rc = SetEvent(completionEvent);
 	if (!rc)
		PrintError("SetEvent failed at ",__FILE__,__LINE__);
	//completed = true;
}

unsigned WINAPI Thread::startThreadRunnable(LPVOID pVoid){
	Thread* runnableThread = static_cast<Thread*> (pVoid);
	assert(runnableThread);
	  {
		runnableThread->result = runnableThread->runnable->run();
		runnableThread->setCompleted();
	  }
	return reinterpret_cast<unsigned>(runnableThread->result);
}	
 
unsigned WINAPI Thread::startThread(LPVOID pVoid) {
	Thread* aThread = static_cast<Thread*> (pVoid);
	assert(aThread);
	  {
		aThread->result = aThread->run();
		aThread->setCompleted();
	  }
	return reinterpret_cast<unsigned>(aThread->result);
}

void Thread::PrintError(LPTSTR lpszFunction,LPSTR fileName, int lineNumber)  { 
	TCHAR szBuf[256]; 
	LPSTR lpErrorBuf;
	DWORD errorCode = GetLastError(); 
	FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | 
	     FORMAT_MESSAGE_FROM_SYSTEM, NULL,  errorCode, MAKELANGID(LANG_NEUTRAL, 		     SUBLANG_DEFAULT),  (LPTSTR) &lpErrorBuf, 0, NULL );
	wsprintf(szBuf, "%s failed at line %d in %s with error %d: %s", lpszFunction, lineNumber, fileName, 
  	     errorCode, lpErrorBuf); 
 	DWORD numWritten;
	WriteFile(GetStdHandle(STD_ERROR_HANDLE),szBuf,strlen(szBuf),&numWritten,FALSE);
	LocalFree(lpErrorBuf);
	exit(errorCode); 
}

