#pragma warning(disable:4786) 
#include <stddef.h>
#include "criticalSectionLock.h"
#include "mutexLocker.h"
#include <map>
#include <string>
#include <fstream>
#include "threadIDGenerator.h"
#include <stdlib.h>
#include "traceOrReplay.h"
#include <assert.h>

    criticalSectionLock ThreadIDGenerator::instanceLock;

	 ThreadIDGenerator& ThreadIDGenerator::instance() {
	  mutexLocker<criticalSectionLock> mutex(instanceLock); 
	  static ThreadIDGenerator instance;
	  return instance;
	 }

	ThreadIDGenerator::ThreadIDGenerator() : numberOfThreads(0), saveNumberOfThreads(0), firstRT(true) {
		char* env = getenv("MODE");
		if (env!=NULL) {
			if (!(strcmp(env,"TRACE")))
				mode = TRACE;
			else if (!(strcmp(env,"REPLAY")))
				mode = REPLAY;
			else if (!(strcmp(env,"TEST")))
				mode = TEST;
			else if (!(strcmp(env,"SPECTEST")) || !(strcmp(env,"COMMTEST")))
				mode = SPECTEST;
			else if (!(strcmp(env,"RT"))) {
				std::cout << "We are currently updating our C++ RT implementation." << std::endl;
				std::cout << "We will post new files shortly and enable RT." << std::endl;
				exit(1);
			}
			else if (!(strcmp(env,"NONE")))
				mode = NONE;
			else mode = NONE;
		}
	    else mode = NONE;
		env = getenv("RACEANALYSIS");
		if (env!=NULL) {
			if (!(strcmp(env,"ON")))
				raceAnalysis = true;
			else raceAnalysis = false;
		}
		else raceAnalysis = false;
	    if (mode == TRACE || mode == RT) {
			outputThreadIDsText.open("ThreadID.txt");
			if (outputThreadIDsText.fail()) {
				std::cout << "Error opening file ThreadID.txt for output." << std::endl;
				exit(1);
			}
		}
		else if (mode == REPLAY || mode == TEST || mode == SPECTEST) {
				inputThreadNameAndIDs.open("ThreadID.txt");
				if (inputThreadNameAndIDs.fail()) {
					std::cout << "Error opening file ThreadID.txt for input." << std::endl;
					exit(1);
      			}
				std::string name; int ID;
				while (inputThreadNameAndIDs >> name >> ID) { // assume unique names and IDS
					numberOfThreads = ID; // last value of numberOfThreads will be used
					std::pair<std::map<std::string,int>::iterator, bool> p1 = 
					     replayIDsByName.insert(std::make_pair(name,ID));
					assert(p1.second);
					std::pair<std::map<int,std::string>::iterator, bool> p2 = 
					     replayIDsByID.insert(std::make_pair(ID,name));
					assert(p2.second);
				}
		}			
	}

    void ThreadIDGenerator::resetIDs() {
		mutexLocker<win32Critical_Section> mutex(dataLock);
   		saveNumberOfThreads = numberOfThreads;
   		numberOfThreads = 0;
   		firstRT = false;
		names.clear();
		IDs.clear();
		replayIDsByName.clear();
		replayIDsByID.clear();
   }

	int ThreadIDGenerator::getNextID() {
	  return ++numberOfThreads;
	}

	bool ThreadIDGenerator::containsName(std::string stringID) {
	  std::map<std::string, int>::iterator pos;
	  pos = names.find(stringID);
	  if (pos != names.end()) 
	    return true;
	  else
		  return false;
	}

	void ThreadIDGenerator::putName(std::string name, int num) {
	// associates name with num
		std::pair<std::map<std::string,int>::iterator, bool> p =
							  names.insert(std::make_pair(name,num));
		if (!p.second) // name already in table so change its value
		  p.first->second = num;
	}

	int ThreadIDGenerator::getNum(std::string name) {
	// gets num associated with name
		std::map<std::string, int>::iterator pos;
	  pos = names.find(name);
		assert(pos != names.end());
	  return(pos->second);
	}

	int ThreadIDGenerator::generateID(std::string name) {
	// associates name with num
		int threadID=0;
		if (mode == TRACE || mode == RT || (mode==NONE && raceAnalysis)) { 
			threadID = getNextID();
			std::pair<std::map<int,std::string>::iterator, bool> p =
			    IDs.insert(std::make_pair(threadID,name));
			assert(p.second);
			if (firstRT)
				outputThreadIDsText << name << " " << threadID << std::endl;
		}
		else if (mode == REPLAY || mode == TEST || mode == SPECTEST) {
 	 		std::map<std::string, int>::iterator pos;
			pos = replayIDsByName.find(name);
			assert(pos != replayIDsByName.end());
			threadID = pos->second;
		}
		return threadID;
	}

	std::string ThreadIDGenerator::getName(int ID) {
		mutexLocker<win32Critical_Section> mutex(dataLock);
		std::map<int,std::string>::iterator pos;
		if (mode == TRACE || mode == RT || (mode==NONE && raceAnalysis)) {
		  pos = IDs.find(ID);
		  assert(pos != IDs.end());
			return (pos->second);	
		}
		else if (mode == REPLAY || mode == TEST || mode == SPECTEST) {
			pos = replayIDsByID.find(ID);
		  assert(pos != replayIDsByID.end());
			return (pos->second);	 	 
		}
		assert(false);
		return ""; // never called unless one of above flags is true
	}

	int ThreadIDGenerator::getID(std::string stringID) {
		mutexLocker<win32Critical_Section> mutex(dataLock);
		// check to see if this is first instance of this thread class
		// id has two tables:
		//	1. instance table of (className, #instances)
		// 2. identifier table of (className+instance#,ID)
		std::string nextName;
		if (!containsName(stringID)){
				// first thread of this class
				nextName = stringID + "1"; // next className+instance# for identifier table
				putName(stringID,1); // add to instance table
			}
		else {
				// at least one instance already exists
			int nextNum = getNum(stringID)+1;

      char buffer[33];
			itoa(nextNum,buffer,10);
			std::string nextNumString(buffer); // (nextNum);

			nextName = stringID + nextNumString; // next className+instance# for identifier table
			putName(stringID,nextNum); // add to instance table
		}
      return generateID(nextName);
	}

	int ThreadIDGenerator::getIDFromUserName(std::string userName) {
		mutexLocker<win32Critical_Section> mutex(dataLock);
		if (!containsName(userName)){
				// first thread of this class
				putName(userName, 1); // add to instance table
			}
		else {
			// at least one instance already exists
			std::cout << "Use of duplicate thread name: " << userName << std::endl;
			exit(1);
		}
      return generateID(userName);
	}

	// if tracing, returns number of IDs that have been created; 
	// if replay or test, returns max ID read from ThreadID file.
	int ThreadIDGenerator::getNumberOfThreads() {
		mutexLocker<win32Critical_Section> mutex(dataLock);
		if( mode == RT) 
			return saveNumberOfThreads;
		else
			return numberOfThreads;
	}
