#ifndef THREADIDGENERATOR_H
#define THREADIDGENERATOR_H

#pragma warning(disable:4786) 
#include <stddef.h>
#include "criticalSectionLock.h"
#include "win32Critical_Section.h"
#include "mutexLocker.h"
#include <map>
#include <string>
#include <fstream>
#include <iostream>

class InvalidThreadName { 
public:
	InvalidThreadName() { }
	InvalidThreadName(std::string msg_): msg(msg_) { } 
private:
  std::string msg; 
};

class ThreadIDGenerator {
private:
	std::map<std::string,int> names;
	std::map<int,std::string> IDs;
	// when replay or test, use IDs generated during trace (not new IDs).
	std::map<std::string,int> replayIDsByName; // keyed on Name
	std::map<int,std::string> replayIDsByID; // keyed on ID
	static criticalSectionLock instanceLock;
	win32Critical_Section dataLock; 
	std::ofstream outputThreadIDsText;
	int numberOfThreads; // number of threads in ThreadID file: 1..numberOfThreads
	std::ifstream inputThreadNameAndIDs; // input sequence of int IDs
	int mode;
	bool raceAnalysis;
	bool firstRT;
	int saveNumberOfThreads;
	ThreadIDGenerator();
	//TBD
	//~ThreadIDGenerator(); // close files?
	//TBD
	void resetIDs();
	int getNextID();
	bool containsName(std::string stringID);
	void putName(std::string name, int num);
	int getNum(std::string name);
	int generateID(std::string name);
public:
	std::string getName(int ID);
	static ThreadIDGenerator& instance();
	int getID(std::string stringID);
	int getIDFromUserName(std::string userName);
	// if tracing, returns number of IDs that have been created; 
	// if replay or test, returns max ID read from ThreadID file.
	int getNumberOfThreads();
};
#endif
