#include "Timestamp.h"
#include <string>

Timestamp::Timestamp (int dimension) {
	this->dimension = dimension;
	values = new int [dimension];
}
Timestamp::Timestamp(const Timestamp& b) {copy(b);}  // copy constructor
Timestamp::~Timestamp() { free(); }   // destructor
const Timestamp& Timestamp::operator=(const Timestamp& b) {  // copy assignment
  if (this != &b) {
    free();
    copy(b);
  }
  return *this;
}
void Timestamp::copy(const Timestamp& b) {
	dimension = b.dimension;
	values = new int [dimension];
	for (int i = 0; i < dimension; i ++) {
		values[i] = b.getValue(i);
	}
}
void Timestamp::free() { delete [] values;}

int Timestamp::getDimension () const{
	return dimension;
}

int Timestamp::getValue (int index) const{
	return values[index];
}

void Timestamp::setValue (int index, int value) {
	values[index] = value;
}

// update methods
void Timestamp::advance (int pid) {
	values[pid]++;
}
 
void Timestamp::advance(Timestamp& stamp) {
	for(int i = 0; i < dimension; i ++) {
		if(values[i] < stamp.getValue(i)) {
			values[i] = stamp.getValue (i);
		}
	}
}

bool Timestamp::equalTo(Timestamp& stamp) const {
	bool rval = true;
	if(dimension != stamp.getDimension()) {
		std::cout << "The two timestamps are not comparable." << std::endl;
	   rval = false;
	}
	else {
	    for (int i = 0; i < dimension; i ++) {
			if (values[i] != stamp.getValue(i)) {
				rval = false;
				break;
			}
	    }
	}
	return rval;
}
    
bool Timestamp::lessThan (Timestamp& ts) const {
	bool rval = true;
	Timestamp stamp = (Timestamp) ts;
	if(dimension != stamp.getDimension ()) {
		std::cout << "The two timestamps are not comparable." << std::endl;
		rval = false;
	}
	else if (equalTo (stamp)) {
	    rval = false;
	}
	else {
	    for (int i = 0; i < dimension; i ++) {
			if (values[i] > stamp.getValue(i)) {
			   rval = false;
			  break;
			}
	    }
	}
	return rval;
} 

Timestamp* Timestamp::deepCopy () {
	Timestamp* rval = new Timestamp(dimension);
	for (int i = 0; i < dimension; i ++) {
		rval->setValue(i, getValue(i));
	}
	return rval;
}


std::ostream& operator<<(std::ostream& os, const Timestamp& t) {
	os << t.toString();
	return os;
}

std::string Timestamp::toString() const {
	std::string rval;
	rval += "[";
	for(int i = 0; i < dimension; i ++) {
		if(i > 0) {
			rval += ", ";
		}
		char buffer[33];
		itoa(values[i],buffer,10);
		rval += buffer;
	}
	rval += "]";
	return rval;
}