#pragma warning(disable:4786) 
#include <stddef.h>
#include "criticalsectionlock.h"
#include "mutexlocker.h"
#include <map>
#include <string>
#include <fstream>
#include "conditionvaridgenerator.h"
#include <stdlib.h>
#include <assert.h>
 
    criticalSectionLock ConditionVarIDGenerator::instanceLock;

	 ConditionVarIDGenerator& ConditionVarIDGenerator::instance() {
	  mutexLocker<criticalSectionLock> mutex(instanceLock); 
	  static ConditionVarIDGenerator instance;
	  return instance;
	 }

	ConditionVarIDGenerator::ConditionVarIDGenerator() {
		outputConditionIDs.open("ConditionID.txt");
		if (outputConditionIDs.fail()) {
			std::cout << "Error opening file ConditionID.txt for output." << std::endl;
			exit(1);
		}
	}

	bool ConditionVarIDGenerator::containsName(std::string stringID) {
		//mutexLocker mutex(dataLock);
	  std::map<std::string, int>::iterator pos;
	  pos = names.find(stringID);
	  if (pos != names.end()) 
	    return true;
	  else
		return false;
	}

	void ConditionVarIDGenerator::putName(std::string name, int num) {
	// associates name with num
		//mutexLocker mutex(dataLock);
		std::pair<std::map<std::string,int>::iterator, bool> p =
							  names.insert(std::make_pair(name,num));
		if (!p.second) // name already in table so change its value
		  p.first->second = num;

		char buffer[20];
		itoa(num,buffer,10);
		std::string numString(buffer); // (nextNum);
		outputConditionIDs << (name+numString) << std::endl;
	}

	void ConditionVarIDGenerator::putUserName(std::string name, int num) {
	// associates name with num
		//mutexLocker mutex(dataLock);
		std::pair<std::map<std::string,int>::iterator, bool> p =
							  names.insert(std::make_pair(name,num));
		if (!p.second) // name already in table so change its value
		  p.first->second = num;
		outputConditionIDs << name << std::endl;
	}

	int ConditionVarIDGenerator::getNum(std::string name) {
	// gets num associated with name
		//mutexLocker mutex(dataLock);
		std::map<std::string, int>::iterator pos;
	  pos = names.find(name);
		assert(pos != names.end());
	  return(pos->second);
	}


	std::string ConditionVarIDGenerator::getID(std::string stringID) {
		mutexLocker<win32Critical_Section> mutex(dataLock);
		// check to see if this is first instance of this thread class
		// id has two tables:
		//	1. instance table of (className, #instances)
		// 2. identifier table of (className+instance#,ID)
		std::string nextName;
		if (!containsName(stringID)){
				// first thread of this class
				nextName = stringID + "1"; // next className+instance# for identifier table
				putName(stringID,1); // add to instance table
			}
		else {
				// at least one instance already exists
			int nextNum = getNum(stringID)+1;

			char buffer[20];
			itoa(nextNum,buffer,10);
			std::string nextNumString(buffer); // (nextNum);

			nextName = stringID + nextNumString; // next className+instance# for identifier table
			putName(stringID,nextNum); // add to instance table
		}
      return nextName;
	}

	std::string ConditionVarIDGenerator::getIDFromUserName(std::string userName) {
		mutexLocker<win32Critical_Section> mutex(dataLock);
		if (!containsName(userName)){
				// first thread of this class
				putUserName(userName, 1); // add to instance table
			}
		else {
			// at least one instance already exists
			std::cout << "Use of duplicate condition variable name: " << userName << std::endl;
			exit(1);
			//throw InvalidConditionName("Use of duplicate thread name: " + userName);
		}
      return userName;
	}

