#ifndef MONITORSU_H
#define MONITORSU_H

#pragma warning(disable:4786) 
#include "binarySemaphore.h"
#include "countingSemaphore.h"
#include "monitorTracingReplayTest.h"
#include "monitorEvent.h"
#include "conditionVarIDGenerator.h"
#include "threadData.h"
#include <string>
#include "vectorTimeStamp.h"
#include "OPenEventX.h"
#include "Control.h"
#include "msgTracingAndReplay.h"
#include "TDThread.h"
#include "deadlockWatchdog.h"

class monitorSU {
// Monitor Toolbox that uses VP() operation and SU signaling

protected:
   	monitorSU();
	monitorSU(std::string monitorName_);
	monitorSU(int mode);
	~monitorSU();
	void enterMonitor(std::string methodName);
	void enterMonitor();
	void exitMonitor();
	void exerciseEvent(std::string eventName);
	
private:
	binarySemaphore mutex;
	binarySemaphore reentry;
	int reentryCount;
	monitorTracingReplayTest* m;


	void requestMPermit(monitorEvent::eventTypes op, int ID, std::string methodName, 
		std::string conditionName);
	void releaseMPermit();
	void requestEntryPermit(int ID);
	void releaseEntryPermit();
	void trace(monitorEvent::eventTypes op, int ID, std::string methodName, 
		                         std::string conditionName);
	void requestMPermitSpecTest(int ID);
	void traceCommEvent(int ID, std::string eventName);

	std::string generateID();
	std::string generateIDFromUserName(std::string userName);
	int generateIDFromUserNameRT(std::string userName);
	int generateIDRT();
	friend class conditionVariable;
	int mode;
	int numControllers;
/* strategy */
	int strategy;
	bool randomDelay;
	bool detectDeadlock;
	std::string monitorName;
	std::string currentMethodName;

	std::vector<OpenEventX> openList; // monitor name (representing always open methods)
	vectorTimeStamp vectorTS;
	int versionNumber;
	int getVersionNumber() { return versionNumber; }
	int getAndIncVersionNumber() {return versionNumber++;}

	Control* control;
		   // Returns the integer clock value associated with this thread.
	int getIntegerTS();
	// Increments the integer clock associated with this thread.
    void updateIntegerTS(); 

	// Updates the integer clock with the value passed in.
	void setIntegerTS(int ts);
	void mergeIntegerTS(int ts);
	// Returns a clone of the vector timestamp associated with this thread
	vectorTimeStamp& getVectorTS(); 
	// Updates vector timestamp with the passed in vector timestamp
	void updateVectorTS(const vectorTimeStamp& newTS);
	void initVectorTS();
	Control* instanceRT(int mode, int numControllers, int strategy, std::string traceFileName);
	int ID;
	int getID() {return ID;}
};

class conditionVariable {
private:
	countingSemaphore threadQueue;
	int numWaitingThreads;
	monitorSU& m;
	int mode;
	bool randomDelay;
	bool detectDeadlock;
	std::string conditionName;
	//std::string generateIDFromUserName(std::string conditionName);
	std::string conditionVariable::generateIDFromUserName(std::string conditionName) {
		//	System.out.println("User Name: " + userName);
		return (ConditionVarIDGenerator::instance().getIDFromUserName(conditionName));
	}
	//std::string generateID();
	std::string conditionVariable::generateID() {
		const std::string thisClassName = "conditionVariable";
		std::string threadName;
		//thisClassName = typeid(*this).name(); // will return "Thread"
		data d = threadData::instance().getThreadData(GetCurrentThreadId());	// return child or main
		//if (d.threadID != -1) 
			threadName = d.threadName;
		//else
		//	threadName = "";
		//int position = threadName.find("_"); 
		//if (position > -1 && (size_t) position < threadName.length())
			// default for main is "main" so just use it
			threadName = threadName + "_" + thisClassName;
		//else	
		//	threadName = "main_" + thisClassName;
		
		std::string myID = ConditionVarIDGenerator::instance().getID(threadName);
		return myID;
	}
public:
	//conditionVariable(monitorSC* mon); 
	//conditionVariable(monitorSC* mon, std::string conditionName);
	conditionVariable::conditionVariable(monitorSU* mon) :threadQueue(0,false), m(*mon), 
		numWaitingThreads(0) {
		char* env = getenv("MODE");
		if (env!=NULL) {
			if (!(strcmp(env,"TRACE")))
				mode = TRACE;
			else if (!(strcmp(env,"REPLAY")))
				mode = REPLAY;
			else if (!(strcmp(env,"TEST")))
				mode = TEST;
			else if (!(strcmp(env,"SPECTEST")) || !(strcmp(env,"COMMTEST")))
				mode = SPECTEST;
			else if (!(strcmp(env,"RT")))
				mode = RT;
			else if (!(strcmp(env,"NONE")))
				mode = NONE;
			else mode = NONE;
		}
		else mode = NONE;
		env = getenv("RANDOMDELAY");
		if (env!=NULL) {	
		if	(!(strcmp(env,"ON")))
			randomDelay = true;
		else
			randomDelay = false;
		}
		else randomDelay = false;

		env = getenv("DEADLOCKDETECTION");
		if (env!=NULL) {
			if (!(strcmp(env,"ON")))
				detectDeadlock = true;
			else
				detectDeadlock = false;
		}
		else detectDeadlock = false;
		if (!(mode==NONE))
			conditionName = generateID();
	}

	conditionVariable::conditionVariable(monitorSU* mon, std::string conditionName_) 
		: threadQueue(0,false), m(*mon), numWaitingThreads(0), conditionName(conditionName_) {
		char* env = getenv("mode");
		if (env!=NULL) {
			if (!(strcmp(env,"TRACE")))
				mode = TRACE;
			else if (!(strcmp(env,"REPLAY")))
				mode = REPLAY;
			else if (!(strcmp(env,"TEST")))
				mode = TEST;
			else if (!(strcmp(env,"SPECTEST")) || !(strcmp(env,"COMMTEST")))
				mode = SPECTEST;
			else if (!(strcmp(env,"RT")))
				mode = RT;
			else if (!(strcmp(env,"NONE")))
				mode = NONE;
			else mode = NONE;
		}
		else mode = NONE;

		env = getenv("RANDOMDELAY");
		if (env!=NULL) {	
			if	(!(strcmp(env,"ON")))
				randomDelay = true;
			else
				randomDelay = false;
		}
		else randomDelay = false;

		env = getenv("DEADLOCKDETECTION");
		if (env!=NULL) {
			if (!(strcmp(env,"ON")))
				detectDeadlock = true;
			else
				detectDeadlock = false;
		}
		else detectDeadlock = false;

		if (!(mode==NONE))
			conditionName = generateIDFromUserName(conditionName);
	}

	void conditionVariable::signalC() {
		std::string saveMethodName = m.currentMethodName;
		int ID=0;
		if (mode == TRACE || mode == TEST || mode == RT) {
			ID = (threadData::instance().getThreadData(GetCurrentThreadId())).threadID;
		}
	
		if (mode == TRACE) {
			m.trace(monitorEvent::SIGNAL,ID,m.currentMethodName,conditionName); 
		}
		else if (mode == TEST) {
			m.requestMPermit(monitorEvent::SIGNAL,ID,m.currentMethodName,conditionName);	
		}
		if (numWaitingThreads > 0) {
			++(m.reentryCount);
			// update monitor timestamp on exit in case send/receive done inside monitor
			if (mode == RT) {
   				m.updateVectorTS((threadData::instance().getThreadData(GetCurrentThreadId())).vectorTS);
			}
			m.reentry.VP(&threadQueue);
			--(m.reentryCount);
			if (detectDeadlock && (mode == RT || mode == TRACE)) {
				TDThread::decNumBlockedThreads();
			}
   		}
		if (mode == TRACE) {
			m.currentMethodName = saveMethodName;		// restore name
		}
		else if (mode == TEST) {
			m.currentMethodName = saveMethodName;
		}
 		if (mode == RT) { 
			m.currentMethodName = saveMethodName;		// restore name
	 		// update vector timestamp since other 
			// threads may have entered while this thread was waiting
   			threadData::instance().getThreadData(GetCurrentThreadId()).vectorTS.updateVectorTS(m.getVectorTS());
		}
	}
	void signalC_and_exitMonitor() {
		std::string saveMethodName = m.currentMethodName;
		int ID=0;
		if (mode == TRACE || mode == TEST || mode == RT) {
			ID = (threadData::instance().getThreadData(GetCurrentThreadId())).threadID;
		}
		if (mode == TRACE) {
			m.trace(monitorEvent::SIGNALANDEXIT,ID,m.currentMethodName,conditionName); 
		}
		else if (mode == TEST) {
			m.requestMPermit(monitorEvent::SIGNALANDEXIT,ID,m.currentMethodName,conditionName);	
		}
		else if (mode == RT) {
			// update monitor timestamp on exit in case send/receive done inside monitor
			m.updateVectorTS((threadData::instance().getThreadData(GetCurrentThreadId())).vectorTS);

		}
		if (numWaitingThreads > 0) {
			threadQueue.V();
			if (detectDeadlock && (mode == RT || mode == TRACE)) {
				TDThread::decNumBlockedThreads();
			}
		}
		else if (m.reentryCount > 0)
    		m.reentry.V();
 		else
    		m.mutex.V();
	}
	
	void conditionVariable::waitC() {
		std::string saveMethodName = m.currentMethodName;
	    int callerForReceiveEvent = -1;
		int ID=0;
		if (mode == TRACE || mode == REPLAY || mode == TEST || mode == SPECTEST || mode == RT) {
			ID = (threadData::instance().getThreadData(GetCurrentThreadId())).threadID;
		}
		//std::cout << ID << "saved: " << saveMethodName << std::endl;
		if (mode == TRACE) {
			m.trace(monitorEvent::WAIT,ID,m.currentMethodName,conditionName); 
		}
		else if (mode == TEST) {
			m.requestMPermit(monitorEvent::WAIT,ID,m.currentMethodName,conditionName);	
		}
  		else if (mode == RT) {
			// update monitor timestamp on exit in case send/receive done inside monitor
   			m.updateVectorTS((threadData::instance().getThreadData(GetCurrentThreadId())).vectorTS);

		}
		if (detectDeadlock && (mode == RT || mode == TRACE)) {
			TDThread::incNumBlockedThreads();
			deadlockWatchdog::changeStatus(GetCurrentThreadId(),"blocked on waitC() in method " + saveMethodName);
		}
 		numWaitingThreads++;
		if (m.reentryCount > 0)
			threadQueue.VP(&(m.reentry));
		else
			threadQueue.VP(&(m.mutex));
		--numWaitingThreads;

		if (detectDeadlock && (mode == RT || mode == TRACE)) {
			deadlockWatchdog::changeStatus(GetCurrentThreadId(),"running");
		}

	
		if (mode == TRACE) {
			m.currentMethodName = saveMethodName;		// restore name;
		}
		else if (mode == TEST) {
			m.currentMethodName = saveMethodName;
		}
		else if (mode == SPECTEST) {
		
		}
	   	else if (mode == RT) {
			m.currentMethodName = saveMethodName;	     		
   			threadData::instance().getThreadData(GetCurrentThreadId()).vectorTS.updateVectorTS(m.getVectorTS());  
		}
	}	

	bool conditionVariable::empty() {
		return (numWaitingThreads == 0);
	}

	int conditionVariable::length() {
		return numWaitingThreads;
	}
};

#endif