#include "monitorTracingReplayTest.h"
#include "traceorreplay.h"
#include "criticalsectionlock.h"
#include "mutexLocker.h"
#include "monitorevent.h"
#include "process.h"
#include <fstream>

criticalSectionLock monitorTracingReplayTest::instanceLock;

std::string eventTypes[] = {"ENTRY","EXIT","WAIT","SIGNAL","SIGNALALL", "REENTRY", "SIGNALANDEXIT", "COMM"};

typedef unsigned (WINAPI *PBEGINTHREADEX_THREADFUNC) (
    LPVOID lpThreadParameter
		);
typedef unsigned *PBEGINTHREADEX_THREADID;

DWORD WINAPI monTestWatchDog(LPVOID pVoid) {
	monitorTracingReplayTest* controller = static_cast<monitorTracingReplayTest*> (pVoid);
		while (controller->index < (int) controller->Msequence.size()) {
			int saveIndex = controller->index;
			Sleep(2000);
			if (saveIndex == controller->index) {
				monitorEvent infeasibleEvent = 
						(monitorEvent) (controller->Msequence[controller->index]);
				std::cout << "Infeasible Sequence - timeout waiting for event " 
					<< (controller->index+1) << ": (" << eventTypes[infeasibleEvent.getEventType()]
					<< "," << infeasibleEvent.getThreadID() 
					<< "," << infeasibleEvent.getMethodName() 
					<< "," << infeasibleEvent.getConditionName() << ")" << std::endl;
				ExitProcess(1);
			}
		}
		return 0;
}

DWORD WINAPI monSpecTestWatchDog(LPVOID pVoid) {
	monitorTracingReplayTest* controller = static_cast<monitorTracingReplayTest*> (pVoid);
		while (controller->index < (int) controller->COMMsequence.size()) {
			int saveIndex = controller->index;
			Sleep(2000);
			if (saveIndex == controller->index) {
				monitorSpecEvent infeasibleEvent = 
						(monitorSpecEvent) (controller->COMMsequence[controller->index]);
				std::cout << "Infeasible Sequence - timeout waiting for event " 
					<< (controller->index+1) << ": (" << infeasibleEvent.getThreadID() 
					<< "," << infeasibleEvent.geteventName() << ")" << std::endl;
				ExitProcess(1);
			}
		}
		return 0;
}

monitorTracingReplayTest* monitorTracingReplayTest::instance(int mode, 
			int numControllers, std::string traceFileName) {
	mutexLocker<criticalSectionLock> mutex(instanceLock);  
	static monitorTracingReplayTest instance(mode,numControllers,traceFileName);
	return &instance;
}

monitorTracingReplayTest::monitorTracingReplayTest(int mode_, int numControllers_,
	std::string traceFileName_) : mode(mode_), numControllers(numControllers_),
		traceFileName(traceFileName_)  {
	 index = 0;

	 if (mode == TRACE) {
	   outputReplayText.open((traceFileName+"-replay.txt").c_str());
	   if (outputReplayText.fail()) {
		   std::cout << "Error opening traceFile for tracing simple M-sequence" << std::endl;
		 exit(1);
	   }
	   outputTestText.open((traceFileName+"-test.txt").c_str());
	   if (outputTestText.fail()) {
		   std::cout << "Error opening traceFile for tracing simple M-sequence" << std::endl;
		 exit(1);
	   }
	   outputSpecTestText.open((traceFileName+"-spectest.txt").c_str());
	   if (outputTestText.fail()) {
		   std::cout << "Error opening traceFile for tracing Comm-sequence" << std::endl;
		 exit(1);
	   }
	 }
	 else if (mode == REPLAY) {
	  inputReplayText.open((traceFileName+"-replay.txt").c_str());
	  if (inputReplayText.fail()) {
		  std::cout << "Error opening traceFile for replay" << std::endl;
		exit(1);
	  }	
	  int ID;
	  int max = 1;  // if sequence is empty max is 1

	  while (inputReplayText >> ID) {
	    SMsequence.push_back(ID);
		if (ID>max) max = ID;
	  }
	  //std::cout << "SM-sequence size is " << SMsequence.size() << std::endl;
	  numberOfThreads = max;
	  inputReplayText.close();
	  threads = new controlConditionVariable*[numberOfThreads+1];
	  //std::cout << "making " << (numberOfThreads+1) << std::endl;
	  for (int i=0; i<(numberOfThreads+2); i++)
			threads[i] = new controlConditionVariable(this);
	 }
	 else if (mode == TEST) {
	  inputTestText.open((traceFileName+"-test.txt").c_str());
	  if (inputTestText.fail()) {
		  std::cout << "Error opening traceFile for testing" << std::endl;
		exit(1);
	  }	
	  monitorEvent m;
	  int max = 1;
	  while (inputTestText >> m) {
	    Msequence.push_back(m);
		if (m.getThreadID()>max) max = m.getThreadID();
	  }
	  if (Msequence.size() == 0) {
	 	std::cout << "Error opening traceFile for testing: empty file" << std::endl;
		exit(1);
	  }
	 // std::cout << "M-sequence size is " << Msequence.size() << std::endl;
	  numberOfThreads = max;
	  inputTestText.close();
	  threads = new controlConditionVariable*[numberOfThreads+2];
	  //std::cout << "making " << (numberOfThreads+1) << std::endl;
	  for (int i=0; i<(numberOfThreads+2); i++)
			threads[i] = new controlConditionVariable(this);
	 }
	 else if (mode == SPECTEST) {
		index = 0;
		inputSpecTestText.open((traceFileName+"-spectest.txt").c_str());
		if (inputSpecTestText.fail()) {
			std::cout << "Error opening traceFile for testing" << std::endl;
			exit(1);
		}	
		monitorSpecEvent m;
		int max = 1;
		while (inputSpecTestText >> m) {
			COMMsequence.push_back(m);
			if (m.getThreadID()>max) max = m.getThreadID();
		}
		if (COMMsequence.size() == 0) {
			std::cout << "Error opening traceFile for testing: empty file" << std::endl;
			exit(1);
		}
		//std::cout << "COMM-sequence size is " << COMMsequenceWrap->size() << std::endl;
		numberOfThreads = max;
		threads = new controlConditionVariable*[numberOfThreads+2];
		inputSpecTestText.close();
		//std::cout << "making " << (numberOfThreads+2) << std::endl;
		for (int i=0; i<(numberOfThreads+2); i++) {
			threads[i] = new controlConditionVariable(this); 
		}

	 if (mode == TEST)
		hTestThread = (HANDLE)_beginthreadex(NULL,0,
			(PBEGINTHREADEX_THREADFUNC) monSpecTestWatchDog,
			(LPVOID) this, 0,
			(PBEGINTHREADEX_THREADID) &winThreadID );

    }
 	if (mode == SPECTEST)
		hSpecTestThread = (HANDLE)_beginthreadex(NULL,0,
			(PBEGINTHREADEX_THREADFUNC) monSpecTestWatchDog,
			(LPVOID) this, 0,
			(PBEGINTHREADEX_THREADID) &winThreadID );

    }
	
	monitorTracingReplayTest::~monitorTracingReplayTest() { 
		if (mode == TRACE) outputReplayText.close();
		if (mode == REPLAY)
			for (int i=0; i<(numberOfThreads+1);i++)
				delete threads[i];
		if (mode == TEST || mode == SPECTEST)
			for (int i=0; i<(numberOfThreads+2);i++)
				delete threads[i];	
		if (mode == TEST) {
			TerminateThread(hTestThread, 0);
			CloseHandle(hTestThread);
		}
		if (mode == SPECTEST) {
			TerminateThread(hSpecTestThread, 0);
			CloseHandle(hSpecTestThread);
		}
	}

	 void monitorTracingReplayTest::requestMPermit(monitorEvent::eventTypes op, int ID, std::string methodName, 
		                         std::string conditionName) {
   		enterMonitor();
		monitorEvent nextEvent;
		if (index < (int) Msequence.size()) {
			nextEvent = Msequence[index];
    		if (ID != nextEvent.getThreadID()) {
				if (ID <= numberOfThreads) {
      				threads[ID]->waitC();
					nextEvent = Msequence[index];  //update fields with correct event
				}
				else {
					threads[numberOfThreads+1]->waitC(); // wait forever if not in sequence
				}
    		}
		}
		else threads[numberOfThreads+1]->waitC(); // wait forever if not in sequence

		if (!(op == nextEvent.getEventType())) {
			std::cout << "Infeasible Sequence at event " << (index+1) <<
				", unexpected event type " << eventTypes[op] << ".  ";
			std::cout << "Expected event " << (index+1) << " was: (" << 
				eventTypes[nextEvent.getEventType()] << "," <<
				nextEvent.getThreadID() << "," << nextEvent.getMethodName() <<
				"," << nextEvent.getConditionName() << ")" << std::endl;
			TerminateThread(hTestThread, 0);
			CloseHandle(hTestThread);
			exit(1);
		}
		if (!(methodName == nextEvent.getMethodName())) {
			std::cout << "Infeasible Sequence at event " << (index+1) <<
				", unexpected method name " << methodName << " on " << eventTypes[op] << " event.  ";
			std::cout << "Expected event " << (index+1) << " was: (" << 
				eventTypes[nextEvent.getEventType()] << "," <<
				nextEvent.getThreadID() << "," << nextEvent.getMethodName() <<
				"," << nextEvent.getConditionName() << ")" << std::endl;
			TerminateThread(hTestThread, 0);
			CloseHandle(hTestThread);
			exit(1);
		}
		if (op == monitorEvent::WAIT || op == monitorEvent::SIGNAL || op == monitorEvent::SIGNALALL || op == monitorEvent::SIGNALANDEXIT) {
			if (!(conditionName == nextEvent.getConditionName())) {
				std::cout << "Infeasible Sequence at event "  << (index+1) <<
					", unexpected condition variable name "  
					<< conditionName <<  " on " << eventTypes[op] << " event.  ";
				std::cout << "Expected event " << (index+1) << " was: (" <<
					eventTypes[nextEvent.getEventType()] << "," <<
					nextEvent.getThreadID() << "," << nextEvent.getMethodName() <<
					"," << nextEvent.getConditionName() << ")" << std::endl;
				TerminateThread(hTestThread, 0);
				CloseHandle(hTestThread);
				exit(1);
			}
			++index;
    		if ((size_t) index < Msequence.size())
      			threads[(Msequence[index]).getThreadID()]->signalC();
			else {
				std::cout << "\n\n***Sequence Completed***\n" << std::endl;
				TerminateThread(hTestThread, 0);
				CloseHandle(hTestThread);
				exit(0);
			}
		}
		else if (op == monitorEvent::EXIT) {
			++index;
    		if ((size_t) index < Msequence.size())
      			threads[(Msequence[index]).getThreadID()]->signalC();
			else {
				std::cout << "\n\n***Sequence Completed***\n" << std::endl;
				TerminateThread(hTestThread, 0);
				CloseHandle(hTestThread);
				exit(0);
			}
		}

    exitMonitor();
	}			
  
    void monitorTracingReplayTest::releaseMPermit() {
   	  enterMonitor();
      ++index;
      if ((size_t) index < Msequence.size()) {
      	threads[Msequence[index].getThreadID()]->signalC();
	  }
	  else
		std::cout << "\n\n***Sequence Completed***\n" << std::endl;
      exitMonitor();
    }

	void monitorTracingReplayTest::requestEntryPermit(int ID) {
      enterMonitor();
	  if (ID < numberOfThreads+1) {
		  if (index < (int) SMsequence.size()) {
			if (ID != SMsequence[index]) {
		        threads[ID]->waitC();
			}
		  }
		  else {
				threads[ID]->waitC();
		  }
	  }
	  else threads[numberOfThreads+1]->waitC();
      exitMonitor();
    }
  
    void monitorTracingReplayTest::releaseEntryPermit() {
      enterMonitor();
	  ++index;
      if ((size_t) index < SMsequence.size()) {
        threads[SMsequence[index]]->signalC();
      }
	  else
		std::cout << "\n\n***Sequence Completed***\n" << std::endl;
	  exitMonitor();
    }


    void monitorTracingReplayTest::trace(monitorEvent m) {
      enterMonitor();
      /* record integer ID in trace file */
	  outputTestText << m << std::endl;
	  if (m.getEventType() == monitorEvent::ENTRY || m.getEventType() == monitorEvent::REENTRY)
		  outputReplayText << m.threadID << std::endl;
      exitMonitor();
    }

	void monitorTracingReplayTest::exerciseEvent(int ID, std::string eventName) {
		//std::cout << ID << " " << eventName << std::endl;
		requestAndReleaseCommPermit(ID, eventName);
	} 

	void monitorTracingReplayTest::requestMPermitSpecTest(int ID) {
   		enterMonitor();
		if (index < (int) COMMsequence.size()) {
			monitorSpecEvent nextEvent = COMMsequence[index];
			if (ID != nextEvent.getThreadID()) {
				if (ID <=numberOfThreads) {
      				threads[ID]->waitC();
					nextEvent = COMMsequence[index];  //update fields with correct event
				}
				else {
					threads[numberOfThreads+1]->waitC();
				}
			}
		}
		else threads[numberOfThreads+1]->waitC();
		exitMonitor();
	}	

	void monitorTracingReplayTest::requestAndReleaseCommPermit(int ID, std::string eventName) {
 		enterMonitor();
		monitorSpecEvent nextEvent;
		if (index < (int) COMMsequence.size()) {
			nextEvent = COMMsequence[index];
    		if (ID != nextEvent.getThreadID()) {
				if (ID <=numberOfThreads) {
		  			threads[ID]->waitC();
					nextEvent = COMMsequence[index];  //update fields with correct event
				}
				else {
					threads[numberOfThreads+1]->waitC();
				}
			}
		}
		else threads[numberOfThreads+1]->waitC();

		if(eventName != nextEvent.geteventName()) {
			std::cout << "Infeasible Sequence at event " << (index+1) <<
				", unexpected event name " << eventName << std::endl;
			std::cout << "Expected event " << (index+1) << " was: (" << 
				nextEvent.getThreadID() << "," << nextEvent.geteventName() << ")" << std::endl;
			TerminateThread(hSpecTestThread, 0);
			CloseHandle(hSpecTestThread);
			exit(1);
		}

		// release
	    ++index;
		if (index < (int) COMMsequence.size()) {
			threads[COMMsequence[index].getThreadID()]->signalC_and_exitMonitor();
			return;
		}
		else {
			printf("\n\n***Sequence Completed***\n");
			//std::cout << "\n\nSequence Completed\n";
			TerminateThread(hSpecTestThread, 0);
			CloseHandle(hSpecTestThread);
			exit(0);
		}
		exitMonitor();
	}
  
	void monitorTracingReplayTest::traceCommEvent(monitorSpecEvent m) {
		enterMonitor();
		outputSpecTestText << m << std::endl;
		exitMonitor();
	}
