// this mutexLock suppoted race analysis but not tracing/replay
#include "mutexLock.h"
#include <windows.h>
#include "criticalSectionLock.h"
#include "MutexIDGenerator.h"
#include "threadData.h"
#include "TDThread.h"
#include "deadlockWatchdog.h"
#include <iostream>
#include <assert.h>

int mutexLock::ID=1;
criticalSectionLock mutexLock::idLock; // mutual exclusion in getID()

mutexLock::mutexLock(): owner(0), count(0), analysisIsOn(true), tracingIsOn(true), mutex(1) {
	char* env = getenv("MODE");
	if (env!=NULL) {
		if (!(strcmp(env,"TRACE")))
			mode = TRACE;
		else if (!(strcmp(env,"REPLAY")))
			mode = REPLAY;
		else if (!(strcmp(env,"TEST")))
			mode = TEST;
		else if (!(strcmp(env,"RT"))) {
			std::cout << "We are currently updating our C++ RT implementation." << std::endl;
			std::cout << "We will post new files shortly and enable RT." << std::endl;
			exit(1);
		}
		else if (!(strcmp(env,"NONE")))
			mode = NONE;
		else mode = NONE;
	}
	else mode = NONE;
 	if (mode != NONE && mode!=TRACE && mode!=REPLAY) {
		std::cout << "Error: Only modes TRACE, REPLAY, or NONE are supported by mutexLocks." << std::endl;
		exit(1);
	}

	env = getenv("DATARACEDETECTION");
	if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			raceAnalysis = true;
		else raceAnalysis = false;
	}
	else raceAnalysis = false;

	env = getenv("CONTROLLERS");
	if (env!=NULL) {
		if (!(strcmp(env,"SINGLE")))
			numControllers = 1;
		else numControllers = 0;
	}
	else numControllers = 1;

	env = getenv("RANDOMDELAY");
	if (env!=NULL) {	
		 if	(!(strcmp(env,"ON")))
			 randomDelay = true;
		 else
			 randomDelay = false;
	}
	else randomDelay = false;

	env = getenv("DEADLOCKDETECTION");
	if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			detectDeadlock = true;
		else
			detectDeadlock = false;
	 }
	 else detectDeadlock = false;

	if (raceAnalysis)
		mutexID = getID();  // getID() is a static function for distributing IDs

	 if (mode) {
		mutexName = generateID();
		std::string traceFileName;
		if (numControllers == 0) 
			traceFileName = mutexName;
		else
			traceFileName = "semaphoreandlock";
		if (mode == TRACE || mode == REPLAY)
			if (numControllers = 1) 
				control = instance(mode,numControllers,traceFileName);
			else
				control = new semaphoreAndLockTracingAndReplay(mode,numControllers,
							traceFileName);
	 }
}

mutexLock::mutexLock(analysisOnOff aOnOff, tracingOnOff tOnOff) : owner(0), count(0), mutex(1) {
	char* env = getenv("mode");
	if (env!=NULL) {
		if (!(strcmp(env,"TRACE")))
			mode = TRACE;
		else if (!(strcmp(env,"REPLAY")))
			mode = REPLAY;
		else if (!(strcmp(env,"TEST")))
			mode = TEST;
		else if (!(strcmp(env,"NONE")))
			mode = NONE;
		else mode = NONE;
	}
	else mode = NONE;
	if (mode != NONE && mode!=TRACE && mode!=REPLAY) {
		std::cout << "Error: Only modes TRACE, REPLAY, or NONE are supported by mutexLocks." << std::endl;
		exit(1);
	}
	env = getenv("DATARACEDETECTION");
	if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			raceAnalysis = true;
		else raceAnalysis = false;
	}
	else raceAnalysis = false;

	env = getenv("CONTROLLERS");
	if (env!=NULL) {
		if (!(strcmp(env,"SINGLE")))
			numControllers = 1;
		else numControllers = 0;
	}
	else numControllers = 1;

	env = getenv("RANDOMDELAY");
	if (env!=NULL) {	
		 if	(!(strcmp(env,"ON")))
			 randomDelay = true;
		 else
			 randomDelay = false;
	}
	else randomDelay = false;

	env = getenv("DEADLOCKDETECTION");
	if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			detectDeadlock = true;
		else
			detectDeadlock = false;
	}
	else detectDeadlock = false;

	if (aOnOff == analysisOn)
		analysisIsOn = true;
	else
		analysisIsOn = false;
	if (tOnOff == traceOn)
		tracingIsOn = true;
	else
		tracingIsOn = false;

	if (raceAnalysis && analysisIsOn)
		mutexID = getID();  // getID() is a static function for distributing IDs

	 if (mode && tracingIsOn) {
		mutexName = generateID();
		std::string traceFileName;
		if (numControllers == 0) 
			traceFileName = mutexName;
		else
			traceFileName = "semaphoreandlock";
		if (mode == TRACE || mode == REPLAY)
			if (numControllers = 1) 
				control = instance(mode,numControllers,traceFileName);
			else
				control = new semaphoreAndLockTracingAndReplay(mode,numControllers,
							traceFileName);
	 }
}

std::string mutexLock::generateID() {
		const std::string thisClassName = "mutexLock";
		std::string mutexName;
		data d = threadData::instance().getThreadData(GetCurrentThreadId());	
		if (d.threadID != -1) 
		  mutexName = d.threadName;
		else
		  mutexName = "";
		int position = mutexName.find("_"); 
		if (position > -1 && (size_t) position < mutexName.length())
		  mutexName = mutexName + "_" + thisClassName;
		else
		  mutexName = "main_" + thisClassName;		
		std::string myID = MutexIDGenerator::instance().getID(mutexName);
		return myID;
}

semaphoreAndLockTracingAndReplay* mutexLock::instance(int mode, 
			int numControllers, std::string traceFileName) {
	return semaphoreAndLockTracingAndReplay::instance(mode, numControllers,
		   traceFileName);
}

mutexLock::~mutexLock() {
	if (mode == TRACE || mode == REPLAY)
		if (numControllers = 0) 
		  delete control;
	// else control is a static shared by many threads - need to ensure it is 
	// deleted only once if you decide to delete it (need to lock control)

}

void mutexLock::lock() {
	if (mode == REPLAY) {
     	control->requestPermit((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
	}
    else if (randomDelay && mode == TRACE) {
	   Sleep(rand()%delayAmount);	// default delayAmount is 750
    }
	mutex.P();
	if (count == 0) {
		count = 1;
		owner = GetCurrentThreadId();
		if (mode == REPLAY)
			control->releasePermit();
		else if (mode == TRACE)
    		control->trace((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
	}
	else if (owner == GetCurrentThreadId()) {
		++count;
		if (mode == REPLAY)
			control->releasePermit();
		else if (mode == TRACE)
    		control->trace((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
	}
	else  {
		win32Semaphore* conditionVar = pool.acquire();
		if (detectDeadlock && (mode == RT || mode == TRACE)) {
			TDThread::incNumBlockedThreads();
			deadlockWatchdog::changeStatus(GetCurrentThreadId(),"blocked on lock operation of "+mutexName);
		}
		waitingLock.push(conditionVar); // otherwise append blocked thread
		mutex.V();
		if (mode == REPLAY) {
			std::cout << "failing for " << (threadData::instance().getThreadData(GetCurrentThreadId())).threadID << std::endl;
			assert(false);
		}
		conditionVar->P();		// returns holding mutex since unlock() didn't release it
		if (detectDeadlock && (mode == RT || mode == TRACE)) {
			deadlockWatchdog::changeStatus(GetCurrentThreadId(),"running");
		 }
		if (mode == REPLAY)
			assert(false);
		if (mode == TRACE) 
			control->trace((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
		count = 1;			
		owner = GetCurrentThreadId();
		pool.release(conditionVar);
	}

	if (raceAnalysis && analysisIsOn) {
		lockData::instance().addLocked(mutexID);
	}

	mutex.V();
}

int mutexLock::unlock() {
	if (mode == REPLAY)
     	control->requestPermit((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
    else if (randomDelay && mode == TRACE) {
	   Sleep(rand()%delayAmount);	// default delayAmount is 750
    }
	int returnCode=0;
	mutex.P();
	if (owner != 0) {
		if (owner == GetCurrentThreadId()) {
				--count;
				if (count == 0) {
					owner = 0;
					if (!waitingLock.empty()) { // any threads waiting for lock?
						win32Semaphore* oldest = waitingLock.front();
						waitingLock.pop();
						if (mode == REPLAY)
							assert(false);
						if (mode == TRACE) 
    						control->trace((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
						if (returnCode==0)
							if (raceAnalysis && analysisIsOn)
								lockData::instance().removeLocked(mutexID); 
						oldest->V();
						if (detectDeadlock && (mode == RT || mode == TRACE)) {
							TDThread::decNumBlockedThreads();
						}
						// do not release mutex so oldest will complete
						return returnCode;
					}
				}
				else {
					if (mode == REPLAY)
						control->releasePermit();
					else if (mode == TRACE) 
    					control->trace((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
					if (returnCode==0)
						if (raceAnalysis && analysisIsOn)
							lockData::instance().removeLocked(mutexID);
					mutex.V();
					return returnCode;
				}
		}
		else {
			returnCode = -1;
		}
	}
	else {
		returnCode = -1;
	}

	if (mode == REPLAY)
		control->releasePermit();
	else if (mode == TRACE) 
    	control->trace((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);

	if (returnCode==0)
		if (raceAnalysis && analysisIsOn)
			lockData::instance().removeLocked(mutexID); 

	mutex.V();
	return returnCode;
}



