#ifndef SELECTABLESYNCHCHANNEL_H
#define SELECTABLESYNCHCHANNEL_H

#include "channel.h"
#include "win32semaphore.h"
#include "mutexLock.h"
#include "messageptr.h"
#include "IselectiveWait.h"
#include "criticalSectionLock.h"
#include <string>
#include "selectable.h"
#include "callMsg.h"
#include <list>

//class selectiveWaitC;

template <class T>
class selectableSynchChannelC /* : public channel<T>, */: public selectable /*<T>*/  {
public:
	//selectableSynchChannelC(): sent(0), received(0) {}
	selectableSynchChannelC() 
		: sent(0), received(0), msg(0), open(false), ready(0), inchoice(NULL),
		guard_(true) {}
	//void send(callMsg<T> sentMsg); 
	//callMsg<T> receive(); 
	//void send();

	void send(callMsg<T>* sentMsg) {
	  	//if (sentMsg->request.get() == NULL) {
		//	throw new NullMessageException("Null message passed to send()");
		//}
		mutexSending.lock();
		msg = sentMsg;
		signal();
		received.P(); 
		mutexSending.unlock();
	}

	callMsg<T>* receive() {
		mutexReceiving.lock();
		sent.P();
		clearReady();
		callMsg<T>* receivedMessage = msg; // msg's dumb pointer set to null;
		arrivals.pop_front();
		received.V();
		mutexReceiving.unlock();
		return receivedMessage;
	}

	void send() {
	// send a null message; msg's dumb pointer is constructed as null, 
	// and it's null after every receive() so use msg as it is
		mutexSending.lock();
		signal();
		received.P();
		mutexSending.unlock();
	}

	void guard(bool g) {
        guard_=g;
    }

	//typedef std::vector<selectable<T>*> Type;

private:
	callMsg<T>* msg;
	mutexLock mutexSending;
	mutexLock mutexReceiving;
	win32Semaphore sent;
	win32Semaphore received;
	win32Critical_Section mutex;

	//typedef std::vector<selectable<T>*> channelPtr;
    //static std::vector<channelPtr> channels;
	//typedef selectiveWaitC* selectiveWaitCPtr;
	//static std::vector<selectiveWaitCPtr> selectiveWaits;

	

	bool open;
	int ready;
	IselectiveWait* inchoice;
	bool guard_;
	static int timeStamp;
	std::list<int> arrivals;
	static criticalSectionLock classLock;
	//static int getTimeStamp();
	static int getTimeStamp() {
		mutexLocker<criticalSectionLock> mutex(classLock); 
		return(++timeStamp);
	}

	//void signal();
	//void setSelect(IselectiveWait* s);
	//int getOldestArrival();
	//int count();
    //bool testReady();
    //void clearReady();
    //void setOpen();
    //void clearOpen();
    //void guard(bool g);
    //bool testGuard();
	//std::string getChannelName();

	void signal() {
		// Q: is this lock necessary? since send() is locked
		mutexLocker<win32Critical_Section> m(mutex); 
		if (inchoice == NULL) {
			++ready; 
			arrivals.push_back(getTimeStamp());
			sent.V(); // if (open) notifyAll();
		} 
		else {
			inchoice->lock();
				++ready;
				arrivals.push_back(getTimeStamp());
				sent.V();		
				if (open)    // is this is an entry that choice is waiting for?
					inchoice->notify(); 
			inchoice->unlock();
		}
	}

	void setSelect(IselectiveWait* s) {
		mutexLocker<win32Critical_Section> m(mutex); 
		inchoice = s;
	}

	int getOldestArrival() {
	// only called from choose() when ready>0, i.e., arrivals.size()>0
		mutexLocker<win32Critical_Section> m(mutex); 
		return (arrivals.front());
	}

	int count() { return ready; }

    bool testReady() {
        return ready>0;
    }

    void clearReady() {
		mutexLocker<win32Critical_Section> m(mutex); 
        --ready;
    }

    void setOpen() {
        open=true;
    }

    void clearOpen() {
         open=false;
    }



    bool testGuard(){
        return guard_;
    }
	std::string getChannelName() { return "foo";}

};

template <class T>
criticalSectionLock selectableSynchChannelC<T>::classLock;
template <class T>
int selectableSynchChannelC<T>::timeStamp;
//template <class T>
//int selectableSynchChannelC<T>::channels;
//template <class T>
//int selectableSynchChannelC<T>::selectiveWaits;

#endif
