#ifndef SREVENT_H
#define SREVENT_H

#include <string>
#include <vector>
#include "OpenEventX.h"
#include "vectorTimeStamp.h"
#include <memory>

class srEvent /*implements Serializable, Comparable */ {
public:
	friend class srThreadEvent;
	enum eventType {SR_SYNCHRONIZATION, ASYNCH_SEND, ASYNCH_RECEIVE, START_ENTRY, 
		FINISH_ENTRY, ELSE_DELAY, SEND_EXCEPTION, RECEIVE_EXCEPTION , 
		UNACCEPTED_SYNCH_SEND, UNACCEPTED_ASYNCH_SEND, UNACCEPTED_RECEIVE,
		THREAD_BASED_SEND, THREAD_BASED_RECEIVE, MONITOR_CALL, MONITOR_ENTRY,
		SEMAPHORE_CALL, SEMAPHORE_COMPLETION, LOCK_CALL, LOCK_COMPLETION, 
		SYNCH_SEND, SYNCH_RECEIVE} ;
protected:
	int caller;
	int called;
	int callerVersionNumber;
	int calledVersionNumber;
	std::string channelName;
	int channelVersionNumber;
	eventType eType;
	std::auto_ptr<vectorTimeStamp> vectorTS;
	std::string label;
	bool isEntry;
	std::vector<OpenEventX> openList;
    eventType eType2;    // added
    bool isRecollected;

	void copy(const srEvent& b);
	void free();


public:
	srEvent(const srEvent& b) {copy(b);}  // copy constructor
	const srEvent& operator=(const srEvent& b);
	~srEvent() {free();}
	srEvent() : isRecollected(false) {}

    srEvent(int caller_, int called_, int callerVersionNumber_,
			int calledVersionNumber_, std::string channelName_, int channelVersionNumber_, 
			eventType eType_, const vectorTimeStamp& vectorTS_)

			:	caller(caller_),	called(called_),	callerVersionNumber(callerVersionNumber_),
				calledVersionNumber(calledVersionNumber_), channelName(channelName_),
				channelVersionNumber(channelVersionNumber_), eType(eType_),
				vectorTS(vectorTS_.clone()),
				label(),isEntry(false), eType2(eType_), isRecollected(false) 
			{}

	srEvent(int caller_,  int called_, int callerVersionNumber_,
			int calledVersionNumber_, std::string channelName_, int channelVersionNumber_,
			eventType eType_, const vectorTimeStamp& vectorTS_,	
			std::vector<OpenEventX> openList_, eventType eType2_)
			: caller(caller_), called(called_), callerVersionNumber(callerVersionNumber_),
			  calledVersionNumber(calledVersionNumber_), channelName(channelName_),
			  channelVersionNumber(channelVersionNumber_), eType(eType_),
			  vectorTS(vectorTS_.clone()),
			  label(), isEntry(false), isRecollected(false),
			  openList(openList_), eType2(eType2_) // I think this is for RT, e.g., to identify that an
													// asynch_receive was really a semaphore_completion, etc.
			 {}


	srEvent(int caller_, int called_, int callerVersionNumber_,
			int calledVersionNumber_, std::string channelName_, int channelVersionNumber_, 
			eventType eType_, const vectorTimeStamp& vectorTS_,
			std::string label_,
			std::vector<OpenEventX> openList_, eventType eType2_) 
			: caller(caller_), called(called_), callerVersionNumber(callerVersionNumber_),
			  calledVersionNumber(calledVersionNumber_), channelName(channelName_),
			  channelVersionNumber(channelVersionNumber_), eType(eType_),
			  vectorTS(vectorTS_.clone()),
			  label(label_), isEntry(false), isRecollected(false),
			  openList(openList_), eType2(eType2_)
	{}

	srEvent(int caller_, int called_, int callerVersionNumber_,
			int calledVersionNumber_, std::string channelName_, int channelVersionNumber_,
			eventType eType_, const vectorTimeStamp& vectorTS_,
			std::string label_, bool isEntry_,
			std::vector<OpenEventX> openList_, eventType eType2_)
			: caller(caller_), called(called_), callerVersionNumber(callerVersionNumber_),
			  calledVersionNumber(calledVersionNumber_),
			  channelName(channelName_), channelVersionNumber(channelVersionNumber_),
			  eType(eType_),
			  vectorTS(vectorTS_.clone()),
			  label(label_), isEntry(isEntry_), isRecollected(false),
			  openList(openList_), eType2(eType2_)
	{}

	void setCaller(int caller_) {
		caller = caller_;
	}

	void setCalled(int called_) {
		called = called_;
	}
	void setCallerVersionNumber(int callerVersionNumber_) {
		callerVersionNumber = callerVersionNumber_;
	}
	void setCalledVersionNumber(int calledVersionNumber_) {
		calledVersionNumber = calledVersionNumber_;
	}
	void setChannelName(std::string channelName_) {
		channelName = channelName_;
	}
	void setChannelVersionNumber(int channelVersionNumber_) {
		channelVersionNumber = channelVersionNumber_;
	}
	void setEventType(eventType eType_) {
		eType = eType_;
	}
	void setLabel(std::string label_) {
	   label = label_;
	}
	
	void setIsEntry(bool isEntry_) {
	   isEntry = isEntry_;
	}
	void setOpenList(std::vector<OpenEventX> openList_) {
	   openList = openList_;
	}
	void setEventType2(eventType eType2_) {
		eType2 = eType2_;
	}
	void setIsRecollected(bool isRecollected_) {
		isRecollected = isRecollected_;
	}
	int getCaller() const {
		return caller;
	}
	int getCalled() const {
		return called;
	}
	int getCallerVersionNumber() const {
		return callerVersionNumber;
	}
	int getCalledVersionNumber() const {
		return calledVersionNumber;
	}
	std::string getChannelName() const  {
		return channelName;
	}
	int getChannelVersionNumber() const {
		return channelVersionNumber;
	}
	eventType getEventType() const {
		return eType;
	}
	vectorTimeStamp& getVectorTS() {
		return *(vectorTS.get());
	}
	std::string getLabel() const  {
		return label;
	}
	bool getIsEntry() const {
		return isEntry;
	}
	std::vector<OpenEventX> getOpenList() const {
		return openList;
	}	
	eventType getEventType2() const  {
		return eType2;
	}
	bool getIsRecollected() const {
		return isRecollected;
	}

	friend std::ostream& operator<<(std::ostream& os, const srEvent& e);

	friend bool operator==(const srEvent& left, const srEvent& right);
	//friend bool operator<(const srEvent& left, const srEvent& right);
};

#endif