#ifndef SRTHREADEVENT_H
#define SRTHREADEVENT_H

#include "srEvent.h"

class srThreadEvent {
private:
	int thisThread;
	int thisThreadVersionNumber;
	int otherThread;
	int otherThreadVersionNumber;
	std::string channelName;
	int channelVersionNumber;
	srEvent::eventType eType;
	std::auto_ptr<vectorTimeStamp> vectorTS;

	void copy(const srThreadEvent& b);
	void free();

public:
	srThreadEvent(const srThreadEvent& b) {copy(b);}  // copy constructor
	const srThreadEvent& operator=(const srThreadEvent& b);
	~srThreadEvent() {free();}
	srThreadEvent() {}
	srThreadEvent(int thisThread_, int thisThreadVersionNumber_, int otherThread_, int otherThreadVersionNumber_,
		std::string channelName_, int channelVersionNumber_, srEvent::eventType eType_,
		vectorTimeStamp vectorTS_) 
		:
		thisThread(thisThread_),
		thisThreadVersionNumber(thisThreadVersionNumber_),
		otherThread(otherThread_),
		otherThreadVersionNumber(otherThreadVersionNumber_),
		channelName(channelName_),
		channelVersionNumber(channelVersionNumber_),
		eType(eType_),
		vectorTS(vectorTS_.clone())
		{}

	void setThisThread(int thisThread_) {
		thisThread = thisThread_;
	}
	void setOtherThread(int otherThread_) {
		otherThread = otherThread_;
	}
	void setThisThreadVersionNumber(int thisThreadVersionNumber_) {
		thisThreadVersionNumber = thisThreadVersionNumber_;
	}
	void setOtherThreadVersionNumber(int otherThreadVersionNumber_) {
		otherThreadVersionNumber = otherThreadVersionNumber_;
	}
	void setChannelVersionNumber(int channelVersionNumber_) {
		channelVersionNumber = channelVersionNumber_;
	}
	void setChannelName(std::string channelName_) {
		channelName = channelName_;
	}
	void setEventType(srEvent::eventType eType_) {
		eType = eType_;
	}
	int getThisThreadVersionNumber() {
		return thisThreadVersionNumber;
	}
	int getThisThread() {
		return thisThread;
	}
	int getOtherThread() {
		return otherThread;
	}
	int getOtherThreadVersionNumber() {
		return otherThreadVersionNumber;
	}
	int getChannelVersionNumber() {
		return channelVersionNumber;
	}
	std::string getChannelName() {
		return channelName;
	}
	srEvent::eventType getEventType() {
		return eType;
	}

	vectorTimeStamp& getVectorTS() {
		return *(vectorTS.get());
	}

	friend std::ostream& operator<<(std::ostream& os, const srThreadEvent& e);
	friend bool operator==(const srThreadEvent& left, const srThreadEvent& right);

};

#endif