#include "vectorTimeStamp.h"
#include "Timestamp.h"
#include <string>

void vectorTimeStamp::makeVectorTimeStamp(Timestamp& t) { 
    // Timestamp is Jeff's RT class, not one above	
	int dimension = t.getDimension();
	for (int i = 0; i < dimension; i ++) {
		std::pair<std::map <int, int >::iterator, bool> result1 = vectorTS->insert (std::pair<int,int>(i, t.getValue(i)));
   	}
}
    
bool vectorTimeStamp::lessThan(vectorTimeStamp* v) const {
	if (isSame(v)) {
	    return false;
	}
	else {	
		if ( v == NULL || v->getSize() == 0) {
		    return false;
		}
		if (getSize() > v->getSize()) {
		    // this means that this timestamp has more knowledge than v
		    return false;
		} 
		std::map< int, int>::iterator pos;
		for (pos = vectorTS->begin(); pos != vectorTS->end(); ++pos) {
			int ID = pos->first;
			int myClockValue = pos->second;
			std::map< int, int>::iterator m;
			if ((m = v->vectorTS->find(ID)) == v->vectorTS->end())
				return false;
			else
				if (myClockValue > m->second)
					return false;
		}
/* TBD
			Set s = vectorTS.entrySet ();
			Iterator it = s.iterator ();
			while (it.hasNext ()) {
			    Map.Entry entry = (Map.Entry) (it.next());
			    Integer ID = (Integer) entry.getKey ();
			    int myClockValue = ((timestampRC) entry.getValue()).clockValue;
			    timestampRC otherTimeStamp = (timestampRC) v.vectorTS.get(ID);
			    if (otherTimeStamp == null) {
				return false;
			    }
			    else {
				if (myClockValue > otherTimeStamp.clockValue) {
				    return false;
				}
			    }
			}
*/
	}
	return true;
} 

vectorTimeStamp::vectorTimeStamp() { 
		// called by clone when copying contents of HashMap
	vectorTS = new std::map< int, int >;
}

vectorTimeStamp::vectorTimeStamp(int ID) { 
	// called by TDthread to initilaize vector TimeStamp for thread ID
	vectorTS = new std::map< int, int >;
	std::pair<std::map <int, int >::iterator, bool> result1 = vectorTS->insert (std::make_pair(ID, 0));
}

void vectorTimeStamp::init(int ID) { 
	// called to initilizee vector TimeStamp for monitors/semaphores
	vectorTS = new std::map< int, int >;
	std::pair<std::map <int, int >::iterator, bool> result1 = vectorTS->insert (std::make_pair(ID, 0));
}

vectorTimeStamp::vectorTimeStamp(const vectorTimeStamp& b) {copy(b);}  // copy constructor
vectorTimeStamp::~vectorTimeStamp() { free(); }   // destructor
const vectorTimeStamp& vectorTimeStamp::operator=(const vectorTimeStamp& b) {  // copy assignment
  if (this != &b) {
    free();
    copy(b);
  }
  return *this;
}
void vectorTimeStamp::copy(const vectorTimeStamp& b) { 
	vectorTS = new std::map< int, int >;
 	std::map< int, int>::iterator pos;
	for (pos = b.vectorTS->begin(); pos != b.vectorTS->end(); ++pos) {
		int otherID = pos->first;
		int otherClockValue = pos->second;
	   	// If the clock value of event v > than that of the event in
	   	// vectorTS with the same ID, replace the old event
	   	// with the new event.
		std::pair<std::map <int, int >::iterator, bool> result1 = vectorTS->insert(std::make_pair(otherID, otherClockValue));
	}
}
void vectorTimeStamp::free() { delete vectorTS;}


void vectorTimeStamp::reset() {
	vectorTS->clear();
}

// added by rhc for SBRT	
void vectorTimeStamp::reset(int ID) {
	vectorTS->clear();
	std::pair<std::map <int, int >::iterator, bool> result1 = vectorTS->insert (std::make_pair(ID, 0));
}

// added by lei
int vectorTimeStamp::getSize () const {
	return vectorTS->size ();
}

vectorTimeStamp* vectorTimeStamp::clone() const{
	return new vectorTimeStamp(*this); // invoke copy constructor to copy *this

/*
	// We can't simply clone HashMap because entries are not cloned. 
	// So create new table with new entries.
		vectorTimeStamp* v = new vectorTimeStamp();
		std::map< int, int>::iterator pos;
		for (pos = vectorTS->begin(); pos != vectorTS->end(); ++pos) {
			int ID = pos->first;
			int myClockValue = pos->second;
			std::pair<std::map <int, int >::iterator, bool> result1 = v->vectorTS->insert(std::make_pair(ID, myClockValue));
		}
		return v;

		Set s = vectorTS.entrySet();
		Iterator i = s.iterator();
		while (i.hasNext()) {
			Map.Entry e = (Map.Entry)(i.next());
			int ID = ((Integer)e.getKey()).intValue();
			int clockValue = ((timestampRC)e.getValue()).clockValue;
			//v.vectorTS.put(new Integer(ID),new timestampRC(clockValue));
			v.vectorTS.put(integerSpace.get(ID),timestampRCSpace.get(clockValue));
		}
*/
}
	
// compares the current vector timestamp with the passed in vector timestamp
bool vectorTimeStamp::isSame(vectorTimeStamp* v) const {

	if (v == NULL || getSize () != v->getSize()) {
	   	return false;
	}
	std::map< int, int>::iterator pos;
	for (pos = vectorTS->begin(); pos != vectorTS->end(); ++pos) {
		int ID = pos->first;
		int myClockValue = pos->second;
		std::map< int, int>::iterator m;
		if ((m = v->vectorTS->find(ID)) == v->vectorTS->end())
			return false;
		else
			if (myClockValue != m->second)
				return false;
	}
/*
	Set s = v.getVectorTS().vectorTS.entrySet();
	Iterator i = s.iterator();
	while (i.hasNext()) {
		Map.Entry e = (Map.Entry)(i.next());
		Integer ID = (Integer)e.getKey();
		int otherClockValue = ((timestampRC)e.getValue()).clockValue;
		timestampRC thisTimeStamp = (timestampRC)vectorTS.get(ID);
		if (thisTimeStamp == null)
			return false; // no clockValue yet for this thread
		else if (otherClockValue != thisTimeStamp.clockValue)
			return false;
	}
*/
	return true;
}

bool vectorTimeStamp::equals (vectorTimeStamp* v) const {
	return isSame(v);
}
					
 
// Returns a clone of internal hash table
const vectorTimeStamp& vectorTimeStamp::getVectorTS()const {
	return *this; // ->clone();
}

// Returns the integer clock associated with the given thread ID
int vectorTimeStamp::getIntegerTS(int ID) const { 

	std::map< int, int>::iterator m;
	if ((m = vectorTS->find(ID)) != vectorTS->end())
		return m->second;
	else
		return -1;
	//timestampRC ts = ((timestampRC)vectorTS.get(integerSpace.get(ID)));
	//if (ts == null)
	//	return -1;
	//else
	//	return ts.clockValue;	
}

// Increments integer clock associated with given thread
void vectorTimeStamp::updateIntegerTS(int ID) {

	std::map< int, int>::iterator m;
	if ((m = vectorTS->find(ID)) != vectorTS->end()) {
		int newClockValue = m->second;
		newClockValue++;
		vectorTS->erase(m);
		std::pair<std::map <int, int >::iterator, bool> result1 = vectorTS->insert (std::make_pair(ID, newClockValue));
	}
	else {
		std::cout << "Internal error in vectorTimeStamp::updateIntegerTS" << std::endl;
		exit(1);
	}
	//timestampRC ts = ((timestampRC)vectorTS.get(integerSpace.get(ID)));
	//if (ts != null) {
	//	//++ts.clockValue;
	//	int newClockValue = ts.clockValue;
	//	newClockValue ++;
	//	vectorTS.put(integerSpace.get(ID),timestampRCSpace.get(newClockValue)); 
	//	// overwrites old value
	//}
}


   // Updates integer clock of the given thread with the given clock value
void vectorTimeStamp::setIntegerTS(int ID, int clockValue) {

	std::map< int, int>::iterator m;
	if ((m = vectorTS->find(ID)) != vectorTS->end()) {
		vectorTS->erase(m);
		std::pair<std::map <int, int >::iterator, bool> result1 = vectorTS->insert(std::make_pair(ID, clockValue));
	}
	else {
		std::cout << "Internal error in vectorTimeStamp::setIntegerTS" << std::endl;
		exit(1);
	}

	//timestampRC ts = (timestampRC)vectorTS.get(integerSpace.get(ID));
	//if (ts != null)
	//	//ts.clockValue = clockValue;
	//	vectorTS.put(integerSpace.get(ID),timestampRCSpace.get(clockValue)); 
	//	// overwrites old value
}

/*
  * Updates the current vector time stamp with the given timestamp.
  * If an entry is not present then adds to the current and
  * updates the entries with given vectorTS entries
*/
 void vectorTimeStamp::updateVectorTS(const vectorTimeStamp& v) {
 	std::map< int, int>::const_iterator pos;
	for (pos = v.vectorTS->begin(); pos != v.vectorTS->end(); ++pos) {
		int otherID = pos->first;
		int otherClockValue = pos->second;
	   	// If the clock value of event v > than that of the event in
	   	// vectorTS with the same ID, replace the old event
	   	// with the new event.
		std::map< int, int>::iterator m;
		if ((m = vectorTS->find(otherID)) == vectorTS->end())
			std::pair<std::map <int, int >::iterator, bool> result1 = vectorTS->insert(std::make_pair(otherID, otherClockValue));
		else if (otherClockValue > m->second) {
				vectorTS->erase(m);
				std::pair<std::map <int, int >::iterator, bool> result1 = vectorTS->insert(std::make_pair(otherID, otherClockValue));
		}

	}
 /*
	Set s = v.getVectorTS().vectorTS.entrySet();
	Iterator i = s.iterator();
	while (i.hasNext()) {
		Map.Entry e = (Map.Entry)(i.next());
		Integer ID = (Integer)e.getKey();
	   	// If the clock value of event > than that of the event in
	   	// vectorTS with the same ID, replace the old event
	   	// with the new event.
		int otherClockValue = ((timestampRC)e.getValue()).clockValue;
		timestampRC thisTimeStamp = (timestampRC)vectorTS.get(ID);
		if (thisTimeStamp == null) {
			//vectorTS.put(ID,new timestampRC(otherClockValue));
			vectorTS.put(ID,timestampRCSpace.get(otherClockValue));
		}
		else if (otherClockValue > thisTimeStamp.clockValue) {
			// overwrites old value
			//vectorTS.put(ID,new timestampRC(otherClockValue)); 
			vectorTS.put(ID,timestampRCSpace.get(otherClockValue)); 
		}
	}
*/
}

std::ostream& operator<<(std::ostream& os, const vectorTimeStamp& v) {
	os << v.toString();
	return os;
}
	
 std::string vectorTimeStamp::toString() const {
	std::map< int, int>::iterator pos;
	int maxID = 0;
	std::string s;
	s.append("[");
	for (pos = vectorTS->begin(); pos != vectorTS->end(); ++pos) {
		int ID = pos->first;
		if (ID > maxID) maxID = ID;
	}
	for (int i=1; i<=maxID; i++) {
		int clockValue = getIntegerTS(i);
		if (clockValue == -1)
			s.append("0");
		else {
			char buffer[33];
			itoa(clockValue,buffer,10);
			s.append(buffer);
		}
		if (i<maxID)
			s.append(",");
	}
	s.append("]");
	return s;

 /*
	StringBuffer stringVectorTS = new StringBuffer("[");
	int clockValue;
	Set s = vectorTS.keySet();
	Integer m = (Integer) Collections.max(s);
	for (int i = 1; i<=m.intValue(); i++) {
		clockValue = getIntegerTS(i);
		if (clockValue == -1)
			stringVectorTS.append("0");
		else
			stringVectorTS.append(clockValue);
		stringVectorTS.append(",");
	}
	stringVectorTS.setCharAt(stringVectorTS.length()-1,']');
	return stringVectorTS.toString();
*/
}