#ifndef VECTORTIMESTAMP_H
#define VECTORTIMESTAMP_H

#pragma warning(disable:4786) 
#include "Timestamp.h"
#include <map>
#include <string>

class vectorTimeStamp {

	//need to clone too + <<
private:
	std::map< int, int>* vectorTS;
	void copy(const vectorTimeStamp& b);
	void free();

public:
     /*
     * The timestamps are stored in a hash table. Thread IDs are used as 
	  * keys and timeStamps are the associated timestamp values.
     */
	vectorTimeStamp(const vectorTimeStamp& b);  // copy constructor
	~vectorTimeStamp();   // destructor
	const vectorTimeStamp& operator=(const vectorTimeStamp& b);
    void makeVectorTimeStamp(Timestamp& t);
    bool lessThan (vectorTimeStamp* v) const;
    vectorTimeStamp();
	vectorTimeStamp(int ID);
	void vectorTimeStamp::init(int ID);
	void reset();
	// added by rhc for SBRT	
	void reset(int ID);
    // added by lei
    int getSize() const;
	vectorTimeStamp* clone() const;
	// compares the current vector timestamp with the passed in vector timestamp
	bool isSame(vectorTimeStamp* v) const;
    bool equals (vectorTimeStamp* v) const;
	// Returns a clone of internal hash table
	const vectorTimeStamp& getVectorTS() const;
    // Returns the integer clock associated with the given thread ID
	int getIntegerTS(int ID) const;
	// Increments integer clock associated with given thread
	void updateIntegerTS(int ID);
	// Updates integer clock of the given thread with the given clock value
	void setIntegerTS(int ID, int clockValue);
   /*
    * Updates the current vector time stamp with the given timestamp.
    * If an entry is not present then adds to the current and
    * updates the entries with given vectorTS entries
	*/
    void updateVectorTS(const vectorTimeStamp& v);
	std::string toString() const;
	friend std::ostream& operator<<(std::ostream& os, const vectorTimeStamp& e);
};

#endif