/* This is an example program for using mailbox and message_ptr */
#include "messageptr.h"
#include "mailbox.h"
#include "thread.h"


class Message {
public:
	int contents;
	Message(int contents_) : contents(contents_){}
	Message(const Message& m) : contents(m.contents) {}
	Message* clone() const {return new Message(*this);}
};
class Producer : public Thread {
private:
	mailbox<Message>& deposit;
public:
	Producer (mailbox<Message>& deposit_) : deposit(deposit_) {}
	virtual void* run () {
	for (int i=0; i<3; i++) {
	message_ptr<Message> m(new Message(i));
	std::cout << "Producing " << i << std::endl;
	deposit.send(m);
	}
	return 0;
	}
};
class Consumer : public Thread {
private:
	mailbox<Message>& withdraw;
public:
	Consumer (mailbox<Message>& withdraw_) : withdraw(withdraw_) {}
	virtual void* run () {
	for (int i=0; i<3; i++) {
	message_ptr<Message> m = withdraw.receive();
	std::cout << "Consumed " << m->contents << std::endl;
	}
		return 0;
	}
};
class Buffer : public Thread {
private:
	mailbox<Message>& deposit; mailbox<Message>& withdraw;
public:
	Buffer (mailbox<Message>& deposit_, mailbox<Message>& withdraw_) 
	: deposit(deposit_), 	withdraw(withdraw_) {}
	virtual void* run () {
	for (int i=0; i<3; i++) {
	message_ptr<Message> m = deposit.receive(); withdraw.send(m);
	}
		return 0;
	}
};
int main () {
	mailbox<Message> deposit;	mailbox<Message> withdraw;
	std::auto_ptr<Producer> producer(new Producer(deposit));
	std::auto_ptr<Consumer> consumer(new Consumer(withdraw));
	std::auto_ptr<Buffer> buffer(new Buffer(deposit,withdraw));	
	producer->start();	 consumer->start(); buffer->start();
	producer->join();  consumer->join(); buffer->join();
	return 0;
}
