import java.net.*;
import java.io.*;

class distributedProcess2 extends Thread {
	private int ID;									// process ID
	private int number;								// sent on requests
	private int replyCount;							// replies received so far
	final private int numberOfProcesses = 4;
	final private int basePort = 2020;
	String processHost = null;
	// requests sent to other processes
	private TCPSender[] sendRequests = new TCPSender[numberOfProcesses];
	private TCPSender[] sendReplies = new TCPSender[numberOfProcesses];
	private TCPMailbox receiveRequests = null;
	private TCPMailbox receiveReplies = null;
	private boolean[] deferred = null;			// true means reply to other process has been deferred
	private Coordinator C;							// monitor
	private Helper helper;							// manages incoming requests


	distributedProcess2(int ID) {
		this.ID = ID;
		try {
			processHost = InetAddress.getLocalHost().getHostName();  
		}
		catch (UnknownHostException e) {e.printStackTrace();System.exit(1);}
		for (int i = 0; i < numberOfProcesses; i++) {
			sendRequests[i] = new TCPSender(processHost,basePort+(2*i));
			sendReplies[i] = new TCPSender(processHost,basePort+(2*i)+1);
		}
		receiveRequests = new TCPMailbox(basePort+(2*ID),"RequestsFor"+ID);
		
		receiveReplies = new TCPMailbox(basePort+(2*ID)+1,"RepliesFor"+ID);
		C = new Coordinator();
		deferred = new boolean[numberOfProcesses];
		for (int i=0; i<numberOfProcesses; i++) deferred[i] = false;
		helper = new Helper(); // ("helper"+ID);
		helper.setDaemon(true); // start helper in run() method
	}

	class Helper extends Thread {
	// manages requests from other distributed processes
		public void run() { // handle requests from other distributedProcesses
			//try {Thread.sleep(1000);} catch(InterruptedException e) {}
			while (true) {
				messageParts msg = (messageParts) receiveRequests.receive();
				requestMessage m = (requestMessage) msg.obj;
				if (!(C.deferrMessage(m))) {
					messageParts reply = new messageParts(new Integer(ID),processHost,basePort+(2*ID));
					sendReplies[m.ID].send(reply);	
				}
			}
		}
	}

class Coordinator extends monitorSC {
// Synchronizes the distributed process and its helper
	private boolean requestingOrExecuting = false;			// true if in or requesting to enter CS
	private int highNumber;										// highest number so far

	public boolean deferrMessage(requestMessage m) {
		enterMonitor();
		highNumber = Math.max(highNumber,m.number);
		boolean deferMessage = requestingOrExecuting && ((number < m.number)
			|| (number == m.number && ID < m.ID));
		if (deferMessage) {
			deferred[m.ID] = true;
		}
		exitMonitor();
		return deferMessage;
	}

	public int chooseNumberAndSetRequesting() {
		enterMonitor();
		requestingOrExecuting = true;
		number = highNumber + 1;
		exitMonitor();
		return number;
	}

	public void resetRequesting() {
		enterMonitor();
		requestingOrExecuting = false;
		exitMonitor();
	}

}

	public void run() {
		int count = 0;
		try {
			Thread.sleep(2000);  // let other processes start
		} catch (InterruptedException e) {e.printStackTrace();System.exit(1);}
		System.out.println("Process " + ID + " starting");

		
		for (int i = 0; i < numberOfProcesses; i++) {
			if (i != ID) {
				sendRequests[i].connect();
				sendReplies[i].connect();
			}
		}
		
		helper.start();
				
		while (count++<3) {
			System.out.println(ID + " Before Critical Section");
			System.out.flush();
			int number = C.chooseNumberAndSetRequesting();
			sendRequests();
			waitForReplies();
			System.out.println(ID + " Entering Critical Section-"+count);
			System.out.println(ID + " Leaving Critical Section-"+count);
			System.out.flush();
			try {Thread.sleep(500);} catch (InterruptedException e) {}
			C.resetRequesting();
			replytoDeferredProcesses();
		}
		
		try {
			Thread.sleep(10000);  // let other processes finish
		} catch (InterruptedException e) {e.printStackTrace();System.exit(1);}
		
		for (int i = 0; i < numberOfProcesses; i++) {
			if (i != ID) {
				sendRequests[i].close();
				sendReplies[i].close();
			}
		}
	}


	public void sendRequests() {
		replyCount = 0;
		for (int i = 0; i < numberOfProcesses; i++) {
			if (i != ID) {
				messageParts msg = new messageParts(new requestMessage(number,ID));
				//System.out.println("sending request to process " + i);
				sendRequests[i].send(msg);
				System.out.println(ID + " sent request to " + i);
				try {Thread.sleep(1000);} catch (InterruptedException e) {}
			}
		}
	}

	public void replytoDeferredProcesses() {
	System.out.println("replying to deferred processes");
		for (int i=0; i < numberOfProcesses; i++)
			if (deferred[i]) {
				deferred[i] = false;
				messageParts msg = new messageParts(new Integer(ID));
				sendReplies[i].send(msg);
			}
	}

	public void waitForReplies() {
		while (true) {
			messageParts m = (messageParts) receiveReplies.receive();
			int receivedID = ((Integer)m.obj).intValue();
			replyCount++;
			if (replyCount == numberOfProcesses-1)
				break;
		}
	}
	
}

class requestMessage implements Serializable {
	public int ID;
	public int number;
	public requestMessage(int number, int ID) {
		this.ID = ID;
		this.number = number;
	}
}
