import java.util.*;
import java.net.*;
import java.io.*;

public class TCPSender {
	String destinationHostname = null;		//  destination host associated with this channel
	int destinationPort;							//  destination port associated with this channel
	Socket socket = null;
	ObjectOutputStream to = null;
	Object lock = new Object();
	messageParts nullMessage;
	int age = 0;									// age is incremented after each send; when age
	int ageLimit = 1000;                   // reaches ageLimit, stream "to" is reset().
														// See file Readme.txt.

   public TCPSender(String destinationHostname, int destinationPort ) {
		this.destinationHostname = destinationHostname;
		this.destinationPort = destinationPort;
		// Whenever we send a nullMessage, we make sure that we send the same object (See file Readme.txt)
		// (The object stream caches the first version of each object that it writes and just reuses it unless you reset the stream.)
		nullMessage = new messageParts(new Object());
	}
   
   public void connect() throws TCPChannelException {
		try {
			socket = new Socket( destinationHostname, destinationPort );
			to = new ObjectOutputStream(socket.getOutputStream());
		} catch (Exception e) { e.printStackTrace(); throw new TCPChannelException(e.getMessage());}
	
	}

	public void send( messageParts message) throws TCPChannelException{
		try {
			synchronized (lock) {
				to.writeObject(message); to.flush();
				age++;
				if (age % ageLimit == 0) {		// occasionally reset the stream
					to.reset(); // Objects previously written to the stream will not be refered to as 
					            // already being in the stream. They will be written to the stream again.
				}
			}
		} 
		catch (NullPointerException e) {
			e.printStackTrace();
			throw new TCPChannelException("null stream - call connect() before sending messages");
		}
		catch (Exception e) { e.printStackTrace(); throw new TCPChannelException(e.getMessage());}
   }
   
	public void send( ) throws TCPChannelException{
		send(nullMessage);
   }
   
   public void shutdownOutput() throws TCPChannelException {
		try {
			if (socket != null) socket.shutdownOutput();
		} catch (Exception e) { e.printStackTrace(); throw new TCPChannelException(e.getMessage());}
	}
  	public void close() throws TCPChannelException {
		try {
			if (to != null) to.close();  
			if (socket != null) socket.close();
		} catch (Exception e) { e.printStackTrace(); throw new TCPChannelException(e.getMessage());}
	}
}
