import java.net.*;
import java.io.*;

class distributedReadersAndWriters {
	public static void main(String[] args) {
		final int READER = 1;
		final int WRITER = 2;
		if (args.length != 2) {
			System.out.println("need 2 arguments: Process ID (IDs start with 0) and Reader/Writer");
			System.exit(1);
		}
		int ID = Integer.parseInt(args[0]);
		String readerWriter = args[1];
		int readerOrWriter=0;
		if (readerWriter.equals("Reader"))
			readerOrWriter = READER;
		else if (readerWriter.equals("Writer"))
			readerOrWriter = WRITER;
		else {
			System.out.println("Second argument must be Reader or Writer");
			System.exit(1);
		}
		new distributedRWProcess(ID,readerOrWriter).start();
	}
}

class distributedRWProcess extends Thread implements operationTypeParameters {
	private int ID;									// process ID

	private int number;								// sent on requests
	private int replyCount;							// replies received so far
	private final int basePort = 2020;
	String processHost = null;
	private int readerOrWriter;
	private final int numberOfProcesses = 5;
	private final int READER = 3;
	private final int WRITER = 2;
	private TCPSender[] sendRequests = new TCPSender[numberOfProcesses];
	private TCPSender[] sendReplies = new TCPSender[numberOfProcesses];
	private TCPMailbox receiveRequests = null;
	private TCPMailbox receiveReplies = null;
	// replies received in response to requests
	private boolean[] deferred = null;			// true means reply to other process has been deferred
	private Coordinator C;							// monitor
	private Helper helper;							// manages incoming requests

	distributedRWProcess(int ID, int readerOrWriter) {
		this.ID = ID;
		this.readerOrWriter = readerOrWriter;
		try {
			processHost = InetAddress.getLocalHost().getHostName();  
		}
		catch (UnknownHostException e) {e.printStackTrace();System.exit(1);}
		System.out.println(ID + " is a " + (readerOrWriter == 1 ? "READER" : "WRITER"));
		C = new Coordinator();
		for (int i = 0; i < numberOfProcesses; i++) {
			sendRequests[i] = new TCPSender(processHost,basePort+(2*i));
			sendReplies[i] = new TCPSender(processHost,basePort+(2*i)+1);
		}
		receiveRequests = new TCPMailbox(basePort+(2*ID));
		receiveReplies = new TCPMailbox(basePort+(2*ID)+1);
		deferred = new boolean[numberOfProcesses];
		for (int i=0; i<numberOfProcesses; i++) deferred[i] = false;
		helper = new Helper(); // ("helper"+ID);
		helper.setDaemon(true); // start helper in run() method
	}

	class Helper extends Thread {
	// manages requests from other distributed processes
		public void run() { // handle requests from other distributedProcesses
				for (int i = 0; i < numberOfProcesses; i++) {
					if (i != ID)
						sendReplies[i].connect();
				}
				while (true) {
				//System.out.println("Helper wating for receiveRequest.receive()");
				messageParts msg = (messageParts) receiveRequests.receive();
				requestMessage m = (requestMessage) msg.obj;
				if (!C.decideAboutDeferral(m)) {
					messageParts reply = new messageParts(new Integer(ID));
					sendReplies[m.ID].send(reply);
				}
			}
		}
	}

class Coordinator extends monitorSC {
// Synchronizes the distributed process and its helper
	private boolean requestingOrExecuting = false;			// true if requesting to enter CS
	private int highNumber;							// highest number so far

	private boolean compatible(int operation) {
		if (readerOrWriter == READER && operation == READER)
			return true; // only reads are compatible
		else
			return false;
	}

	public boolean decideAboutDeferral(requestMessage m) {
		enterMonitor();
		highNumber = Math.max(highNumber,m.number);
		boolean deferMessage = requestingOrExecuting && !compatible(m.operation) &&
			((number < m.number)	|| (number == m.number && ID < m.ID));
		if (deferMessage) {
			deferred[m.ID] = true;
		}
		exitMonitor();
		return deferMessage;
	}

	public int chooseNumberAndSetRequesting() {
		enterMonitor();
		requestingOrExecuting = true;
		number = highNumber + 1;
		exitMonitor();
		return number;
	}

	public void resetRequesting() {
		enterMonitor();
		requestingOrExecuting = false;
		exitMonitor();
	}

}

	public void run() {
		int count = 0;
		try {
			Thread.sleep(6000);  // let other processes start
		} catch (InterruptedException e) {e.printStackTrace();System.exit(1);}
		System.out.println("Process " + ID + " starting");

		System.out.println("Get Connections");		
		for (int i = 0; i < numberOfProcesses; i++) {
			if (i != ID) {
				sendRequests[i].connect();
			}
		}
		
		helper.start();
		
		while (count++<3) {
			System.out.println(ID + " Before " + (readerOrWriter == 1 ? "READ" : "WRITE") + " Operation");
			System.out.flush();
			try {Thread.sleep(500);} catch (InterruptedException e) {}
			int number = C.chooseNumberAndSetRequesting();
			sendRequests();
			waitForReplies();
			System.out.println(ID + " After " + (readerOrWriter == 1 ? "READ" : "WRITE") + " Operation");
			System.out.flush();
			try {Thread.sleep(500);} catch (InterruptedException e) {}
			C.resetRequesting();
			replytoDeferredProcesses();
		}
		
		try {
			Thread.sleep(10000);  // let other processes finish
		} catch (InterruptedException e) {e.printStackTrace();System.exit(1);}
		
		for (int i = 0; i < numberOfProcesses; i++) {
			if (i != ID) {
				sendRequests[i].close();
				sendReplies[i].close();
			}
		}
		
	}


	public void sendRequests() {
		replyCount = 0;
		for (int i = 0; i < numberOfProcesses; i++) 
			if (i != ID) {
				messageParts msg = new messageParts(new requestMessage(number,ID,readerOrWriter));
				sendRequests[i].send(msg);
			}
	}

	public void replytoDeferredProcesses() {
		for (int i=0; i < numberOfProcesses; i++)
			if (deferred[i]) {
				deferred[i] = false;
				messageParts msg = new messageParts(new Integer(ID));
				sendReplies[i].send(msg);
			}
	}

	public void waitForReplies() {
		while (true) {
			messageParts m = (messageParts) receiveReplies.receive();
			int receivedID = ((Integer)m.obj).intValue();
			replyCount++;
			if (replyCount == numberOfProcesses-1)
				break;
		}
	}
	
}

class requestMessage implements Serializable {
	public int ID;
	public int number;
	public int operation;
	public requestMessage(int number, int ID, int operation) {
		this.ID = ID;
		this.number = number;
		this.operation = operation;
	}
}




