import java.util.*;
import java.net.*;
import java.io.*;

public class TCPMailbox implements Runnable {
	private ServerSocket listen = null;
	private messageBuffer buffer = null;
	Socket socket = null;
	int port;
	String channelName = "";
	ObjectInputStream from = null;  
	private boolean closed = false;
	Object lock = new Object();
	
   public TCPMailbox( int port) throws TCPChannelException {
   	this(port,"");
   }
   
   public TCPMailbox( int port, String channelName ) throws TCPChannelException {
   	this.port = port;
   	this.channelName = channelName;
		try {listen = new ServerSocket( port ); } 
		catch (IOException e) { e.printStackTrace(); throw new TCPChannelException(e.getMessage());}
      buffer = new messageBuffer(100);
		Thread internal = new Thread(this);
		internal.start(); 
	}
	
	class messageBuffer extends monitorSU {  // SU monitor
		private int fullSlots = 0; 			// # of full slots in the buffer
		private int capacity = 0;
		private conditionVariable notFull = new conditionVariable(); 
		private conditionVariable notEmpty = new conditionVariable();
	   private LinkedList buffer;
	   public messageBuffer( int capacity ) {
			this.capacity = capacity;
			buffer = new LinkedList();
   	}
	   public void deposit( messageParts msg ) {
	    	enterMonitor();	
			if (fullSlots == capacity) 
				notFull.waitC();	
	      buffer.addLast(msg);
			++fullSlots;
			notEmpty.signalC_and_exitMonitor();
	   }
	   public messageParts withdraw() {
			enterMonitor();
			if (fullSlots == 0) 
				notEmpty.waitC();
	      messageParts msg = (messageParts) buffer.getFirst();
			buffer.removeFirst();
			--fullSlots;
			notFull.signalC_and_exitMonitor();
	      return(msg);
	   }
	}

	public messageParts receive( ) throws TCPChannelException {
		try {
		   synchronized(lock) {
				messageParts msg = buffer.withdraw();
				return msg;
			} // end synchronized lock
		} catch (Exception e) {
			e.printStackTrace(); 
			throw new TCPChannelException(e.getMessage());
		  }
	}
	public synchronized void close() throws TCPChannelException {
		try {
			if (listen != null) {
				closed = true;
				listen.close();
			}
		}
		catch (Exception e) {e.printStackTrace(); throw new TCPChannelException(e.getMessage());}
	}
		
   public void run() {
		try {
			while (true) {
				Socket socket = listen.accept(); // listen for connection requests from senders
				(new connectionHandler(socket)).start();
	  		}
		}
		catch (SocketException e) {if (!closed) e.printStackTrace(); }
   	catch (Exception e) {e.printStackTrace();}
		finally {
    		try { if (listen != null) listen.close(); } 
			catch (IOException e2) { e2.printStackTrace(); }
		}
	}
		
	class connectionHandler extends Thread {
		private ObjectInputStream from; 
		private Socket socket;
		connectionHandler(Socket socket) throws IOException {
			this.socket = socket;
			from = new ObjectInputStream(socket.getInputStream());
		}
		public void run() {
		 try { // run
			messageParts closeMessage = null;	
			while (true) { // read objects until get EOF
				messageParts msg = null;
				try {
					msg = (messageParts) from.readObject(); // read message
					buffer.deposit(msg);
				}
				catch (EOFException e) {
					break; 
				} // end catch
			} // end while true
			from.close(); 
			socket.close(); // issue passive close (after active close by sender)
		} // end try run
		catch (SocketException e) { if (!closed) e.printStackTrace(); }
     	catch (Exception e) {e.printStackTrace();}
		finally {
   		try { if (from != null) from.close(); } 
			catch (IOException e2) { e2.printStackTrace(); }
   		try { if (socket != null) socket.close(); } 
			catch (IOException e3) { e3.printStackTrace(); }
		}
	  } // run
	} // connectionHandler	
}
