import java.util.*;

public final class binarySemaphore extends semaphore {
// a FIFO counting semaphore

	private ArrayList waitingP = new ArrayList(); // queue of threads blocked on P
	private ArrayList waitingV = new ArrayList(); // queue of threads blocked on V

	public binarySemaphore(int initialPermits) {
		super(initialPermits);
		if (initialPermits != 0 && initialPermits != 1)
			throw new IllegalArgumentException("initial value must be 0 or 1");
	}

	public void P() {
		Object conditionVar = new Object(); // each thread blocks on its own conditionVar object
		synchronized (conditionVar) { //lock conditionVar 
			synchronized (this) { // lock semaphore
				if (permits==1)  { // then no need to block thread
						if (waitingV.size() > 0) {  // signal thread blocked in V()
							Object oldest = waitingV.get(0);
							waitingV.remove(0);
							synchronized(oldest) {oldest.notify();}
						}
						else
							permits=0;
						return;
				}
				waitingP.add(conditionVar); // otherwise append blocked thread
			} // end synchronized (this) to avoid conditionVar.wait() while holding lock on this
			try {conditionVar.wait();} catch (InterruptedException ex) {}
		} // end synchronized (conditionVar)
	} // end P()

	public void V() {
		Object conditionVar = new Object(); // each thread blocks on its own conditionVar object
		synchronized (conditionVar) { // lock conditionVar
			synchronized (this) { // lock semaphore
				if (permits==0) { // then no need to block thread
					if (waitingP.size() > 0) {  // signal thread blocked in P()
						Object oldest = waitingP.get(0);
						waitingP.remove(0);
						synchronized(oldest) {oldest.notify();}
					}
					else
						permits=1;
					return;
				}
				waitingV.add(conditionVar); // otherwise append blocked thread
			} // end synchronized (this) to avoid conditionVar.wait() while holding lock on this
			try {conditionVar.wait();} catch (InterruptedException ex) {}
		} // end synchronized (conditionVar)
	} // end V()

	protected boolean doP() {
	// Called by VP() operation; checks permit and unblocks a thread waiting
	// in V() if necessary. Returns true if P() should block; false otherwise.
		if (permits==1)  { // P() does not block
			if (waitingV.size() > 0) {  // signal thread blocked in V()
				Object oldest = waitingV.get(0);
				waitingV.remove(0);
				synchronized(oldest) {oldest.notify();}
			}
			else
				permits=0;
			return false;
		} // P() does not block
		else
			return true;	// P() does block
	}

	public int getSemaphoreID() { return semaphoreID;}

	public void VP(semaphore vSem) {
	// execute {vSem.V(); this.P();} without any intervening P() or V() operations on this or vSem.	

		// lock semaphores in ascending order of IDs to prevent circular deadlock (i.e. T1 holds
		// this's lock and waits for vSem's lock while T2 holds vSem's lock and waits for this's lock.)
		semaphore first = this;
 		semaphore second = vSem;
		if (this.getSemaphoreID() > vSem.getSemaphoreID()) { 
			first = vSem;
			second = this;
		}
		Object conditionVar = new Object(); // each thread blocks on its own conditionVar
		synchronized(conditionVar) {
			synchronized(first) { 
				synchronized(second) { 

					//vSem.V() must not block
					if (vSem instanceof binarySemaphore && vSem.permits == 1)
						throw new IllegalArgumentException("V() will block");

					// perform vSem.V()
					vSem.V(); 		// it's okay that we already hold vSem's lock
					// perform this.P()
					boolean blockingP = doP();
					if (!blockingP)
					return;
					/*		permit--;
							if (permit>=0) // then no need to block thread
								return;
					*/
					waitingP.add(conditionVar); // append blocked thread
				} // end synchronized(second)
			} // end synchronized(first)
			try {conditionVar.wait();} catch (InterruptedException ex) {}
		} // end synchronized (conditionVar)
	} // end VP()

}// end  countingSemaphore

