import java.util.*;
import java.net.*;
import java.io.*;

public class TCPSelectableMailbox implements Runnable, selectable {
	private ServerSocket listen = null;
	private messageBuffer buffer = null;
	Socket socket = null;
	int port;
	String channelName = "";
	ObjectInputStream from = null;  
	private boolean closed = false;
	Object lock = new Object();
	
	private boolean open = false;
	private int ready = 0;
	private IselectiveWait inchoice = null;
	private boolean guard_ = true;
	private static int timeStamp = 0;
	private Vector arrivals = new Vector();

	
   public TCPSelectableMailbox( int port) throws TCPChannelException {
   	this(port,"");
   }
   
   public TCPSelectableMailbox( int port, String channelName ) throws TCPChannelException {
   	this.port = port;
   	this.channelName = channelName;
		try {listen = new ServerSocket( port ); } 
		catch (IOException e) { e.printStackTrace(); throw new TCPChannelException(e.getMessage());}
      buffer = new messageBuffer(100);
		Thread internal = new Thread(this);
		internal.start(); 
	}
	
	class messageBuffer extends monitorSU {  // SU monitor
		private int fullSlots = 0; 			// # of full slots in the buffer
		private int capacity = 0;
		private conditionVariable notFull = new conditionVariable(); 
		private conditionVariable notEmpty = new conditionVariable();
	   private LinkedList buffer;
	   public messageBuffer( int capacity ) {
			this.capacity = capacity;
			buffer = new LinkedList();
   	}
	   public void deposit( messageParts msg ) {
	    	enterMonitor();	
			if (fullSlots == capacity) 
				notFull.waitC();	
	      buffer.addLast(msg);
			++fullSlots;
			notEmpty.signalC_and_exitMonitor();
	   }
	   public messageParts withdraw() {
			enterMonitor();
			if (fullSlots == 0) 
				notEmpty.waitC();
	      messageParts msg = (messageParts) buffer.getFirst();
			buffer.removeFirst();
			--fullSlots;
			notFull.signalC_and_exitMonitor();
	      return(msg);
	   }
	}

	public messageParts receive( ) throws TCPChannelException {
		try {
		   synchronized(lock) {
				messageParts msg = buffer.withdraw();
				decReady();
				arrivals.removeElementAt(0); // vector's methods are synched
				return msg;
			} // end synchronized lock
		} catch (Exception e) {
			e.printStackTrace(); 
			throw new TCPChannelException(e.getMessage());
		  }
	}
	public synchronized void close() throws TCPChannelException {
		try {
			if (listen != null) {
				closed = true;
				listen.close();
			}
		}
		catch (Exception e) {e.printStackTrace(); throw new TCPChannelException(e.getMessage());}
	}
		
   public void run() {
		try {
			while (true) {
				Socket socket = listen.accept(); // listen for connection requests from senders
				(new TCPReceiver(socket)).start();
	  		}
		}
		catch (SocketException e) { if (!closed) e.printStackTrace(); }
   	catch (Exception e) {e.printStackTrace();}
		finally {
    		try { if (listen != null) listen.close(); } 
			catch (IOException e2) { e2.printStackTrace(); }
		}
	}
		
	class TCPReceiver extends Thread {
		private ObjectInputStream from; 
		private Socket socket;
		TCPReceiver(Socket socket) throws IOException {
			this.socket = socket;
			from = new ObjectInputStream(socket.getInputStream());
		}
		public void run() {
		 try { // run
			messageParts closeMessage = null;	
			while (true) { // read objects until get EOF
				messageParts msg = null;
				try {
					msg = (messageParts) from.readObject(); // read message
					buffer.deposit(msg);
					arrivals.addElement(new Integer(getTimeStamp())); // vector's methods are synched
					incReady(); // make sure arrival available before ready is incremented
					if (inchoice != null) {
						synchronized (inchoice) {
							if (open)    // method choice() in selectiveWait is waiting for arrival
							// Method run() could notify select twice before select has chance to run
							// but "extra" notifies do not cause a problem (they are neither "lost" 
                       // nor do they cause select to do extra actions).
								inchoice.notify(); 
						}
					}
				}
				catch (EOFException e) {
					break; 
				} // end catch
			} // end while true
			from.close(); 
			socket.close(); // issue passive close (after active close by sender)
		} // end try run
		// These exceptions (pertaining to trace/test/replay) are not visible to users
		// They happen, e.g., if user program terminates while connected to controller.
		// If trace/test/replay fails, it will be detected by upper levels
		catch (SocketException e) { /*if (!closed) e.printStackTrace();*/ }
     	catch (Exception e) {/*e.printStackTrace();*/}
		finally {
   		try { if (from != null) from.close(); } 
			catch (IOException e2) { e2.printStackTrace(); }
   		try { if (socket != null) socket.close(); } 
			catch (IOException e3) { e3.printStackTrace(); }
		}
	  } // run
	} // TCPReceiver	
	
	private static synchronized int getTimeStamp() {
   // all selectableTCPReceivers share this clock, so FIFO imposed over all arrivals
		return(++timeStamp); 
	}

	public void setSelect(IselectiveWait s) {
		inchoice = s;
	}

	public int getOldestArrival() {
	// only called from choose() when ready>0, i.e., arrivals.size()>0
		return (((Integer)arrivals.firstElement()).intValue()); // vector's methods are synched
	}

	public int count() { return ready; }

    public boolean testReady() {
        return ready>0;
    }

    public synchronized void decReady() {
        --ready;
    }

    public synchronized void incReady() {
        ++ready;
    }

    public void setOpen() {
        open=true;
    }

    public void clearOpen() {
         open=false;
    }

    public void guard(boolean g) {
        guard_=g;
    }

    public boolean testGuard(){
        return guard_;
    }
}
